/*
 * Copyright (c) 1998-2000 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup;

import java.util.*;

import debug.Debug;

/**
 * This class is an helper utility that provides application developers with
 * additional information about members of their current view. It can be
 * used as an hash table, to associate application-dependent information to
 * member identifier. It can be used to obtain information about the state
 * of member identifier with respect to both the current view and previous
 * installed views (for example, method <t>getState</t> may be used to
 * discover which members survived from the previous view). The table maintains
 * also information about the relative position of a member identifier in
 * the member identifier array associated to the last installed view. <p>
 *
 * The table maintains information about a member until it crashes. A member
 * is declared crashed when a member identifier from the same host, but with
 * an higher incarnation identifier is inserted in the table. This means
 * that the JVM hosting the member has crashed, and a new one has started.
 * The state of crashed members is set to <t>CRASHED</t>. The information
 * associated to the member is maintained until the next view change,
 * after which the member identifier and all its associated information
 * is removed from the table. <p>
 *
 * This class implements the <t>MembershipListener</t> interface, so it may be
 * passed to the <t>MembershipService</t> to be notified whenever a view
 * change occurs. Otherwise, the application developer may take care
 * of updating the table by explicitly calling method <t>viewChange</t>.
 *
 * @author  Alberto Montresor
 * @since   Jgroup 0.9
 */
public class MemberTable
implements MembershipListener
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * State assigned to member identifiers not included in the table
   */
  public final static int NOTMEMBER   = 0;

  /**
   *  State assigned to partitioned members for which no information is
   *  available.
   */
  public final static int PARTITIONED = 1;

  /**
   * State assigned to members that have been declared crashed
   * because a member identifier from the same host but with a new
   * incarnation number has been inserted in the table.
   */
  public final static int CRASHED     = 2;

  /**
   * State assigned to member identifiers whose incarnation
   * number is greater than the incarnation number of other
   * members previously contained in the table.
   */
  public final static int RECOVERING  = 3;

  /**
   * State assigned to members that survived from the previous view
   */
  public final static int SURVIVED    = 4;

  /**
   * State assigned to members that merged the view after a previous
   * partitioning
   */
  public final static int MERGING     = 5;

  /**
   * State assigned to new members
   */
  public final static int NEWMEMBER   = 6;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  private HashMap       table;          // Hashtable with MemberIds as keys
  private HashMap       iptable;        // Hashtable with IP addresses as keys


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Build a new empty <code>MemberTable</code>.
   */
  public MemberTable()
  {
    table = new HashMap(101);
    iptable = new HashMap(101);
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////


  /**
   * Returns the state associated to the member identifier. The returned value
   * is one of the integer constants defined in this class.
   */
  synchronized public int getState(MemberId id)
  {
    Record r = (Record) table.get(id);
    if (r != null)
      return r.status;
    else
      return NOTMEMBER;
  }

  /**
   * Returns the position of the specified member identifier
   * in the last view for which method <t>viewChange</t>
   * has been called; return -1 if the member is not
   * member of that view.
   */
  synchronized public int getIndex(MemberId id)
  {
    Record r = (Record) table.get(id);
    if (r != null)
      return r.pos;
    else
      return -1;
  }

  /**
   * Associates object <t>value</t> to the specified member
   * identifier. This association is maintained in the
   * table until it is explicitly removed from the table, or
   * the member is declared crashed. Values may
   * be retrieved using method <t>get</t>.
   */
  synchronized public void put(MemberId id, Object value)
  {
    Record r = (Record) table.get(id);
    if (r != null)
      r.value = value;
    // Cosa fare quando non c'e'? Lanciare qualche eccezione?
  }

  /**
   * Returns the value associated by this <tt>MemberTable</tt>
   * to the specified member id. Returns null if the map contains
   * no mapping for this member id. A return value of null does
   * not necessarily indicate that the map contains no mapping for the
   * key; it is possible that the map explicitly maps the key to null.
   *
   * @param id the member identifier used as key
   * @return the associated value
   */
  synchronized public Object get(MemberId id)
  {
    Record r = (Record) table.get(id);
    if (r != null)
      return r.value;
    else
      return null;
  }

  /**
   *
   */
  synchronized public void remove(MemberId id)
  {
    Record r = (Record) table.get(id);
    if (r != null)
      table.remove(id);
    ArrayList l = (ArrayList) iptable.get(id.getAddress());
    if (l != null) {
      for (int i=0; i < l.size(); i++) {
        r = (Record) l.get(i);
        if (r.id.equals(id)) {
          l.remove(i);
          break;
        }
      }
    }
  }


  /**
   *  Returns an array containing the values associated to all members
   *  contained in this <t>MemberTable<t>, including members
   *  crashed or partitioned.
   */
  synchronized public Object[] elements()
  {
    Object[] elements = new Object[table.size()];
    int i = 0;
    Iterator iterator = table.values().iterator();
    while (iterator.hasNext()) {
      Record r = (Record) iterator.next();
      elements[i++] = r.value;
    }
    return elements;
  }


  /**
   *  Returns an array containing the identifiers of all members
   *  contained in this <t>MemberTable<t>, including members
   *  crashed or partitioned.
   */
  synchronized public MemberId[] members()
  {
    MemberId[] elements = new MemberId[table.size()];
    int i = 0;
    Iterator iterator = table.values().iterator();
    while (iterator.hasNext()) {
      Record r = (Record) iterator.next();
      elements[i++] = r.id;
    }
    return elements;
  }


  /**
   * Returns true if the specified member identifier is
   * member of the last view for which method <t>viewChange</t>
   * has been called; return false otherwise.
   *
   * @param id the member whose membership is tested.
   */
  synchronized public boolean isMember(MemberId id)
  {
    Record r = (Record) table.get(id);
    if (r != null)
      return (r.status != CRASHED && r.status != PARTITIONED);
    else
      return false;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from MembershipListener
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Update the information in the table with the information
   *  maintained in the new view.
   */
  synchronized public void viewChange(View view)
  {
    Record    r;
    ArrayList l;

    if (Debug.DEBUG)
      Debug.println(50, "MemberTable view change");
    // Remove crashed elements from table
    Record[] records = (Record[]) table.values().toArray(new Record[table.size()]);
    for (int i=0; i < records.length; i++) {
      r = records[i];
      r.pos = -1;
      r.prev = r.status;
      r.status = PARTITIONED;
      if (r.prev == CRASHED)
        table.remove(r.id);
    }

    //
    MemberId[] members = view.getMembers();
    for (int i=0; i < members.length; i++) {
      r = (Record) table.get(members[i]);
      if (r != null) {

        /*
         * The member id is already present in the table; if
         * the member was partitioned, now its status is MERGING.
         * In all other cases (NEWMEMBER, RECOVERING, SURVIVED, MERGING),
         * the new status is SURVIVED
         */
        //
        if (r.prev == PARTITIONED)
          r.status = MERGING;
        else
          r.status = SURVIVED;

      } else {

        /*
         * The member id is not present in the table; may be
         * a new member or a recovered member. For each host
         * identifier, we maintain a list of member identifier
         * associated to that host. This is useful to discover
         * if this is just a new member, or it is a recovering
         * member from a previous JVM.
         */
        l = (ArrayList) iptable.get(members[i].getAddress());
        if (l == null) {

          /*
           * First member from that host; new member for sure
           */
          l = new ArrayList();
          iptable.put(members[i].getAddress(), l);
          r = new Record(NEWMEMBER, members[i]);
          l.add(r);
          table.put(members[i], r);

        } else {
          r = (Record) l.get(0);

          if (members[i].isNewer(r.id)) {
            /*
             * New incarnation! Move old member ids to crashed status;
             * Create a new empty record and put it in the tables.
             */
            for (int k=0; k < l.size(); k++) {
              r = (Record) l.get(k);
              r.status = CRASHED;
              r.pos = -1;
            }
            l.clear();
            r = new Record(RECOVERING, members[i]);
            l.add(r);
            table.put(members[i], r);

          } else {
            /*
             *  This member id has the same incarnation number of those
             *  contained in the arraylist, but its member identifier
             *  is not present in the table. If members from the same
             *  host are recovering, so is this. Otherwise, it is a
             *  new member.
             */
            if (r.status == RECOVERING)
              r = new Record(RECOVERING, members[i]);
            else
              r = new Record(NEWMEMBER, members[i]);
            l.add(r);
            table.put(members[i], r);
          }
        }
      }
      r.pos = i;
    }
  }

  /**
   * No effect.
   */
  public void leaved()
  {
  }


  /**
   * No effect.
   */
  public void prepareChange()
  {
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Internal data structure
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Data structure used to maintain an element in the hashtable and its
   *  status.
   */
  private final class Record {

    MemberId  id;         // MemberId to which this value has been
    int       pos;        // Position of the member in the view
    int       status;     // State (RECOVERED, SURVIVED, MERGED, NEWMEMBER, CRASHED)
    int       prev;       // Previous status
    Object    value;      // The object stored by users of this data structure

    Record(int status, MemberId id) {
      this.id     = id;
      this.status = status;
      this.pos    = -1;
      this.prev   = status;
      this.value  = null;
    }

  } // END Record

} // END MemberTable









