/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are subject to the Sun Community Source License
 * v 3.0/Jini Technology Specific Attachment v 1.0 (the "License"). You may not
 * use this file except in compliance with the License. You may obtain a copy
 * of the License at http://www.sun.com/jini/ . Software distributed under the
 * License is distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing rights and limitations under the License.
 * 
 * The Reference Code is Jini Technology Core Platform code, v 1.1. The
 * Developer of the Reference Code is Sun Microsystems, Inc.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * The contents of this file comply with the Jini Technology Core Platform
 * Compatibility Kit, v 1.1A.
 * 
 * Tester(s): Sun Microsystems, Inc.
 * 
 * Test Platform(s):
 * 
 * 	Java 2 SDK, Standard Edition, V 1.2.2_006 Solaris
 * 	   Reference Implementation Release
 * 
 * 	Java 2 SDK, Standard Edition, V 1.2.2_05a Solaris
 * 	   Production Release
 * 
 * 	Java 2 SDK, Standard Edition, V 1.2.2_006 Windows 95/98/NT
 * 	   Production Release
 * 	   
 * Version 1.1
 * 
 */
package jgroup.jini;

import net.jini.core.lease.*;
import net.jini.core.entry.Entry;
import net.jini.core.discovery.*;
import net.jini.core.lookup.*;
import net.jini.discovery.*;
import net.jini.lease.*;
import net.jini.lookup.ServiceIDListener;

import java.io.IOException;
import java.rmi.RemoteException;
import java.util.*;
import com.sun.jini.lookup.entry.LookupAttributes;

/**
 * A goal of any well-behaved service is to advertise the facilities and
 * functions it provides by requesting residency within at least one lookup
 * service. Making such a request of a lookup service is known as registering
 * with, or "joining", a lookup service. To demonstrate this good behavior,
 * a service must comply with both the multicast discovery protocol and the
 * unicast discovery protocol in order to discover the lookup services it is
 * interested in joining. The service must also comply with the join protocol
 * to register with the desired lookup services. 
 * <p>
 * In order for a service to maintain its residency in the lookup services
 * it has joined, the service must provide for the coordination, systematic
 * renewal, and overall management of all leases on that residency. In
 * addition to handling all discovery and join duties, as well as managing
 * all leases on lookup service residency, the service must also provide
 * for the coordination and management of any attribute sets with which
 * it may have registered with the lookup services in which it resides.
 * <p>
 * This class performs all of the functions related to discovery, joining,
 * service lease renewal, and attribute management which is required of a
 * well-behaved service. Each of these activities is intimately involved
 * with the maintenance of a service's residency in one or more lookup
 * services (the service's join state), thus the name <code>JoinManager</code>.
 * <p>
 * This class should be employed by services, not clients. The use of this
 * class in a wide variety of services can help minimize the work resulting
 * from having to repeatedly implement this required functionality in each
 * service. Note that this class is not remote. Services that wish to use
 * this class will create an instance of this class in the service's address
 * space to manage the entity's join state locally.
 * 
 * @see net.jini.discovery.DiscoveryManagement
 * @see net.jini.lease.LeaseRenewalManager
 *
 * @author Zane Pan
 */
public class JoinManager {
	private ServiceIDListener callback;
	
	private  Entry[] lookupAttr = null;
	private ServiceItem serviceItem;
	private final ArrayList joinSet = new ArrayList();
	private final ArrayList listeners = new ArrayList();
	private final ArrayList task = new ArrayList();
	private TaskThread taskThread = null;
	private LeaseRenewalManager leaseRenewalMgr = null;
	private DiscoveryManagement  discMgr = null;
	private DiscMgrListener discMgrListener = new DiscMgrListener();
	private boolean bTerminated = false;
	private boolean bCreateDiscMgr = false;
	private class DiscMgrListener implements DiscoveryListener {
		public void discovered(DiscoveryEvent e) {
			//System.out.println("discovered..." + from);
			ServiceRegistrar[] proxys = (ServiceRegistrar[])e.getRegistrars();
			RegisterTask t = new RegisterTask(proxys);
			createTaskThread(t);
			
		}
		
		public void discarded(DiscoveryEvent e) {
			//System.out.println("remove...");
			ServiceRegistrar[] proxys = (ServiceRegistrar[])e.getRegistrars();
			DiscardProxyTask t = 
				new DiscardProxyTask(proxys);
			createTaskThread(t);
		}
	}
	
	/** 
	 * Constructs an instance of this class that will notify the service
	 * when a service ID is assigned by a lookup service with which the
	 * service is registered. This constructor is typically used by services
	 * that have not yet been assigned a service ID, but which have been
	 * pre-configured to join lookup services that the service identifies
	 * through the initialization of a discovery manager.
	 *
	 * @param obj         a reference to the service requesting the services
	 *                    of this class.
	 * @param attrSets    array of <code>Entry</code> consisting of the
	 *                    attribute sets with which to register the service.
	 * @param callback    a reference to an object that implements the
	 *                    <code>ServiceIDListener</code> interface, which 
	 *                    will be sent an event containing the service ID
	 *                    assigned to the service by the lookup service.
	 * @param discoverMgr a reference to an object that implements the 
	 *                    <code>DiscoveryManagement</code> interface which 
	 *                    this class will use to manage the service's 
	 *                    discovery duties. If <code>null</code> is input
	 *                    to this argument, an instance of the utility class
	 *                    <code>LookupDiscoveryManager</code> will be used to
	 *                    listen for events announcing the discovery of only
	 *                    those lookup services that are members of the public
	 *                    group.
	 * @param leaseMgr    an instance of the <code>LeaseRenewalManager</code>
	 *                    utility class. If <code>null</code> is input
	 *                    to this argument, an instance of the
	 *                    <code>LeaseRenewalManager</code> class will be
	 *                    created, initially managing no leases. This 
	 *                    feature allows a service that employs the 
	 *                    <code>JoinManager</code> to use either a single
	 *                    entity to manage all of its leases, or to use
	 *                    separate entities: one to manage the leases
	 *                    unrelated to the join process, and one to manage the
	 *                    leases that result from the join process and which
	 *                    are accessible only within the 
	 *
	 * @throws java.io.IOException because construction of this class may
	 *         initiate the discovery process, which can throw an
	 *         <code>IOException</code> when socket allocation occurs.
	 *
	 * @see net.jini.lookup.ServiceIDListener
	 * @see net.jini.discovery.DiscoveryManagement
	 * @see net.jini.discovery.LookupDiscoveryManager
	 * @see net.jini.lease.LeaseRenewalManager
	 */
	public JoinManager(Object obj,
		Entry[] attrSets,
		ServiceIDListener callback,
		DiscoveryManagement discoverMgr,
		LeaseRenewalManager leaseMgr)    throws IOException
	{
		discMgr = discoverMgr;
		createJoinManager(null, obj, attrSets, callback, leaseMgr);	
	}
	
	/** 
	 * Constructs an instance of this class that will register the service,
	 * using the supplied service ID, with the lookup services indicated
	 * through the <code>discoveryMgr</code> parameter. This constructor is
	 * typically used by services that have already been assigned a service
	 * ID (possibly by the service provider itself or as a result of a prior
	 * registration with some lookup service), and which may or may not have
	 * been pre-configured to join lookup services identified by group or by
	 * specific location.
	 * <p>
	 * Note that this form of the constructor applies the same semantics
	 * to the <code>attrSets</code>, <code>discoveryMgr</code>, and
	 * <code>leaseMgr</code> parameters as is applied by the first form
	 * of the constructor.
	 *
	 * @param obj         a reference to the service requesting the services
	 *                    of this class.
	 * @param attrSets    array of <code>Entry</code> consisting of the
	 *                    attribute sets with which to register the service.
	 * @param serviceID   an instance of <code>ServiceID</code> with which to
	 *                    register the service with all desired lookup services
	 * @param discoverMgr a reference to an object that implements the 
	 *                    <code>DiscoveryManagement</code> interface which 
	 *                    this class will use to manage the service's 
	 *                    discovery duties. If <code>null</code> is input
	 *                    to this argument, an instance of the utility class
	 *                    <code>LookupDiscoveryManager</code> will be used to
	 *                    listen for events announcing the discovery of only
	 *                    those lookup services that are members of the public
	 *                    group.
	 * @param leaseMgr    an instance of the <code>LeaseRenewalManager</code>
	 *                    utility class. If <code>null</code> is input
	 *                    to this argument, an instance of the
	 *                    <code>LeaseRenewalManager</code> class will be
	 *                    created, initially managing no leases. This 
	 *                    feature allows a service that employs the 
	 *                    <code>JoinManager</code> to use either a single
	 *                    entity to manage all of its leases, or to use
	 *                    separate entities: one to manage the leases
	 *                    unrelated to the join process, and one to manage the
	 *                    leases that result from the join process and which
	 *                    are accessible only within the 
	 *
	 * @throws java.io.IOException because construction of this class may
	 *         initiate the discovery process, which can throw an
	 *         <code>IOException</code> when socket allocation occurs.
	 *
	 * @see net.jini.core.lookup.ServiceID
	 * @see net.jini.discovery.DiscoveryManagement
	 * @see net.jini.discovery.LookupDiscoveryManager
	 * @see net.jini.lease.LeaseRenewalManager
	 */
	public JoinManager(Object obj,
		Entry[] attrSets,
		ServiceID serviceID,
		DiscoveryManagement discoverMgr,
		LeaseRenewalManager leaseMgr)    throws IOException
	{
		discMgr = discoverMgr;
		createJoinManager(serviceID, obj, attrSets, callback, leaseMgr);
	}
	
	private void discoveredNotifyListener(DiscoveryListener l,
		ArrayList set)
		
	{
		DiscoveryEvent evt =  
			new DiscoveryEvent(this, buildServiceRegistrar(set));
		l.discovered(evt);
	}
	
	private ServiceRegistrar[] buildServiceRegistrar(ArrayList set) {
		int k = 0;
		ServiceRegistrar[] proxys = 
			new ServiceRegistrar[set.size()];
		Iterator iter = set.iterator();
		while(iter.hasNext()) {
			ProxyReg reg = (ProxyReg)iter.next();
			proxys[k++] = reg.proxy;
		}
		return proxys;
	}
	
	private void createJoinManager(ServiceID serviceID, 
		Object obj,Entry[] attrSets, 
		ServiceIDListener callback,
		LeaseRenewalManager leaseMgr) 
		throws IOException
	{
		if(!(obj instanceof java.io.Serializable) && 
			!(obj instanceof jgroup.ExternalGMIListener)) {
			if(obj instanceof java.rmi.Remote) {
				try {
					java.rmi.server.RemoteObject.toStub((java.rmi.Remote)obj);
				} catch (java.rmi.NoSuchObjectException e) {
					throw new IllegalArgumentException
						("obj must be Serializable or a RemoteStub "
						+"or an exported RemoteObject");
				}
			}else
				throw new IllegalArgumentException
				("obj must be Serializable or a RemoteStub "
				+"or an exported RemoteObject");
		}		
		
		if(attrSets == null)
			lookupAttr = new Entry[0];
		else
			lookupAttr = (Entry[])attrSets.clone();
		
		serviceItem = new ServiceItem(serviceID, obj, lookupAttr);
		leaseRenewalMgr = ( leaseMgr != null) ? leaseMgr :
			new LeaseRenewalManager();
		this.callback = callback;
		if(discMgr == null) {
			bCreateDiscMgr = true;
			/* Create a discovery manager configured for the public group */
			discMgr = new LookupDiscoveryManager(new String[] {""},null,null);
		}
		discMgr.addDiscoveryListener(discMgrListener);
	}
	
	/** 
	 * Returns the instance of <code>DiscoveryManagement</code> that was
	 * either passed into the constructor, or that was created as a result
	 * of <code>null</code> being input to that parameter.
	 * <p>
	 * The object returned by this method encapsulates the mechanism by which
	 * either the <code>JoinManager</code> or the entity itself can set
	 * discovery listeners and discard previously discovered lookup services
	 * when they are found to be unavailable.
	 *
	 * @return the instance of the <code>DiscoveryManagement</code> interface
	 *         that was either passed into the constructor, or that was
	 *         created as a result of <code>null</code> being input to that
	 *         parameter.
	 *
	 * @see net.jini.discovery.DiscoveryManagement
	 * @see net.jini.discovery.LookupDiscoveryManager
	 */
	public DiscoveryManagement getDiscoveryManager(){
		return discMgr; 
		
	}
	
	/** returns the LeaseRenewalManager */
	/** 
	 * Returns the instance of the <code>LeaseRenewalManager</code> class
	 * that was either passed into the constructor, or that was created
	 * as a result of <code>null</code> being input to that parameter.
	 * <p>
	 * The object returned by this method manages the leases requested and
	 * held by the <code>JoinManager</code>. Although it may also manage
	 * leases unrelated to the join process that are requested and held by
	 * the service itself, the leases with which the <code>JoinManager</code>
	 * is concerned are the leases that correspond to the service registration
	 * requests made with each lookup service the service wishes to join.
	 *
	 * @return the instance of the <code>LeaseRenewalManager</code> class
	 *         that was either passed into the constructor, or that was
	 *         created as a result of <code>null</code> being input to that
	 *         parameter.
	 *
	 * @see net.jini.discovery.DiscoveryManagement
	 * @see net.jini.lease.LeaseRenewalManager
	 */
	public LeaseRenewalManager getLeaseRenewalManager(){
		return leaseRenewalMgr; 
		
	}
	
	/** 
	 * Returns an array of <code>ServiceRegistrar</code> objects, each
	 * corresponding to a lookup service with which the service is currently
	 * registered (joined). If there are no lookup services with which the
	 * service is currently registered, this method returns the empty array.
	 * This method returns a new array upon each invocation.
	 *
	 * @return array of instances of <code>ServiceRegistrar</code>, each
	 *         corresponding to a lookup service with which the service is
	 *         currently registered 
	 *
	 * @see net.jini.core.lookup.ServiceRegistrar
	 */
	public ServiceRegistrar[] getJoinSet() {
		synchronized(joinSet) {
			ServiceRegistrar[] ret = new ServiceRegistrar[joinSet.size()];
			int k = 0;
			for (Iterator iter = joinSet.iterator(); iter.hasNext(); ) {
				ret[k++] = ((ProxyReg)iter.next()).proxy;
			}
			return ret;
		}
	}    
	
	/** 
	 * Returns an array containing the set of attributes currently associated
	 * with the service. If the service is not currently associated with an
	 * attribute set, this method returns the empty array. This method returns
	 * a new array upon each invocation.
	 *
	 * @return array of instances of <code>Entry</code> consisting of the
	 *         set of attributes with which the service is registered in
	 *         each lookup service that it has joined
	 *
	 * @see net.jini.core.entry.Entry
	 * @see #setAttributes
	 */
	public Entry[] getAttributes() {
		synchronized(serviceItem) {
			return (Entry[])lookupAttr.clone();
		}
	}
	
	/** 
	 * Associates a new set of attributes with the service, in addition to
	 * the service's current set of attributes. The association of this new
	 * set of attributes with the service will be propagated to each lookup
	 * service with which the service is registered. Note that this
	 * propagation is performed asynchronously, thus there is no guarantee
	 * that the propagation of the attributes to all lookup services with
	 * which the service is registered will have completed upon return from
	 * this method.
	 * <p>
	 * An invocation of this method with duplicate elements in the 
	 * <code>attrSets</code> parameter (where duplication means attribute
	 * equality as defined by calling the <code>MarshalledObject.equals</code>
	 * method on field values) is equivalent to performing the invocation
	 * with the duplicates removed from that parameter.
	 * <p>
	 * Note that because there is no guarantee that attribute propagation
	 * will have completed upon return from this method, services that 
	 * invoke this method must take care not to modify the contents of the 
	 * <code>attrSets</code> parameter. Doing so could cause the service's
	 * attribute state to be corrupted or inconsistent on a subset of the
	 * lookup services with which the service is registered as compared with
	 * the state reflected on the remaining lookup services. It is for this
	 * reason that the effects of modifying the contents of the
	 * <code>attrSets</code> parameter, after this method is invoked, are
	 * undefined.
	 *
	 * @param attrSets array of <code>Entry</code> consisting of the
	 *                 attribute sets with which to augment the service's
	 *                 current set of attributes
	 *
	 * @throws java.lang.NullPointerException this exception occurs when
	 *         either <code>null</code> is input to the <code>attrSets</code>
	 *         parameter, or one or more of the elements of the
	 *         <code>attrSets</code> parameter is <code>null</code>.
	 *
	 * @see net.jini.core.entry.Entry
	 */
	public void addAttributes(Entry[] attrSets) {
		addAttributes(attrSets, false);
	}
	
	/** 
	 * Associates a new set of attributes with the service, in addition to
	 * the service's current set of attributes. The association of this new
	 * set of attributes with the service will be propagated to each lookup
	 * service with which the service is registered. Note that this
	 * propagation is performed asynchronously, thus there is no guarantee
	 * that the propagation of the attributes to all lookup services with
	 * which the service is registered will have completed upon return from
	 * this method.
	 * <p>
	 * An invocation of this method with duplicate elements in the 
	 * <code>attrSets</code> parameter (where duplication means attribute
	 * equality as defined by calling the <code>MarshalledObject.equals</code>
	 * method on field values) is equivalent to performing the invocation
	 * with the duplicates removed from that parameter.
	 * <p>
	 * Note that because there is no guarantee that attribute propagation
	 * will have completed upon return from this method, services that 
	 * invoke this method must take care not to modify the contents of the 
	 * <code>attrSets</code> parameter. Doing so could cause the service's
	 * attribute state to be corrupted or inconsistent on a subset of the
	 * lookup services with which the service is registered as compared with
	 * the state reflected on the remaining lookup services. It is for this
	 * reason that the effects of modifying the contents of the
	 * <code>attrSets</code> parameter, after this method is invoked, are
	 * undefined.
	 * <p>
	 * A service typically employs this version of <code>addAttributes</code> 
	 * to prevent clients or other services from attempting to add what are
	 * referred to as "service controlled attributes" to the service's set.
	 * A service controlled attribute is an attribute that implements the
	 * <code>ServiceControlled</code> marker interface.
	 * <p>
	 * Consider a printer service. With printers, there are often times error
	 * conditions, that only the printer can detect (for example, a paper
	 * jam or a toner low condition). To report conditions such as these to
	 * interested parties, the printer typically adds an attribute to its
	 * attribute set, resulting in an event being sent that notifies clients
	 * that have registered interest in such events. When the condition is
	 * corrected, the printer would then remove the attribute from its set
	 * by invoking the <code>modifyAttributes</code> method in the appropriate
	 * manner.
	 * <p>
	 * Attributes representing conditions that only the service can know about
	 * or control are good candidates for being defined as service controlled
	 * attributes. That is, the service provider (the developer of the printer
	 * service for example) would define the attributes that represent
	 * conditions such as those just described to implement the
	 * <code>ServiceControlled</code> marker interface. Thus, when other
	 * entities attempt to add new attributes, services that wish to employ
	 * such service controlled attributes should ultimately invoke only this
	 * version of <code>addAttributes</code> (with the <code>checkSC</code>
	 * parameter set to <code>true</code>), resulting in a
	 * <code>SecurityException</code> if any of the attributes being added
	 * happen to be service controlled attributes. In this way, only the
	 * printer itself would be able to set a "paper jammed" or "toner low"
	 * attribute, not some arbitrary client.
	 *
	 * @param attrSets array of <code>Entry</code> consisting of the
	 *                 attribute sets with which to augment the service's
	 *                 current set of attributes
	 * @param checkSC  <code>boolean</code> flag indicating whether the
	 *                 elements of the set of attributes to add should be
	 *                 checked to determine if they are service controlled
	 *                 attributes
	 *
	 * @throws java.lang.NullPointerException this exception occurs when
	 *         either <code>null</code> is input to the <code>attrSets</code>
	 *         parameter, or one or more of the elements of the
	 *         <code>attrSets</code> parameter is <code>null</code>.
	 *
	 * @throws java.lang.SecurityException this exception occurs when
	 *         the <code>checkSC</code> parameter is <code>true</code>, and
	 *         at least one of the attributes to be added is an instance
	 *         of the <code>ServiceControlled</code> marker interface.
	 *
	 * @see net.jini.core.entry.Entry
	 * @see net.jini.lookup.entry.ServiceControlled
	 */
	public void addAttributes(Entry[] attrSets, boolean checkSC) {
		Task t;
		synchronized(serviceItem) {
			lookupAttr = LookupAttributes.add(lookupAttr, attrSets, checkSC);
			t = new addAttributesTask(attrSets, lookupAttr);
		}
		createTaskThread(t);
	}
	
	/** 
	 * Replaces the service's current set of attributes with a new set of
	 * attributes. The association of this new set of attributes with the
	 * service will be propagated to each lookup service with which the
	 * service is registered. Note that this propagation is performed
	 * asynchronously, thus there is no guarantee that the propagation of
	 * the attributes to all lookup services with which the service is
	 * registered will have completed upon return from this method.
	 * <p>
	 * An invocation of this method with duplicate elements in the 
	 * <code>attrSets</code> parameter (where duplication means attribute
	 * equality as defined by calling the <code>MarshalledObject.equals</code>
	 * method on field values) is equivalent to performing the invocation
	 * with the duplicates removed from that parameter.
	 * <p>
	 * Note that because there is no guarantee that attribute propagation
	 * will have completed upon return from this method, services that 
	 * invoke this method must take care not to modify the contents of the 
	 * <code>attrSets</code> parameter. Doing so could cause the service's
	 * attribute state to be corrupted or inconsistent on a subset of the
	 * lookup services with which the service is registered as compared with
	 * the state reflected on the remaining lookup services. It is for this
	 * reason that the effects of modifying the contents of the
	 * <code>attrSets</code> parameter, after this method is invoked, are
	 * undefined.
	 *
	 * @param attrSets array of <code>Entry</code> consisting of the
	 *                 attribute sets with which to replace the service's
	 *                 current set of attributes
	 *
	 * @throws java.lang.NullPointerException this exception occurs when
	 *         at least one of the elements of the <code>attrSets</code>
	 *         parameter is <code>null</code>.
	 *
	 * @see net.jini.core.entry.Entry
	 * @see #getAttributes
	 */
	public void setAttributes(Entry[] attrSets){
		Task t;
		synchronized(serviceItem) {
			lookupAttr = (Entry[]) attrSets.clone();
			t = new setAttributesTask(lookupAttr);
		}
		createTaskThread(t);
	}
	
	/** 
	 * Changes the service's current set of attributes using the same
	 * semantics as the <code>modifyAttributes</code> method of the
	 * <code>ServiceRegistration</code> class.
	 * <p>
	 * The association of the new set of attributes with the service will
	 * be propagated to each lookup service with which the service is
	 * registered. Note that this propagation is performed asynchronously,
	 * thus there is no guarantee that the propagation of the attributes to
	 * all lookup services with which the service is registered will have
	 * completed upon return from this method.
	 * <p>
	 * Note that if the length of the array containing the templates does
	 * not equal the length of the array containing the modifications, an
	 * <code>IllegalArgumentException</code> will be thrown and propagated
	 * through this method.
	 * <p>
	 * Note also that because there is no guarantee that attribute propagation
	 * will have completed upon return from this method, services that 
	 * invoke this method must take care not to modify the contents of the 
	 * <code>attrSets</code> parameter. Doing so could cause the service's
	 * attribute state to be corrupted or inconsistent on a subset of the
	 * lookup services with which the service is registered as compared with
	 * the state reflected on the remaining lookup services. It is for this
	 * reason that the effects of modifying the contents of the
	 * <code>attrSets</code> parameter, after this method is invoked, are
	 * undefined.
	 *
	 * @param attrSetTemplates array of <code>Entry</code> used to identify
	 *                         which elements to modify from the service's
	 *                         current set of attributes
	 * @param attrSets         array of <code>Entry</code> containing the
	 *                         actual modifications to make in the matching
	 *                         sets found using the 
	 *                         <code>attrSetTemplates</code> parameter
	 *
	 * @throws java.lang.IllegalArgumentException this exception occurs when
	 *         the array containing the templates does not equal the length
	 *         of the array containing the modifications.
	 *
	 * @see net.jini.core.entry.Entry
	 * @see net.jini.core.lookup.ServiceRegistration#modifyAttributes
	 */
	public void modifyAttributes(Entry[] attrSetTemplates, Entry[] attrSets) {
		modifyAttributes(attrSetTemplates, attrSets, false);
	}
	
	/** 
	 * Changes the service's current set of attributes using the same
	 * semantics as the <code>modifyAttributes</code> method of the
	 * <code>ServiceRegistration</code> class.
	 * <p>
	 * The association of the new set of attributes with the service will
	 * be propagated to each lookup service with which the service is
	 * registered. Note that this propagation is performed asynchronously,
	 * thus there is no guarantee that the propagation of the attributes to
	 * all lookup services with which the service is registered will have
	 * completed upon return from this method.
	 * <p>
	 * Note that if the length of the array containing the templates does
	 * not equal the length of the array containing the modifications, an
	 * <code>IllegalArgumentException</code> will be thrown and propagated
	 * through this method.
	 * <p>
	 * Note also that because there is no guarantee that attribute propagation
	 * will have completed upon return from this method, services that 
	 * invoke this method must take care not to modify the contents of the 
	 * <code>attrSets</code> parameter. Doing so could cause the service's
	 * attribute state to be corrupted or inconsistent on a subset of the
	 * lookup services with which the service is registered as compared with
	 * the state reflected on the remaining lookup services. It is for this
	 * reason that the effects of modifying the contents of the
	 * <code>attrSets</code> parameter, after this method is invoked, are
	 * undefined.
	 * <p>
	 * A service typically employs this version of 
	 * <code>modifyAttributes</code> to prevent clients or other services
	 * from attempting to modify what are referred to as "service controlled
	 * attributes" in the service's set. A service controlled attribute is an
	 * attribute that implements the <code>ServiceControlled</code> marker
	 * interface.
	 * <p>
	 * Attributes representing conditions that only the service can know about
	 * or control are good candidates for being defined as service controlled
	 * attributes. When other entities attempt to modify a service's
	 * attributes, if the service wishes to employ such service controlled
	 * attributes, the service should ultimately invoke only this version 
	 * of <code>modifyAttributes</code> (with the <code>checkSC</code>
	 * parameter set to <code>true</code>), resulting in a
	 * <code>SecurityException</code> if any of the attributes being modified
	 * happen to be service controlled attributes.
	 *
	 * @param attrSetTemplates array of <code>Entry</code> used to identify
	 *                         which elements to modify from the service's
	 *                         current set of attributes
	 * @param attrSets         array of <code>Entry</code> containing the
	 *                         actual modifications to make in the matching
	 *                         sets found using the 
	 *                         <code>attrSetTemplates</code> parameter
	 * @param checkSC          <code>boolean</code> flag indicating whether the
	 *                         elements of the set of attributes to modify
	 *                         should be checked to determine if they are
	 *                         service controlled attributes
	 *
	 * @throws java.lang.IllegalArgumentException this exception occurs when
	 *         the array containing the templates does not equal the length
	 *         of the array containing the modifications.
	 *
	 * @throws java.lang.SecurityException this exception occurs when
	 *         the <code>checkSC</code> parameter is <code>true</code>, and
	 *         at least one of the attributes to be modified is an instance
	 *         of the <code>ServiceControlled</code> marker interface.
	 *
	 * @see net.jini.core.entry.Entry
	 * @see net.jini.core.lookup.ServiceRegistration#modifyAttributes
	 * @see net.jini.lookup.entry.ServiceControlled
	 */
	public void modifyAttributes(Entry[] attrSetTemplates,
		Entry[] attrSets,
		boolean checkSC)
	{
		Task t;
		synchronized(serviceItem) {
			lookupAttr = LookupAttributes.modify(lookupAttr,
				attrSetTemplates, attrSets, checkSC);
			t = new modifyAttributesTask(attrSetTemplates,
				attrSets,lookupAttr);
		}
		createTaskThread(t);
	}
	
	/**
	 * Performs cleanup duties related to the termination of the lookup
	 * service discovery event mechanism, as well as the lease and 
	 * thread management performed by the <code>JoinManager</code>. This
	 * method will cancel all of the service's managed leases that were
	 * granted by the lookup services with which the service is registered,
	 * and will terminate all threads that have been created.
	 * <p>
	 * Note that if the discovery manager employed by the instance of this
	 * class that is being terminated was created by the instance itself,
	 * this method will terminate all discovery processing being performed by
	 * that manager object on behalf of the service; otherwise, the discovery
	 * manager supplied by the service is still valid.
	 * <p>
	 * Whether an instance of the <code>LeaseRenewalManager</code> class was
	 * supplied by the service or created by the <code>JoinManager</code>
	 * itself, any reference to that object obtained by the service prior to
	 * termination will still be valid after termination.
	 * Note also this class makes certain concurrency guarantees with respect
	 * to an invocation of the terminate method while other method invocations
	 * are in progress. The termination process will not begin until
	 * completion of all invocations of the methods defined in the public
	 * interface of this class. Furthermore, once the termination process has
	 * begun, no further remote method invocations will be made by this class,
	 * and all other method invocations made on this class will not return
	 * until the termination process has completed.
	 * <p>
	 * Upon completion of the termination process, the semantics of all
	 * current and future method invocations on the instance of this class
	 * that was just terminated are undefined; although the reference to the
	 * <code>LeaseRenewalManager</code> object employed by that instance
	 * of <code>JoinManager</code> is still valid.
	 */
	public void terminate() {
		discMgr.removeDiscoveryListener(discMgrListener);
		if(bCreateDiscMgr)
			discMgr.terminate();
		TaskThread tmpThread;
		synchronized(task) {
			bTerminated = true;
			tmpThread = taskThread;
		}
		if(tmpThread != null) {
			tmpThread.interrupt();
			try {
				tmpThread.join();
			} catch (Exception e) {
			}
		}
		
		synchronized(joinSet) {
			for (Iterator iter = joinSet.iterator(); iter.hasNext(); ) {
				try {
					leaseRenewalMgr.cancel(
						((ProxyReg)iter.next()).srvcRegistration.getLease());
				} catch (Exception ex) {
				}
			}
			joinSet.clear();
		}
	}
	
	private void createTaskThread(Task t) {
		synchronized (task) {
			if(bTerminated )
				return;
			task.add(t);
			if(taskThread == null ) {
				taskThread =  new TaskThread();
				taskThread.start();
			}
		}
	}
	
	private interface Task {
		void exec();
	}
	
	private class ProxyReg  {
		private class DiscLeaseListener implements LeaseListener {
			public void notify(LeaseRenewalEvent e ) {
				//System.out.println("**********************lease expired");
				if (e.getException() instanceof UnknownLeaseException) {
					LeaseExpireNotifyTask t =
						new LeaseExpireNotifyTask(ProxyReg.this);
					createTaskThread(t);
				} else 
					fail(e.getException());
			}
			
		}
		public ServiceRegistrar proxy;
		public ServiceRegistration srvcRegistration = null;
		private DiscLeaseListener dListener = new DiscLeaseListener();
		public ProxyReg(ServiceRegistrar proxy) {
			if(proxy == null)
				throw new IllegalArgumentException(
				"proxy can't be null");
			this.proxy = proxy;
		}	    
		
		public boolean equals(Object obj) {
			if (obj instanceof ProxyReg){
				return proxy.equals(((ProxyReg)obj).proxy);
			} else return false;
		}
		
		public int hashCode() {
			return proxy.hashCode();
		}
		
		public void register() {
			try {
				if( proxy == null )
					throw new RuntimeException("proxy is null");
				//System.out.println("register:"+proxy.getLocator().getHost());
				
				srvcRegistration = proxy.register(serviceItem,
					Long.MAX_VALUE);
				Lease lease = srvcRegistration.getLease();
				leaseRenewalMgr.renewUntil(srvcRegistration.getLease(),
					Long.MAX_VALUE, dListener);
				addJoinSet(this);
				if(serviceItem.serviceID == null ) {
					serviceItem.serviceID = srvcRegistration.getServiceID();
					if(callback != null )
						callback.serviceIDNotify(serviceItem.serviceID);
				}
				// System.out.println("register...succeed");
			} catch (Exception e) {
				fail(e);
			}
		}
		
		public void addAttributes(Entry[] attSet) {
			try {
				srvcRegistration.addAttributes(attSet);
			} catch (Exception e) {
				fail(e);
			}		    
		}
		
		public void modifyAttributes(Entry[] templ, Entry[] attSet) {
			try {
				srvcRegistration.modifyAttributes(templ, attSet);
			} catch (Exception e) {
				fail(e);
			}
		}		    
		
		public void setAttributes(Entry[] attSet) {
			try {
				srvcRegistration.setAttributes(attSet);
			} catch (Exception e) {
				fail(e);
			}		    
		}
		
		public void fail(Throwable e) {
			e.printStackTrace();
			try {
				if (System.getProperty("com.sun.jini.join.debug") != null)
					e.printStackTrace();
			} catch (SecurityException ee) {
			}
			discMgr.discard(proxy);
		}
	}
	
	private class addAttributesTask implements Task {
		protected Entry[] attrSets, finalAttrSet;
		addAttributesTask(Entry[] attrSetsToAdd, Entry[] finalAttrSet ) {
			this.attrSets = (Entry[])attrSetsToAdd.clone();
			this.finalAttrSet = finalAttrSet;
		}
		
		protected void doAttributes(ProxyReg reg) {
			reg.addAttributes(attrSets);
		}
		
		public void exec() {
			for (Iterator iter = joinSet.iterator(); iter.hasNext(); ) {
				ProxyReg reg =(ProxyReg)iter.next();
				doAttributes(reg);
			}
			serviceItem.attributeSets = finalAttrSet;
		}
	}
	
	private final class setAttributesTask extends addAttributesTask {
		setAttributesTask(Entry[] finalAttrSet) {
			super(finalAttrSet, finalAttrSet);
		}
		
		protected void doAttributes(ProxyReg reg) {
			reg.setAttributes(attrSets);
		}
	}
	
	private final class modifyAttributesTask extends addAttributesTask {
		private Entry[] attrSetTemplate;
		modifyAttributesTask(Entry[] attrSetTemplate,
			Entry[] attrSets, Entry[] finalAttrSet ) {
			super(attrSets, finalAttrSet );
			this.attrSetTemplate = (Entry[])attrSetTemplate.clone();
		}
		
		protected void doAttributes(ProxyReg reg) {
			reg.modifyAttributes(attrSetTemplate, attrSets);
		}
	}
	
	private class RegisterTask implements Task {
		private ServiceRegistrar proxys[];
		RegisterTask(ServiceRegistrar proxys[]){
			this.proxys = proxys;
		}
		
		public void exec() {
			for(int i=0; i<proxys.length; i++) {
				//if the serviceItem is a lookup service, don't register.
				if(!proxys[i].equals(serviceItem.service)) {
					ProxyReg reg = new ProxyReg(proxys[i]);
					if(!joinSet.contains(reg))
						reg.register();
				}
			}
		}
		
	}
	
	private class LeaseExpireNotifyTask implements Task {
		protected ProxyReg reg;
		LeaseExpireNotifyTask(ProxyReg reg){
			this.reg = reg;
		}
		
		public void exec() {
			if(joinSet.contains(reg))
				reg.register();
		}
		
	}
	
	private class DiscardProxyTask implements Task {
		private ServiceRegistrar proxys[];
		public DiscardProxyTask(ServiceRegistrar[] proxys){
			this.proxys = proxys;
		}
		
		public void exec() {
			for(int i=0; i<proxys.length; i++) {
				ProxyReg reg = findReg(proxys[i]);
				if(reg != null ) 
					removeJoinSet(reg);
				
			}
		}
	}
	
	private class TaskThread extends Thread {
		public TaskThread() {
			super("taskThread");
			setDaemon(true);
		}
		
		public void run(){
			while(!isInterrupted()) {
				Task t;
				synchronized (task) {
					if(task.isEmpty()) {
						taskThread = null;
						return;
					}
					t = (Task)task.remove(0);
				}
				t.exec(); 
			}
		}
	}
	
	private void addJoinSet(ProxyReg reg) {
		if(reg.proxy == null || reg.srvcRegistration == null)
			throw new IllegalArgumentException(
			"bad proxy or ServiceRegistration");
		synchronized(joinSet) {
			if(!joinSet.contains(reg)) {
				joinSet.add(reg);
			}
		}
		ArrayList notifies;
		synchronized(listeners) {
			if(listeners.isEmpty())
				return;
			notifies = (ArrayList)listeners.clone();
		}
		discoveredNotifyListener(reg.proxy, notifies);
	}
	
	private void removeJoinSet(ProxyReg reg) {
		if(reg.proxy == null || reg.srvcRegistration == null)
			throw new IllegalArgumentException(
			"bad proxy or ServiceRegistration");
		synchronized(joinSet) {
			joinSet.remove(reg);
			try {
				leaseRenewalMgr.cancel(reg.srvcRegistration.getLease());
			} catch (Exception e) {
			}
		}
		ArrayList notifies;
		synchronized(listeners) {
			if(listeners.isEmpty())
				return;
			notifies = (ArrayList)listeners.clone();
		}
		discardedNotifyListener(reg.proxy, notifies);
	}
	
	private void discoveredNotifyListener(ServiceRegistrar proxy,
		ArrayList notifies) 
	{
		notifyListener(proxy, notifies, true);
	}
	
	private void discardedNotifyListener(ServiceRegistrar proxy,
		ArrayList notifies) 
	{
		notifyListener(proxy, notifies, false);
	}
	private void notifyListener(ServiceRegistrar proxy,
		ArrayList notifies,
		boolean bDiscovered) 
	{
		ServiceRegistrar[] proxys = new ServiceRegistrar[]{proxy};
		DiscoveryEvent evt = new DiscoveryEvent(this, proxys);
		Iterator iter = notifies.iterator();
		while(iter.hasNext()) {
			if(bDiscovered)
				((DiscoveryListener)iter.next()).discovered(evt);
			else
				((DiscoveryListener)iter.next()).discarded(evt);
		}
	}
	
	ProxyReg findReg(ServiceRegistrar proxy) {
		for (Iterator iter = joinSet.iterator(); iter.hasNext(); ) {
			ProxyReg reg =(ProxyReg)iter.next();
			if(reg.proxy.equals(proxy))
				return reg;
		}
		return null;
	}
}
