/*
 * Copyright (c) 1998-2000 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.registry;

import java.io.*;
import java.net.*;
import java.io.*;
import java.util.*;
import java.rmi.*;
import java.rmi.server.*;

import jgroup.*;
import jgroup.reference.*;


/**
 * Class <code>RegistryFactory</code> is used to obtain a proxy
 * to the dependable registry service, or to create a dependable
 * registry replica on the local machine. <p>
 *
 * // Questo dipende da come lo implemento io; nel futuro,
 * potrebbe essere diverso (basato su Jini).
 * Every instance of the dependable registry service is associated
 * with a standard RMI registry, which is used to retrieve a
 * unicast proxy for the replica.
 *
 * to obtain used to obtain a bootstrap reference to
 * the Jgroup dependable registry, or to create a Jgroup dependable registry instance
 * on the local machine. Every instance of the Jgroup registry service is based on
 * two remote registries: the standard local registry and the replicated one. The
 * standard local registry is used to retrieve a reference for the replicated
 * remote registry. When the argument <code> port </code> is used in a getRegistry
 * or a createRegistry method, it specifies the port used by the local registry.
 * The port used by the replicated registry is <code> port+1 </code>.
 *
 * @author      Alberto Montresor
 * @since       Jgroup 0.8
 */
public final class  RegistryFactory
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static section
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** */
  public static final String RESOURCE = "META-INF/Jgroup/RegistryLocator";

  /** Group manager locator needed to obtain group managers */
  private static RegistryLocator  loc;

  /** Construct locator */
  private static void setLocator()
    throws RemoteException
  {
    String classname = null;
    Class clazz = null;

    try {

      Enumeration e = ClassLoader.getSystemResources(RESOURCE);
      URL url = (URL) e.nextElement();
      if (e.hasMoreElements())
        throw new JgroupException("Unable to obtain a registry reference: more than one Jgroup implementation found");
      BufferedReader reader = new BufferedReader(new InputStreamReader(url.openStream()));
      classname = reader.readLine();
      if (classname == null)
        throw new JgroupException("Unable to obtain a registry reference: resource " + RESOURCE + " is empty");
      clazz = Class.forName(classname);
      loc = (RegistryLocator) clazz.newInstance();

    } catch (NoSuchElementException e) {
      throw new RemoteException("Unable to obtain a registry reference: resource " + RESOURCE + "not found");
    } catch (IOException e) {
      throw new RemoteException("Unable to obtain a registry reference: I/O error when reading resource " + RESOURCE);
    } catch (ClassNotFoundException e) {
      throw new RemoteException("Unable to obtain a registry reference: class " + classname + " not found");
    } catch (ClassCastException e) {
      throw new RemoteException("Unable to obtain a registry reference: class " + classname + " does not extend jgroup.reference.RegistryLocator");
    } catch (Exception e) {
      throw new RemoteException("Unable to obtain a registry reference");
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Private constructor to disable public construction.
   */
  private RegistryFactory() {}


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Public methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Returns a proxy for the dependable registry service on the specified
   * distributed system.
   * @param ds          the distributed system in which the dependable registry
   *                    replicas are expected to run
   * @return            a proxy for the dependable registry service
   * @exception RemoteException if the proxy could not be obtained
   */
  public static DependableRegistry getRegistry(DistributedSystem ds)
    throws RemoteException
  {
    if (loc == null)
      setLocator();
    return loc.getRegistry(ds);
  }

  /**
   * Returns a proxy for the dependable registry service on the specified
   * distributed system and port.
   *
   * @param ds          the distributed system in which the dependable registry
   *                    replicas are expected to run
   * @param port        port on which distributed replicas are listening
   * @return            a proxy for the dependable registry service
   * @exception RemoteException if the proxy could not be obtained
   */
  public static DependableRegistry getRegistry(DistributedSystem ds, int port)
        throws RemoteException
  {
    if (loc == null)
      setLocator();
    return loc.getRegistry(ds, port);
  }


  /**
   * Returns a proxy for the dependable registry service on the specified
   * distributed system and port. <p>
   *
   * Communication with this remote registry will use the supplied
   * <t>RMIClientSocketFactory csf</t> to create socket connections
   * to the dependable registry on the distributed system and port.
   * @param ds          the distributed system in which the dependable registry
   *                    replicas are expected to run
   * @param port        port on which distributed replicas are listening
   * @param csf         client-side Socket factory used to make connections to
   *                    the dependable registry. If <t>csf</t> is null, then the
   *                    default client-side socket factory will be used
   * @return            a proxy for the dependable registry service
   * @exception RemoteException if the proxy could not be obtained
   */
  public static DependableRegistry getRegistry(DistributedSystem ds, int port,
                                     RMIClientSocketFactory csf)
        throws RemoteException
  {
    if (loc == null)
      setLocator();
    return loc.getRegistry(ds, port, csf);
  }


  /**
   * Creates and exports a dependable registry replica on the local host that
   * listen to incoming requests on an anonymous communication port.
   * Objects contained <tt>conf</tt> are used by Jgroup to configure the
   * group communication; the distributed system description must be
   * inclued in this array.
   *
   * @param conf        configuration information
   * @return            the new dependable registry replica.
   *
   * @exception RemoteException if the registry could not be exported
   * @exception JgroupException
   *            Generic exception raised by Jgroup.
   * @exception DSException
   *            raised when the distributed system configuration object contains an error
   */
  public static DependableRegistry createRegistry(Object[] conf)
        throws RemoteException, JgroupException
  {
    if (loc == null)
      setLocator();
    return loc.createRegistry(conf);
  }


  /**
   * Creates and exports a dependable registry replica on the local host that
   * listen to incoming requests on  the given <code>port</code> communication
   * port.
   * Objects contained <tt>conf</tt> are used by Jgroup to configure the
   * group communication; the distributed system description must be
   * inclued in this array.
   *
   * @param conf        configuration information
   * @param port        port used for RMI communications
   * @return            the new dependable registry replica.
   * @exception RemoteException if the registry could not be exported
   * @exception JgroupException
   *            Generic exception raised by Jgroup.
   * @exception DSException
   *            raised when the distributed system configuration object contains an error
   */
  public static DependableRegistry createRegistry(Object[] conf, int port)
        throws RemoteException, JgroupException
  {
    if (loc == null)
      setLocator();
    return loc.createRegistry(conf, port);
  }


  /**
   * Creates and exports a dependable registry replica on the local host that
   * listen to incoming requests on  the given <code>port</code> communication
   * port.
   * Objects contained <tt>conf</tt> are used by Jgroup to configure the
   * group communication; the distributed system description must be
   * inclued in this array.
   * The new instance listens to incoming requests using a
   * <code>ServerSocket</code> created from the supplied
   * <code>RMIServerSocketFactory</code>.  A client
   * that receives a reference to this registry will use a
   * <code>Socket</code> created from the supplied
   * <code>RMIClientSocketFactory</code>.
   *
   * @param conf        configuration information
   * @param port        port used for RMI communications
   * @param csf         client-side <code>Socket</code> factory used to
   *                    make connections to the registry
   * @param ssf         server-side <code>ServerSocket</code> factory
   *                    used to accept connections to the registry
   * @return            the new dependable registry replica.
   * @exception RemoteException if the registry could not be exported
   * @exception JgroupException
   *            Generic exception raised by Jgroup.
   * @exception DSException
   *            raised when the distributed system configuration object contains an error
   */
  public static DependableRegistry createRegistry(Object[] conf, int port,
                                                  RMIClientSocketFactory csf,
                                                  RMIServerSocketFactory ssf)
        throws RemoteException, JgroupException
  {
    if (loc == null)
      setLocator();
    return loc.createRegistry(conf, port, csf, ssf);
  }

} // END RegistryFactory



