/*
 * Copyright (c) 1998-2000 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package relacs;

import jgroup.TransportConfig;

/**
 * The <code> Config</code> class contains the configuration
 * parameters needed by the underlying transport protocol used in
 * Relacs.
 *
 * @author  Alberto Montresor
 * @since   Jgroup 0.1
 */
public class    RelacsConfig
extends         TransportConfig
{


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  private static final int PAYLOAD      = 8192;         // Default payload
  private static final int LAMBDA       = 4;            // Default lambda
  private static final int ALFAN        = 7;            // Default alfan
  private static final int ALFAD        = 8;            // Default alfad
  private static final int RTIMEOUT     = 400;          // Default rtimeout


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  private int           payload;        // Payload length of a mss message
  private int           lambda;         // Failure detector parameter
  private int           alfan;          // Failure detector parameter
  private int           alfad;          // Failure detector parameter
  private int           rtimeout;       // Failure detector parameter


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Constructs a <code> Config</code> using default values and the Jgroup port.
   *
   */
  public RelacsConfig()
  {
    super();
    this.payload  = PAYLOAD;
    this.lambda   = LAMBDA;
    this.alfan    = ALFAN;
    this.alfad    = ALFAD;
    this.rtimeout = RTIMEOUT;
  }

  /**
   * Constructs a <code> Config</code> using default values and the Jgroup port.
   *
   */
  public RelacsConfig(int port)
  {
    super(port);
    this.payload  = PAYLOAD;
    this.lambda   = LAMBDA;
    this.alfan    = ALFAN;
    this.alfad    = ALFAD;
    this.rtimeout = RTIMEOUT;
  }

  /**
   * Constructs a <code> Config</code> using default values and the Jgroup port.
   *
   */
  public RelacsConfig(TransportConfig conf)
  {
    super(conf.getPort());
    this.payload  = PAYLOAD;
    this.lambda   = LAMBDA;
    this.alfan    = ALFAN;
    this.alfad    = ALFAD;
    this.rtimeout = RTIMEOUT;
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Set the size of the payload field of application packets.
   *
   * @param payload     new payload value
   */
  public void setPayload(int payload)
  {
    this.payload = payload;
  }

  /**
   * Get the size of the payload field of application packets.
   */
  public int getPayload()
  {
    return payload;
  }


  /**
   * Set the lambda parameter, which determines the number of routing
   * messages that could be lost before a remote host is suspected.
   *
   * @param lambda      new lambda value
   */
  public void setLambda(int lambda)
  {
    this.lambda = lambda;
  }

  /**
   * Get the lambda parameter.
   */
  public int getLambda()
  {
    return lambda;
  }

  /**
   * Set the <code> alfa</code> parameter, whose value is given by <code>
   * alfan / alfad</code>. The <code> alfa</code> parameter is used in
   * the computation of the average round-trip time (RTT), and it determines
   * the weight given to the old values. The formula used is the following:
   * <p>
   * <code>
   * RTT = alfa * RTT + (1 - alfa) * M
   *</code>
   * <p>
   * where <code> M</code> is the last measured round-trip time.
   *
   * @param alfan       new alfan value
   * @param alfad       new alfad value
   */
  public void setAlfa(int alfan, int alfad)
  {
    this.alfan = alfan;
    this.alfad = alfad;
  }

  /**
   * Get the <code> alfan</code> parameter.
   */
  public int getAlfan()
  {
    return alfan;
  }

  /**
   * Get the <code> alfad</code> parameter.
   */
  public int getAlfad()
  {
    return alfad;
  }

  /**
   * Set the time interval between two routing messages sent
   * by a daemon.
   *
   * @param rtimeout    new routing timeout value
   */
  public void setRtimeout(int rtimeout)
  {
    this.rtimeout = rtimeout;
  }

  /**
   * Get the time interval between two routing messages sent
   * by a daemon.
   */
  public int getRtimeout()
  {
    return rtimeout;
  }

} // END RelacsConfig
