/*
 * Copyright (c) 1998-2000 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package relacs;

import java.net.*;
import java.io.*;
import java.util.*;

import jgroup.*;


/**
 * A <code> DistributedSystem</code> object contains the description of the
 * distributed system in which the distributed computation is carried on.
 * Currently, a distributed system is described as a collection of hosts
 * and may be constructed in two ways: by adding a collection of host names
 * to a initially empty distributed system, or by reading a file containing
 * a list of host names.
 *
 * @author  Alberto Montresor
 * @since   Jgroup 0.2
 */
public final class      RelacsDS
extends                 DistributedSystem
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  private static InetAddress    local = null;   // Local host

  static {
    byte[]      address;

    try {
      local = InetAddress.getLocalHost();
      address = local.getAddress();
      if (address[0] == 127 && address[1] == 0 && address[2] == 0 && address[3] == 1)
        throw new UnknownHostException();
    }
    catch (UnknownHostException e1) {
      local = null;
      String name = System.getProperty("jgroup.hostname");
      if (name != null)
        try {
          local = InetAddress.getByName(System.getProperty("jgroup.hostname"));
        } catch (Exception e2) { }
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  private ArrayList   names;          // Host list


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Constructs a distributed system with no hosts.
   */
  public RelacsDS()
  {
    super("");
    names = new ArrayList();
  }

  /**
   * Constructs a distributed system whose description is
   * contained in <code> hostfile</code>. The file must contain
   * a list of host names, each line containing one host name.
   *
   * @param hostfile    The pathname of the file containing the
   *                    description of the distributed system.
   */
  public RelacsDS(String hostfile)
    throws JgroupException
  {
    super(hostfile);
    try {
      names = new ArrayList();
      BufferedReader in = new BufferedReader(new FileReader(hostfile));
      String hostname;
      while ((hostname = in.readLine()) != null)
        add(hostname);
    } catch (IOException e) {
      throw new JgroupException(
        "Caught IOException when reading the distributed system configuration: " +e);
    }
  }

  /**
   * Constructs a distributed system whose description is
   * contained in a generic <code>DistributedSystem</code>.
   *
   * @param hostfile    The pathname of the file containing the
   *                    description of the distributed system.
   */
  public RelacsDS(DistributedSystem ds)
    throws JgroupException
  {
    this(ds.getConfigfile());
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Public methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns the number of hosts contained in the distributed system.
   */
  public int size()
  {
    return names.size();
  }

  /**
   *  Returns an array of host names representing the hosts contained in
   *  the distributed system.
   */
  public String[] getNames()
  {
    return (String[]) names.toArray(new String[names.size()]);
  }

  /**
   *  Add a host name to the distributed system.
   *
   * @param hostname    the name of the host to add
   */
  public void add(String hostname)
  {
    names.add(hostname);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns the internet address of the local host.
   */
  public static InetAddress getLocalHost()
    throws UnknownHostException
  {
    if (local == null)
      throw new UnknownHostException();
    return local;
  }

} // END  DistributedSystem













