/*
 * Copyright (c) 1998-2000 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package relacs.daemon;

import java.io.*;
import java.net.*;
import java.util.*;

import jgroup.*;
import jgroup.multicast.*;
import relacs.*;
import relacs.gm.*;
import relacs.mss.*;
import relacs.general.*;

import debug.Debug;

/**
 * The <code> Daemon</code> class
 *
 * @author  Alberto Montresor
 * @since   Jgroup 0.1
 */
public final class      Daemon
implements              MssUser, Tag
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static section
  ////////////////////////////////////////////////////////////////////////////////////////////

  private static Hashtable      daemons;        // Daemon table
  private static int            members = 0;    // Number of members created so far


  static
  {
    daemons = new Hashtable(101);
  }


  /**
   * Obtain a daemon object identified by <code> index</code>.
   *
   * @param     port    index of the daemon object
   * @param     ds      distributed system configuration
   * @param     conf    mss configuration parameters
   */
  public static Daemon getDaemon(RelacsDS ds, RelacsConfig conf)
       throws JgroupException
  {
    Integer     key;
    Daemon      daemon;
    int         port = conf.getPort();

    if (port < 0 || port > 65535)
      throw new JgroupException("Invalid daemon index: "+port);
    key = new Integer(port);
    daemon = (Daemon) daemons.get(key);
    if (daemon == null) {
      daemon = new Daemon(ds, conf);
      daemons.put(key, daemon);
    }
    return daemon;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  private GroupTable    grouptable;     // Group table
  private HostTable     hosttable;      // Host table
  private Mss           mss;            // Mss
  private MsgFactory    mfactory;       // Msg factory used to obtain messages
  private Host          me;             // My host info

  private int           nsuspects;      // Number of suspect events
  private MssHost[]     trset;          // Reachable set of last suspect event
  private MssHost[]     nrset;          // New hosts in reachable set of last suspect event
  private MssHost[]     nuset;          // New hosts in unreachable set of last suspect event


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  private Daemon(RelacsDS ds, RelacsConfig conf)
        throws JgroupException
  {
    Host[]      hosts;
    String[]    names;

    // Initialize group table
    grouptable = new GroupTable();

    // Initialize host table
    hosts = new Host[ds.size()];
    names = ds.getNames();
    try {
      for (int i=0; i < hosts.length; i++)
        hosts[i] = new Host(names[i]);
    } catch (IOException e) {
      throw new JgroupException(e.getMessage());
    }
    hosttable = new HostTable(hosts);
    me = (Host) hosttable.getMe();
    if (me == null)
      throw new DSException("Missing local host");

    // Initialize suspect information
    nsuspects = 0;
    trset = new Host[1];
    trset[0] = me;
    nrset = new Host[0];
    nuset = new Host[0];

    // Initialize mss
    try {
      mss = new Mss("RelacsDaemon", this, hosttable, conf, conf.getPort());
    } catch (IOException e) {
      throw new JgroupException(e.getMessage());
    }
    mfactory = mss.getMsgFactory();


  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Communication with members
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *
   */
  public MemberId getMemberId()
  {
    return new MemberIdImpl(me.getInetAddress(), me.getIncarnation(), ++members);
  }

  /**
   *
   */
  public OutputStream getMessage()
  {
    return MsgMcast.createOutputStream(mfactory);
  }


  /**
   *
   * @param sid Identifier of the sender of the message for which this acknowledgment is
   *  provided.
   */
  public void notifyDlvrAck(int gid, int hpos, int mpos, int vpos, int mid, long sid,
    Object result)
  {
    mss.lsend(new UpcallDlvrAck(gid, hpos, mpos, vpos, mid, sid, result));
  }

  /**
   *
   */
  public void notifyPrepareAck(int gid, int mpos)
  {
    mss.lsend(new UpcallAck(L_PREPAREACK, gid, mpos, 0));
  }

  /**
   *
   */
  public void notifyViewAck(int gid, int mpos, int last)
  {
    mss.lsend(new UpcallAck(L_VIEWACK, gid, mpos, last));
  }

  /**
   *
   */
  public void join(int gid, int lid, BaseLayer gm)
  {
    mss.lsend(new UpcallGroup(L_JOIN, gid, lid, gm));
  }

  /**
   *
   */
  public void leave(int gid, int lid)
  {
    mss.lsend(new UpcallGroup(L_LEAVE, gid, lid, null));
  }

  /**
   * @param sid Identifier of the sender of the message, obtained through
   *   getLongIdentifier() of MemberId. 
   */
  public void mcast(int gid, boolean flag, OutMessage stream, long sid, AckListener ackl)
  {
    mss.lsend(new UpcallMcast(gid, flag, stream, sid, ackl));
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Implementations of methods inherited by interface MssUser
  ////////////////////////////////////////////////////////////////////////////////////////////


  /**
   *  Handles a message m-received by a remote host.
   */
  public void mreceive(byte tag, InMessage stream, MssHost src)
  {
    int                 gid;            // Group identifier number
    Group               group;          // Reference to the destination group

    gid = Types.readGid(stream);
    if ((group = grouptable.lookup(gid)) != null) {
      switch(tag) {

      case R_MCAST:
        if (Debug.DEBUG)
          stream.print(0);
        group.handleMsgMcast  (MsgMcast.decode(stream),   (Host) src);
        break;
      case R_ACK:
        group.handleMsgAck    (MsgAck.decode(stream),     (Host) src);
        break;
      case R_PROP:
        group.handleMsgProp   (MsgProp.decode(stream),    (Host) src);
        break;
      case R_ESTIM:
        group.handleMsgEstim  (MsgEstim.decode(stream),   (Host) src);
        break;
      case R_FORWARD:
        group.handleMsgDlvrd  (MsgMcast.decode(stream),   (Host) src);
        break;
      case R_SYNCH:
        group.handleMsgSynch  (MsgSynch.decode(stream),   (Host) src);
        break;
      case R_SYMM:
        group.handleMsgSymm   (MsgSymm.decode(stream),    (Host) src);
        break;
      case R_VIEW:
        group.handleMsgView   (MsgView.decode(stream),    (Host) src);
        break;
      case R_RESULT:
        group.handleMsgResult (MsgResult.decode(stream),  (Host) src);
        break;
      case R_LEAVE:
        group.leave((Host) src);
        break;
      case R_JOIN:
      case R_JOINACK:
        handleMsgJoin(tag, group, MsgGroup.decode(stream), (Host) src);
        break;
      default:
        // Gestione errore
        break;

      } // END switch
      if (group.tmembers.isEmpty())
        grouptable.remove(group.key);
    }
    else  {
      // None of the members joined the group
      // mss.msend(R_LEAVE, MsgGroup.encode(mfactory, gid, null), src);
    }
  }

  /**
   *  Handles a message received by a local member.
   */
  public void lreceive(Upcall upcall)
  {
    Group               group;          // Reference to the destination group
    MemberData          md;             // Reference to new member additional info
    HostData            hd;             // Reference to new host additional info
    long                vid;            // View identifier
    ViewImpl            view;           // New view
    InetAddress[]       hosts;          //
    int[]               vindex;         //

    if ((group = grouptable.lookup(upcall.getId())) != null) {
      switch (upcall.getTag()) {
        case L_MCAST:
          group.handleLocalMcast        ((UpcallMcast) upcall);
          break;
        case L_DLVRACK:
          group.handleLocalDlvrAck      ((UpcallDlvrAck) upcall);
          break;
        case L_VIEWACK:
          group.handleLocalViewAck      ((UpcallAck) upcall);
          break;
        case L_PREPAREACK:
          group.handleLocalPrepareAck   ((UpcallAck) upcall);
          break;
        case L_JOIN:
          group.handleLocalJoin         ((UpcallGroup) upcall);
          break;
        case L_LEAVE:
          group.handleLocalLeave        ((UpcallGroup) upcall);
          break;
     }
    }
    else if (upcall.getTag() == L_JOIN) {
      UpcallGroup ug = (UpcallGroup) upcall;
      vid = Types.createsVid(me.getAddress(), 0);
      MemberIdImpl id = new MemberIdImpl(me.getInetAddress(), me.getIncarnation(), ug.lid);
      view  = new ViewImpl(ug.getId(), vid, id);
      md    = new MemberData(ug.lid, ug.layer);
      hd    = new HostData(me);
      group = new Group(mss, ug.getId(), S_IDLE, view, hd, md);
      grouptable.insert(group);

      hosts = hosttable.getHostAddressArray();
      vindex = new int[hosts.length];
      for (int j=0; j < vindex.length; j++)
        vindex[j] = -1;

      md.layer.notifyView(ug.getId(), view, group.ninstalled, group.hpos, 0);
      mss.msend(R_JOIN, MsgGroup.encode(mfactory, group.key, group.thosts));
    }
  }

  /**
   * Handles variations in the mss-reachability
   *
   * @param     trset   set of all reachables hosts
   * @param     nrset   set of reachable hosts that were unreachable
   * @param     nuset   set of unreachable hosts that were reachable
   */
  public void suspect(MssHost[] trset, MssHost[] nrset, MssHost[] nuset, MssHost newinc)
  {
    Group               scan;           // Used to scan lists of groups reference
    HostData            hl;             // Used to scan host in a group

    if (Debug.DEBUG) {
      Debug.print(20, "Daemon: Suspect Rset:  ");
      for (int i=0; i < trset.length; i++)
        Debug.print(20, trset[i].getName() + " ");
      Debug.println(20, "");
    }
    for(int i=0; i < nrset.length; i++) {
      if (Debug.DEBUG) {
        Debug.print(20, "setSymset: ");
        for (int j=0; j<this.trset.length; j++)
          Debug.print(20, (this.trset[j].getAddress() & 0xFF) + " ");
        Debug.println(20, "");
      }
      ((Host) nrset[i]).symset = this.trset;
    }
    this.trset = trset;
    this.nrset = nrset;
    this.nuset = nuset;

    if (Debug.DEBUG) {
      Debug.print(20, "Daemon: this.trset:  ");
      for (int i=0; i < this.trset.length; i++)
        Debug.print(20, this.trset[i].getName() + " ");
      Debug.println(20, "");
    }

    Group[] groups = grouptable.getGroups();
    for (int i=0; i < groups.length; i++)
      mss.msend(R_JOIN, MsgGroup.encode(mfactory, groups[i].key, groups[i].thosts), nrset);
    nsuspects++;

    // Updates version numbers in case of new incarnations
    if (newinc != null) {
      if (Debug.DEBUG)
        Debug.println(20, "Newinc: " + (newinc.getAddress() & 0xFF));
      scan = (Group) ((Host) newinc).grouplist.getFirst();
      while (scan != null) {
        if (Debug.DEBUG)
          Debug.println(20, "Group: " + scan.key);
        HostData hd = (HostData) scan.thosts.lookup(newinc.getAddress());
        if (hd != null)
          hd.newIncarnation();
        scan = (Group) scan.getNext();
      }
    }
    activateGroups(nrset, trset, nrset, nuset);
    activateGroups(nuset, trset, nrset, nuset);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // General functions
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *
   */
  private void activateGroups(MssHost[] hosts, MssHost[] trset, MssHost[] nrset,
                              MssHost[] nuset)
  {
    Group           scan;           // Used to scan lists of groups reference

    for (int i=0; i < hosts.length; i++) {
      scan = (Group) ((Host) hosts[i]).grouplist.getFirst();
      while (scan != null) {
        scan.handleSuspect(nsuspects, trset, nrset, nuset);
        scan = (Group) scan.getNext();
      }
    }
  }


  /**
   *
   * @param     tag     the tag associated with the message; could be R_JOIN or R_JOINACK
   * @param     group   the Group data structure
   * @param     msg     the m-received message.
   * @param     src     the sender of the message.
   */
  private void handleMsgJoin (byte tag, Group group, MsgGroup msg, Host src)
  {
    if (Debug.DEBUG)
      msg.print();
    if (tag != R_JOINACK)
      mss.msend(R_JOINACK, MsgGroup.encode(mfactory, group.key, group.thosts), src);
    Host[] hosts = new Host[msg.hosts.length];
    for (int i=0; i < hosts.length; i++)
      hosts[i] = (Host) hosttable.lookup(msg.hosts[i]);
    group.join(src, hosts, nsuspects, trset, nrset, nuset);
  }

} // END Daemon

