/*
 * Copyright (c) 1998-2000 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package relacs.daemon;

import relacs.mss.MssHost;
import relacs.mss.Mss;
import relacs.general.*;
import debug.Debug;

/**
 * The <code> HostData</code> class
 *
 * @author  Alberto Montresor
 * @since   Jgroup 0.3
 */
final class     Estimate
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  Host[]                hlist;          // Array of hosts contained in the estimate
  HostData[]            hdata;          // Additional data associated to hosts in <hlist>
  int                   hsize;          // N. of hosts in hlist/hdata (hsize <= hlist.length)

  private boolean       changed;        // True if estimate has changed
  private int           nsynchronized;  // Number of synchronized elements in estimate
  private SendBuffer    sbuffer;        // Send buffer used to notify acknowledgment

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  Estimate(SendBuffer sbuffer)
  {
    this.sbuffer = sbuffer;
    hdata = new HostData[0];
    hlist = new Host[0];
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /*
   * Creates a new estimate by intersecting the total list of hosts contained in
   * <thosts> and the list of reachable host contained in <rset>.
   *
   * @param thosts      the total list of hosts
   * @param rset        the reachable set
   */
  void init(HostData thosts, MssHost[] rset)
  {
    HostData            scan;           // Used to scan total
    int                 i=0;            // Used to scan rset

    if (hdata.length < rset.length) {
      hdata = new HostData[rset.length];
      hlist = new Host[rset.length];
    }

    if (Debug.DEBUG)
      Debug.print(20, "Estimate: ");
    scan = (HostData) thosts.getFirst();
    hsize = 0;
    while (scan != null) {
      if (i == rset.length || scan.key < rset[i].getAddress() || scan.leaving) {
        scan.inEstimate = false;
        if (scan.viewIndex > 0)   // ??? Controllare questo
          sbuffer.suspect(scan.viewIndex);

        scan = (HostData) scan.getNext();
      }
      else if (scan.key == rset[i].getAddress() && !scan.leaving) {
        if (Debug.DEBUG)
          Debug.print(20, " " + scan.host.getName());
        hdata[hsize] = scan;
        hlist[hsize] = scan.host;
        scan.inEstimate = true;
        scan.inSynchronized = false;
        scan = (HostData) scan.getNext();
        hsize++;
        i++;
      }
      else {
        i++;
      }
    }
    if (Debug.DEBUG)
      Debug.println(20, " " +hsize);

    changed = true;
    nsynchronized = 1;
  }


  /*
   * Computes the intersection between the current estimate and the set <hosts>
   *
   * @param hosts       the hosts to be intersected with the current estimate
   * @param agreed      the agreed version numbers received with the hosts
   */
  void intersect(int[] hosts, int[] agreed)
  {
    int         nsize=0;                // New hsize of the estimate
    int         i=0, j=0;               // Counters

    while (i < hsize && j < hosts.length) {
      if (hdata[i].key == hosts[j]) {
        // Do not remove hdata[i]/hlist[i]
        hdata[nsize] = hdata[i];
        hlist[nsize] = hlist[i];
        if (agreed != null)
          hdata[nsize].agreed = agreed[j];
        nsize++;
        i++; j++;
      } else if (hdata[i].key < hosts[j]) {
        // Remove hdata[i]/hlist[i]
        if (hdata[i].viewIndex > 0)
          sbuffer.suspect(hdata[i].viewIndex);
        hdata[i].inEstimate = false;
        if (hdata[i].inSynchronized)
          // If synchronized, it has to be removed from the count
          nsynchronized--;
        i++;
      } else {
        j++;
      }
    }

    changed = (nsize < hsize);
    hsize     = nsize;
  }


  /*
   * Remove the host described by <host> from the current estimate.
   *
   * @param host        the host to be removed
   */
  void remove(MssHost host)
  {
    int         i=0, j=0;               // Counters

    changed = false;
    while (i < hsize) {
      if (hlist[i] == host) {
        // Remove hdata[i]/hlist[i]
        if (hdata[i].viewIndex > 0)
          sbuffer.suspect(hdata[i].viewIndex);
        changed = true;
        hsize--;
        hdata[i].inEstimate = false;
        if (hdata[i].inSynchronized)
          // If synchronized, it has to be removed from the count
          nsynchronized--;
      } else {
        // Do not remove hdata[i]/hlist[i]
        hdata[j] = hdata[i];
        hlist[j] = hlist[i];
        j++;
      }
      i++;
    }
  }


  /*
   * Remove from the estimate the hosts contained in the <uset> array.
   */
  void remove(MssHost[] uset)
  {
    int         nsize=0;                // New hsize of the estimate
    int         i=0, j=0;               // Counters

    while (i < hsize) {
      if (j == uset.length || hdata[i].key < uset[j].getAddress()) {
        // Do not remove hdata[i]/hlist[i]
        hdata[nsize] = hdata[i];
        hlist[nsize] = hlist[i];
        nsize++;
        i++;
      } else if (hdata[i].key == uset[j].getAddress()) {
        // Remove hdata[i]/hlist[i]
        if (hdata[i].viewIndex > 0)
          sbuffer.suspect(hdata[i].viewIndex);
        hdata[i].inEstimate = false;
        if (hdata[i].inSynchronized)
          // If synchronized, it has to be removed from the count
          nsynchronized--;
        i++; j++;
      } else
        j++;
    }

    changed = (nsize < hsize);
    hsize     = nsize;
  }


  /*
   * Remove from the estimate the hosts whose addresses are contained in the <uset>
   * array.
   */
  void remove(int[] uset)
  {
    int         nsize=0;                // New hsize of the estimate
    int         i=0, j=0;               // Counters

    nsize = 0;
    while (i < hsize) {
      if (j == uset.length || hdata[i].key < uset[j]) {              // heVcomp[i].key is not removed
        // Do not remove hdata[i]/hlist[i]
        hdata[nsize] = hdata[i];
        hlist[nsize] = hlist[i];
        nsize++;
        i++;
      } else if (hdata[i].key == uset[j]) {                    // heVcomp[i].key must be removed
        // Remove hdata[i]/hlist[i]
        if (hdata[i].viewIndex > 0)
          sbuffer.suspect(hdata[i].viewIndex);
        hdata[i].inEstimate = false;
        if (hdata[i].inSynchronized)
          // If synchronized, it has to be removed from the count
          nsynchronized--;
        i++; j++;
      } else {
        j++;
      }
    }

    changed = (nsize < hsize);
    hsize    = nsize;
  }


  /*
   * Return true if the host identified by <address> is contained in the estimate.
   */
  HostData lookup(int address)
  {
    int                 i;              // Counter

    for (i=0; i < hsize && hdata[i].key < address; i++);
    return ((i<hsize && hdata[i].key == address) ? hdata[i] : null);
  }

  /**
   * If hosts contained in <hi> belongs to the estimate and is not synchronized,
   * change its <inSynchronized> field and update the counter of synchronized
   * hosts.
   */
  void setSynchronized(HostData hd)
  {
    if (hd.inEstimate && !hd.inSynchronized) {
      hd.inSynchronized = true;
      nsynchronized++;
    }
  }

  boolean isSynchronized()
  {
    return (hsize <= nsynchronized);
  }

  boolean hasChanged()
  {
    return changed;
  }

  Host getCoordinator()
  {
    return hlist[0];
  }

} // END Estimate
