/*
 * Copyright (c) 1998-2000 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package relacs.daemon;

import java.net.*;
import java.util.*;

import jgroup.*;
import jgroup.multicast.*;
import relacs.gm.*;
import relacs.mss.*;
import relacs.general.*;
import debug.Debug;


/**
 * The <code> Group</code> class
 *
 * @author  Alberto Montresor
 * @since   Jgroup 0.1
 */
final class     Group
extends         OrderedList
implements      Tag
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  // General fields
  Mss                   mss;            // Mss used to send
  MsgFactory            mfactory;       // Message factory
  int                   status;         // Current mode
  MssHost[]             rset;           // Current reachable set
  HostData              me;             // Information related to this host
  int                   myaddress;      // My 32 bit address
  int                   ninstalled;     // Number of views installed so far
  int                   ncoordinated;   // Number of views coordinated so far
  int                   nsuspects;      // Number of suspect events
  HostData              thosts;         // Complete list of hosts interested in the group
  MemberData            tmembers;       // Complete list of members interested in the group

  // Current view information
  long                  cvid;           // Complete view identifier
  int[]                 cview;          // Complete view composition (host addresses)
  long                  pvid;           // Partial view identifier
  PartialView           pview;          // Partial view composition
  int                   hpos;           // This host position in <pview>.<hosts>

  // Agreement phase data
  Queue                 delayedSent;    // Messages whose sending is postponed
  Queue                 delayedRcvd;    // Messages whose delivering is postponed
  Estimate              estimate;       // Hosts in the current estimate
  int                   nmembers;       // N. of members in the current estimate
  int                   nproposals;     // Number of proposals sent during this view
  MsgProp               proposal;       // Current proposal
  boolean               msgsStable;     // True if message buffer is consistent
  boolean               viewStable;     // True if view delivered is consistent

  SendBuffer            sbuffer;        // Send buffer

  // Group table links
  Group                 tnext;          // Next element in the general group table
  Group                 tprev;          // Previous element in the general group table
  Group                 hnext;          // Next element in the hash group table
  Group                 hprev;          // Previous element in the hash group table


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Creates an empty group used in the GroupTable structure
   */
  Group()
  {
  }


  /**
   * Creates a group identified by the group identifier <code> key</code>.
   *
   * @param     key             group identifier
   * @param     status          initial status of group
   * @param     vid             inital view identifier (both complete and partial)
   * @param     hd              this host additional info
   * @param     md              joining member additional info
   */
  Group(Mss mss, int gid, int status, ViewImpl view, HostData hd, MemberData md)
  {
    // Group identifier initialization
    this.key     = gid;

    // General fields
    this.mss      = mss;
    this.mfactory = mss.getMsgFactory();
    this.status   = status;
    rset          = new Host[1];
    rset[0]       = hd.host;
    me            = hd;
    myaddress     = hd.host.getAddress();
    ncoordinated  = 0;
    ninstalled    = 0;
    nsuspects     = 0;
    thosts        = new HostData();
    tmembers      = new MemberData();
    thosts.insert(hd);
    tmembers.insert(md);

    sbuffer = new SendBuffer();

    // Initialization of the first view
    cvid         = view.getVid();
    pvid         = cvid;
    cview        = new int[1];
    cview[0]     = hd.key;
    pview        = new PartialView(hd, md);
    hpos         = 0;
    sbuffer.setView(view, 1);

    // Initialization of agreement phase info
    delayedSent = new Queue();
    delayedRcvd = new Queue();
    estimate    = new Estimate(sbuffer);
    nmembers    = 1;
    proposal    = new MsgProp();
    msgsStable  = true;
    viewStable  = false;

  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Group management
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *
   */
  void join(Host src, Host[] hosts, int nsuspects, MssHost[] trset,
            MssHost[] nrset, MssHost[] nuset)
  {
    HostData            sender;         // Reference to additional sender info
    HostData            scan;           // Reference to

    if (Debug.DEBUG)
      Debug.println(20, "I'm interested");
    sender = (HostData) thosts.lookup(src.getAddress());
    if (sender == null) {
      if (Debug.DEBUG)
        Debug.println(20, "New host for the group");
      thosts.insert(new HostData(src));
      src.grouplist.insert(this);
      for (int i=0; i < hosts.length; i++) {
        scan = (HostData) thosts.lookup(hosts[i].getAddress());
        if (scan == null) {
          thosts.insert(new HostData(hosts[i]));
          hosts[i].grouplist.insert(this);
        }
      }
      if (Debug.DEBUG) {
        Debug.print(20, "Trset:");
        for (int i=0; i <trset.length; i++)
          Debug.print(20, " " + trset[i].getInetAddress());
      }
      if (Util.in(trset, src))
        handleSuspect(nsuspects, trset, nrset, nuset);
    } else if (sender.leaving) {
      sender.leaving = false;
      if (status == S_IDLE && Util.in(rset, src))
        synchronizationPhaseInit();
    }
  }

  /**
   *
   */
  void leave(Host src)
  {
    HostData sender = (HostData) thosts.lookup(src.getAddress());
    if (sender != null) {
      sender.leaving = true;
      switch (status) {
        case S_IDLE:                                          // IDLE PHASE
          synchronizationPhaseInit();
          break;
        case S_SYNCH:                                         // SYNCHRONIZATION PHASE
          estimate.remove(src);
          if (estimate.isSynchronized())
            estimateExchangePhaseInit();
          break;
        case S_ESTIMATE:                                      // ESTIMATE EXCHANGE PHASE
          estimate.remove(src);
          sendEstimate();
          break;
      }
    }
  }

  /**
   * Chiamato da activateGroups
   */
  void handleSuspect(int nsuspects, MssHost[] trset, MssHost[] nrset, MssHost[] nuset)
  {
    if (nsuspects > this.nsuspects) {
      if (Debug.DEBUG)
        Debug.println(20, "handleSuspect");
      this.nsuspects = nsuspects;
      mss.msend(R_SYMM, MsgSymm.encode(mfactory, key, me.version, thosts.extract(nrset), rset), nrset);
      rset = trset;

      switch (status) {

        case S_IDLE:                                            // IDLE PHASE
          synchronizationPhaseInit();
          break;

        case S_SYNCH:                                           // SYNCHRONIZATION PHASE
          estimate.remove(nuset);
          if (estimate.isSynchronized())
            estimateExchangePhaseInit();
          break;

        case S_ESTIMATE:                                        // ESTIMATE EXCHANGE PHASE
          estimate.remove(nuset);
          sendEstimate();
          break;
      }
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Interactions with members
  ////////////////////////////////////////////////////////////////////////////////////////////

  void handleLocalDlvrAck(UpcallDlvrAck upcall)
  {
    HostData            sender;         // Original sender of the acknowledged message
    HostData            hd;             // Original sender of the acknowledged message
    boolean             carryon;        // Boolean flag to interrupt loops
    int                 i;              // Counter

    if (Debug.DEBUG)
      Debug.println(20, "handleLocalDlvr: msg.pos " + upcall.hpos + " " + upcall.mpos);
   sender = pview.hdata[upcall.hpos];
    sender.mack[upcall.mpos] = upcall.mid;
    OutMessage mresult = MsgResult.encode(mfactory, key, pvid, upcall.mid, upcall.sid, 
                          sender.delivered, upcall.hpos, upcall.vpos, upcall.result);
    mss.msend(R_RESULT, mresult, sender.host);
    mss.msend(R_RESULT, mresult); // FIXME Optimize this!!!!

    if (upcall.mid == sender.delivered+1) {      // Maybe the n. of dlvrd msgs has increased
      carryon = true;
      for (i=0; i<sender.mack.length && carryon; i++)
        carryon = (pview.mdata[i].leaving || sender.mack[i] >= upcall.mid);
      if (carryon) {
        sender.delivered = upcall.mid;
        for (i=0; i<pview.hsize && !carryon; i++) {
          hd = pview.hdata[i];
          carryon = (hd.hack[hpos] == hd.delivered);
        }
        if (carryon) {
          msgsStable = true;
          if (status == S_ESTIMATE) {
            proposal.encodeMsgsProp(estimate.hdata, estimate.hsize);
            sendProposal();
          }
        }
      }
    }
  }

  void handleLocalMcast(UpcallMcast upcall)
  {
    MsgMcast         msg;           // Message to send

    if (Debug.DEBUG)
      Debug.println(20, "handleLocalMcast");
    msg = new MsgMcast(upcall.stream, upcall.getId(), upcall.flag, myaddress, upcall.ackl, upcall.sid);
    if (status == S_IDLE)
      acceptMessage(msg);
    else if (!viewStable)
      acceptLateMessage(msg);
    else
      delayedSent.insert(msg);
  }

  void handleLocalPrepareAck(UpcallAck upcall)
  {
    if (Debug.DEBUG)
      Debug.println(20, "handleLocalPrepareAck");
    // TO BE IMPLEMENTED...
  }


  void handleLocalViewAck(UpcallAck upcall)
  {
    if (Debug.DEBUG)
      Debug.println(20, "handleLocalViewAck");
    pview.mdata[upcall.mpos].lastack = upcall.last;
    viewStable = true;
    for (int i=0; i < pview.msize && viewStable; i++)
      viewStable = (pview.mdata[i].leaving || pview.mdata[i].lastack == ninstalled);
    if (status == S_ESTIMATE)
      sendProposal();
  }

  void handleLocalJoin(UpcallGroup upcall)
  {
    if (Debug.DEBUG)
      Debug.println(20, "handleLocalJoin");
    if (tmembers.insert(new MemberData(upcall.lid, upcall.layer))) {
      nmembers++;
      if (status == S_IDLE) {
        synchronizationPhaseInit();
      } else if (status == S_ESTIMATE) {
        proposal.encodeMembProp(tmembers, nmembers);
        sendProposal();
      }
    }
  }

  void handleLocalLeave(UpcallGroup upcall)
  {
    MemberData          md;             // Sender member and additional data

    if (Debug.DEBUG)
      Debug.println(20, "handleLocalLeave");
    md = (MemberData) tmembers.lookup(upcall.lid);
    if (!md.leaving) {
      md.leaving = true;
      nmembers--;
      if (status == S_IDLE)
        synchronizationPhaseInit();
      else if (status == S_ESTIMATE) {
        proposal.encodeMembProp(tmembers, nmembers);
        sendProposal();
      }
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Interactions with remote hosts (group-related messages)
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Handles a mcast invocation received by a remote host; if the host is in Estimate
   * Exchange Phase, the message is buffered; otherwise, if the view of the message
   * corresponds to the current one, the message is locally delivered and an ack
   * message is m-sent.
   *
   * @param     msg     the m-received message.
   * @param     src     the sender of the message.
   */
  void handleMsgMcast(MsgMcast msg, Host src)
  {
    if (Debug.DEBUG)
      msg.print();
    if (status == S_ESTIMATE) {
      delayedRcvd.insert(msg);
    }
    else if (pvid == msg.vid) {
      localDeliver(msg);
      pview.hdata[msg.hpos].checkAck();
      mss.msend(R_ACK, MsgAck.encode(mfactory, key, pvid, hpos, pview.hdata, pview.hsize), pview.hlist, pview.hsize);
    }
  }

  /**
   *
   *
   * @param     msg     the m-received message.
   * @param     src     the sender of the message.
   */
  void handleMsgAck(MsgAck msg, Host src)
  {
    HostData            hd;             // Reference to additional host data

    if (Debug.DEBUG)
      msg.print();
    if  (pvid == msg.vid) {
      if (msg.ack.length == pview.hsize) {
        for (int i=0; i<msg.ack.length; i++) {
          hd = pview.hdata[i];
          if (hd.hack[msg.hpos] < msg.ack[i]) {
            hd.hack[msg.hpos] = msg.ack[i];
            hd.checkAck();
          }
        }
      } else {
        Util.panic("handleMsgAck");
      }
    }
  }


  /**
   * First of all, the method checks whether (i) the group is known; (ii) the view associated
   * to the message is equal to the current one; (iii) the messge is unknown; (iv) the
   * forwarding sender is still in the estimate. If so, the message is delivered and an
   * ack message is sent. The message is forwarded again to the remaining members.
   *
   * @param     msg     the m-received message.
   * @param     src     the sender of the message.
   */
  void handleMsgDlvrd(MsgMcast msg, Host src)
  {
    if (Debug.DEBUG) {
      msg.print();
      Debug.println(20, "msg.hpos: " + msg.hpos + "hpos" + hpos);
      Debug.println(20, "Group.pvid: " + pvid + " msg.vid: " + msg.vid + " msg.mid " + msg.mid);
    }
    if  ( pvid == msg.vid && estimate.lookup(src.getAddress()) != null &&
          msg.mid > pview.hdata[msg.hpos].hack[hpos] ) {
      localDeliver(msg);
      pview.hdata[msg.hpos].checkAck();
      OutMessage ack = MsgAck.encode(mfactory, key, pvid, hpos, pview.hdata, pview.hsize);
      mss.msend(R_ACK, ack, pview.hlist, pview.hsize);
      if (status == S_ESTIMATE)
        mss.msend(R_FORWARD, msg.outmsg, estimate.hlist, estimate.hsize);
    }
  }

  /**
   *
   * @param     msg     the m-received message.
   * @param     src     the sender of the message.
   */
  void handleMsgSynch(MsgSynch msg, Host src)
  {
    HostData            sender;         // Reference to HostData object of the sender

    if (Debug.DEBUG)
      msg.print();
    if ((sender = (HostData) thosts.lookup(src.getAddress())) != null)

      switch(status) {

      case S_IDLE:                                            // IDLE PHASE
        if (Debug.DEBUG)
          Debug.println(20, "Idle phase");
        if (sender.version < msg.vsend) {
          sender.version = msg.vsend;
          if (Util.in(rset, sender.host))
            synchronizationPhaseInit();
        }
        break;

      case S_SYNCH:                                           // SYNCHRONIZATION PHASE
        if (Debug.DEBUG)
          Debug.println(20, "S phase " + sender.version + " " + msg.vsend);
        if (me.version == msg.vdest) {
          estimate.setSynchronized(sender);
          if (sender.agreed < msg.vsend)
            sender.agreed = msg.vsend;
        }
        if (sender.version < msg.vsend) {
          sender.version = msg.vsend;
          OutMessage msynch =
            MsgSynch.encode(mfactory, key, me.version, sender.version, sender.host.symset);
          mss.msend(R_SYNCH, msynch, src);

        }
        if (estimate.isSynchronized())
          estimateExchangePhaseInit();
        break;

      case S_ESTIMATE:                                        // ESTIMATE EXCHANGE PHASE
        if (Debug.DEBUG)
          Debug.println(20, "EE phase ");
        if (sender.version < msg.vsend)
          sender.version = msg.vsend;
        if (Debug.DEBUG)
          Debug.println(20, "sender.agreed " + sender.agreed + " msg.vsend " + msg.vsend );
        if (sender.agreed < msg.vsend && estimate.lookup(src.getAddress()) != null) {
          estimate.remove(msg.rset);
          sendEstimate();
        }

      } // END switch
  }


  /**
   *
   * @param     msg     the m-received message.
   * @param     src     the sender of the message.
   */
  void handleMsgSymm(MsgSymm msg, Host src)
  {
    HostData            sender;         // Reference to sender host information
    int                 vdest;          // Version number of the destination
    int                 pos;            // My position in the list of messages

    if (Debug.DEBUG)
      msg.print();
    sender = estimate.lookup(src.getAddress());
    pos = Util.valueAt(msg.hosts, myaddress);
    if (pos >= 0) {
      vdest = msg.version[pos];

      if (status == S_SYNCH) {             // SYNCHRONIZATION PHASE
        if (Debug.DEBUG)
          Debug.println(20, "MsgSymm: S PHASE");
        if (me.version == vdest) {
          if (Debug.DEBUG)
            Debug.println(20, "MsgSymm: accepted");
          estimate.remove(msg.rset);
          if (estimate.isSynchronized())
            estimateExchangePhaseInit();
        }

      } else if (status == S_ESTIMATE) {   // ESTIMATE EXCHANGE PHASE
        if (Debug.DEBUG)
          Debug.println(20, "MsgSymm: EE PHASE");
        if (sender != null && me.agreed == vdest && sender.agreed == msg.vsend) {
          if (Debug.DEBUG)
            Debug.println(20, "MsgSymm: accepted");
          estimate.remove(msg.rset);
          sendEstimate();
        }
      }
    }
  }

  /**
   *
   * @param     msg     the m-received message.
   * @param     src     the sender of the message.
   */
  void handleMsgEstim(MsgEstim msg, Host src)
  {
    HostData            sender;         // Reference to sender information
    int                 vdest;          // Version number of destination

    if (Debug.DEBUG)
      msg.print();
    sender = (HostData) thosts.lookup(src.getAddress());
    vdest = msg.agreed[Util.valueAt(msg.hosts, myaddress)];
    if (Debug.DEBUG) {
      Debug.println(20, "vdest " + vdest);
      Debug.println(20, "me.version " + me.version);
    }

    if (status == S_SYNCH) {           // SYNCHRONIZATION PHASE

      if (sender.version < msg.vsend)
        sender.version = msg.vsend;
      if (!sender.inEstimate) {
        OutMessage msymm =
          MsgSymm.encode(mfactory, key, me.version, sender, estimate.hlist, estimate.hsize);
        mss.msend(R_SYMM, msymm, src);
      } else if (me.version == vdest) {
        estimate.intersect(msg.hosts, msg.agreed);
        estimateExchangePhaseInit();
      }

    } else if (status == S_ESTIMATE) {         // ESTIMATE EXCHANGE PHASE

      if (sender.inEstimate && checkAgreed(msg.hosts, msg.agreed)) {
        estimate.intersect(msg.hosts, null);
        sendEstimate();
      }
    }
  }


  /**
   *
   * @param     msg     the m-received message.
   * @param     src     the sender of the message.
   */
  void handleMsgProp(MsgProp msg, Host src)
  {
   // Variables containg the new view description (both for message and local use)
    long                cvid;           // New complete view identifier
    int[]               cvcomp;         // New complete view composition
    int[][]             members;        // Decoded members
    int[]               pvids;          // Local view indexes
    int                 pvid;           // Index of subview id of this host

    // Variable used to build the view message information
    HostData            sender;         // Reference to host information of the sender
    HostData            scan;           // Used to scan tcomp
    MsgProp[]           props;          // Array of proposals
    Host[]              dlist;          // Destination list of view message
    int[][]             prev_vcomp;     // Previous complete view compositions
    long[]              prev_vid;       // Previous view id
    boolean             partial;        // True if partial views are needed
    int                 npartial;       // Number of different partial views
    long[]              diff_vid;       // Distinct previous view id
    int                 len;            // Short for new_vcomp.length
    int                 i,j,k;          // Counters


    if (Debug.DEBUG)
      msg.print();
    if ( status != S_IDLE &&
         (sender = (HostData) thosts.lookup(src.getAddress())) != null &&
         sender.inEstimate ) {
      sender.ctbl = msg;
      if (sender == me)
        nproposals--;
      if (status == S_ESTIMATE &&
          (sender != me || nproposals == 0) &&
          checkAgreement()
          ) {

        // Creates a new complete view identifier and decodes the next view composition
        cvid   = Types.createsVid(myaddress, ++ncoordinated);
        cvcomp = me.ctbl.decodeHostProp_hosts();

        // Allocates arrays
        len = cvcomp.length;
        props        = new MsgProp[len];
        prev_vid     = new long[len];
        prev_vcomp   = new int [len][];
        members      = new int [len][];
        dlist        = new Host[len];

        // Initializes props and dlist
        scan = (HostData) thosts;
        try {
        for (i=0; i < len; i++) {
          scan = (HostData) scan.lookup(cvcomp[i]);
          if (Debug.DEBUG) {
            Debug.print(20, " " + (cvcomp[i] & 0xFF));
            Debug.print(20, " " + (scan.host.getAddress() & 0xFF));
          }
          props[i] = scan.ctbl;
          dlist[i] = scan.host;
        }
        } catch (Exception e) {
          if (Debug.DEBUG)
            Util.panic("Corrupted proposals", e);
        }
        if (Debug.DEBUG)
          Debug.println(20, "");

        // Decodes proposals
        i = 0;
        for (i=0; i < len; i++) {
          prev_vid[i]   = props[i].cvid;
          prev_vcomp[i] = props[i].decodeLastProp();
          members[i]    = props[i].decodeMembProp();
        }


        // Checks whether partial views are needed
        // ??? Si noti che questo puo' essere ottimizzato
        // in modo da essere sicuri che che sia O(n^2)
        partial = false;
        for (i=0; i < len && !partial; i++)
          for (j=0; j < prev_vcomp[i].length && !partial; j++) {
            k = Util.valueAt(cvcomp, prev_vcomp[i][j]);
            partial = ( k >= 0 && prev_vid[i] != prev_vid[k] );
          }

        // Constructs partial view index
        pvids = new int[len];
        pvid = 0;
        if (partial) {
          npartial = 0;
          diff_vid = new long[len];
          for (i=0; i < len; i++) {
            for (j=0; j < npartial && prev_vid[i] != diff_vid[j]; j++);
            if (j == npartial) {
              diff_vid[j] = prev_vid[i];
              npartial++;
            }
            pvids[i] = j;
            if (cvcomp[i] == myaddress) {
              pvid = j;
            }
          }
          ncoordinated += npartial;
        }

        // Reliable m-send the view message containing the information
        mss.rmsend(R_VIEW, MsgView.encode(mfactory, key, cvid, me.ctbl, pvids, props), dlist);

        // Locally installs the view
        int[] incarns = me.ctbl.decodeHostProp_incarns();
        installView(cvid, cvcomp, incarns, members, pvids, pvid);
      }
    }
  }

  /**
   * Locates the position of this host in the complete view and checks whether the
   * version number in the message is equal to the current one.
   *
   * @param     msg     the m-received message.
   * @param     src     the sender of the message.
   */
  void handleMsgView(MsgView msg, Host src)
  {
    int                 pos;            // Counter

    if (Debug.DEBUG)
      msg.print();
    if (status == S_ESTIMATE && estimate.lookup(src.getAddress()) != null) {
      pos = Util.valueAt(msg.cvcomp, myaddress);
      if (pos >= 0 && msg.agreed[pos] == ninstalled+1)
        installView(msg.cvid, msg.cvcomp, msg.incarns, msg.members, msg.pvids, msg.pvids[pos]);
    }
  }

  /**
   *
   * @param     msg     the m-received message.
   * @param     src     the sender of the message.
   */
  void handleMsgResult(MsgResult msg, Host src)
  {
    sbuffer.insertMsgResult(msg);
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods related to the total composition of the group
  ////////////////////////////////////////////////////////////////////////////////////////////

  /*
   * Removes MemberData item which have requested to exit and
   * have been excluded from the current view.
   */
  private void removeLeavingMembers(int gid, int last)
  {
    MemberData          scan;                   // Used to scan the list of members
    MemberData          prev;                   // Scan = prev.next

    prev = (MemberData) tmembers;
    scan = (MemberData) prev.getFirst();
    while (scan != null) {
      if (scan.leaving && scan.last != last) {
        scan.layer.leaveAck(gid);
        prev.remove(scan.key);
      } else
        prev = scan;
      scan = (MemberData) scan.getNext();
    }
  }

  /*
   *
   */
  private void removeLeavingHosts(int key, int last)
  {
    HostData            scan;                   // Used to scan the list of members
    HostData            prev;                   // scan = prev.next

    prev = (HostData) thosts;
    scan = (HostData) prev.getFirst();
    while (scan != null) {
      if (scan.leaving && scan.last != last) {
        mss.msend(Tag.R_LEAVE, MsgGroup.encode(mfactory, key, this), scan.host);
        prev.remove(scan.key);
      } else
        prev = scan;
      scan = (HostData) scan.getNext();
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods related to the current view estimate
  ////////////////////////////////////////////////////////////////////////////////////////////

  /*
   *
   */
  private void synchronizationPhaseInit()
  {
    HostData            hd;             // Optimization variable

    if (Debug.DEBUG)
      Debug.println(20, "synchronizationPhaseInit");
    status = S_SYNCH;
    estimate.init(thosts, rset);
    me.inSynchronized = true;
    me.version++;
    me.agreed++;
    for (int i=0; i < estimate.hsize; i++)
      if ((hd = estimate.hdata[i]) != me) {
        OutMessage msynch =
          MsgSynch.encode(mfactory, key, me.version, hd.version, hd.host.symset);
        mss.msend(R_SYNCH, msynch, hd.host);
      }
    if (estimate.isSynchronized())
      estimateExchangePhaseInit();
  }


  /**
   *
   */
  private void estimateExchangePhaseInit()
  {
    Object[]        buffer;         // Messages still in the buffer
    MsgMcast        msg;            // MsgDlvrd message

    if (Debug.DEBUG)
      Debug.println(20, "estimateExchangePhaseInit");
    status = S_ESTIMATE;
    mss.msend(R_ESTIM, MsgEstim.encode(mfactory, key, me.version, estimate.hdata,
              estimate.hsize), estimate.hlist, estimate.hsize);
    proposal.encodeLastProp(cview, cview.length);
    proposal.encodeHostProp(estimate.hdata, estimate.hsize);
    proposal.encodeMembProp(tmembers, nmembers);
    if (msgsStable) {
      proposal.encodeMsgsProp(estimate.hdata, estimate.hsize);
      sendProposal();
    }
    for (int i=0; i< pview.hsize; i++) {
      buffer = pview.hdata[i].buffer.toArray();
      for (int j=0; j < buffer.length; j++) {
        msg = (MsgMcast) buffer[j];
        mss.msend(R_FORWARD, msg.outmsg, pview.hlist, pview.hsize);
      }
    }
  }


  /*
   *
   */
  private void sendProposal()
  {
    if (msgsStable && viewStable) {
      mss.msend(R_PROP, proposal.encode(mfactory, key, cvid, pvid), estimate.getCoordinator());
      if (estimate.getCoordinator() == me.host)
        nproposals++;
    }
  }


  /*
   *
   */
  private void sendEstimate()
  {
    if (Debug.DEBUG) {
      Debug.print(20, "Estimate: ");
      for (int i=0; i<estimate.hsize; i++)
        Debug.print(20, (estimate.hdata[i].host.getAddress() & 0xFF) + " ");
      Debug.println(20, "");
    }
    if (estimate.hasChanged()) {
      OutMessage mestim =
        MsgEstim.encode(mfactory, key, me.version, estimate.hdata, estimate.hsize);
      mss.msend(R_ESTIM, mestim, estimate.hlist, estimate.hsize);
      proposal.encodeHostProp(estimate.hdata, estimate.hsize);
      sendProposal();
    }
  }

  /*
   *
   */
  private boolean checkAgreement()
  {
    MsgProp          pp;             // Proposal of this host
    MsgProp          pq;             // Proposal of compared host
    int              i,j;            // Counters

    if ((pp = me.ctbl) == null)
      return false;
    for (i=0; i < estimate.hsize; i++) {
      pq = estimate.hdata[i].ctbl;
      if (!pp.checkEstimate(pq))
        return false;
      for (j=i+1; j < estimate.hsize; j++)
        if (!pq.checkMessages(estimate.hdata[j].ctbl))
          return false;
    }
    return true;
  }



  /*
   * Returns true if the agreed version numbers of intersection are equal ???
   */
  boolean checkAgreed(int[] hosts, int[] agreed)
  {
    int         i=0, j=0;               // Counters

    while (i < hosts.length && j < estimate.hsize)
      if (hosts[i] == estimate.hdata[j].key) {
        if (agreed[i] != estimate.hdata[j].agreed)
          return false;
        i++; j++;
      }
      else if (hosts[i] < estimate.hdata[j].key) {
        i++;
      }
      else {
        j++;
      }

    return true;
  }


  /**
   *
   */
  private void localDeliver(MsgMcast msg)
  {
    MemberData          md;             // Used to scan <m_vcomp>
    HostData            hd;             // Additional info of the original sender

    if (Debug.DEBUG)
      Debug.println(20, "Daemon: localDeliver");
    for (int i=0; i<pview.msize; i++) {
      md = pview.mdata[i];
      if (!md.leaving)
        md.layer.notifyDlvr(key, msg.flag, msg.vid, msg.hpos, 
          msg.mid, msg.sid, msg.ackr, msg.getInputStream());
    }
    hd = pview.hdata[msg.hpos];
    hd.hack[hpos] = msg.mid;
    hd.hack[msg.hpos] = msg.mid;
    hd.buffer.insert(msg);
    msgsStable = false;
  }


  /*
   *
   */
  private void acceptMessage(MsgMcast msg)
  {
    if (Debug.DEBUG)
      Debug.println(20, "Daemon: acceptMessage");
    
    msg.complete(hpos, pvid, sbuffer.getMid());
    sbuffer.insertMsgMcast(msg);
    localDeliver(msg);
    me.checkAck();
    if (Debug.DEBUG)
      msg.outmsg.print(-10);
    
    mss.msend(R_MCAST, msg.outmsg, pview.hlist, pview.hsize);
  }

  private void acceptLateMessage(MsgMcast msg)
  {
    if (Debug.DEBUG)
      Debug.println(20, "Daemon: acceptLateMessage");

    msg.complete(hpos, pvid, sbuffer.getMid());
    sbuffer.insertMsgMcast(msg);
    localDeliver(msg);
    me.checkAck();
    /* FIXME: modify in order to send only to members remained in estimate */ 
    mss.msend(R_FORWARD, msg.outmsg, pview.hlist, pview.hsize);
  }


  /**
   *
   */
  private void installView(long cvid, int[] hosts, int[] incarns,
                           int[][] members, int[] pids, int pid)
  {
    MemberId[]    memberIds;      //
    int[]         localMembers;   //
    ViewImpl      view;           // View to be passed to hosts
    HostData      hd;             // Used to scan thosts
    MemberData    md;             // Used to scan m_vcomp
    boolean       stable;         // True if the current view is stable

     // Stores complete view information
    this.cvid = cvid;
    this.cview = (int[]) cview.clone();

    // Set view index to -1 to all hosts; this would indicate
    // that no host is in the current view. Then, set view index
    // to the right value for hosts included in the current view
    hd = (HostData) thosts.getNext();
    while (hd != null) {
      hd.viewIndex = -1;
      hd = (HostData) hd.getNext();
    }

    // Computes host position in the current view and removes info related to
    // hosts and members not participating in the next (partial) view.
    ArrayList v = new ArrayList(hosts.length);
    localMembers = null;
    int hostCount = 0;
    for (int i=0; i < hosts.length; i++) {
      if (pids[i] == pid) {
        if (hosts[i] == myaddress) {
          hpos = hostCount;
          localMembers = members[i];
        }
        hd = (HostData) thosts.lookup(hosts[i]);
        for (int j=0; j < members[i].length; j++)
          v.add(new MemberIdImpl(hd.host.getInetAddress(), incarns[i], members[i][j]));
        hosts[hostCount] = hosts[i];
        hd.viewIndex = hostCount++;
      }
    }
    memberIds = (MemberId[]) v.toArray(new MemberId[v.size()]);

    // Stores current view information
    ninstalled++;
    stable = pview.init(thosts, hosts, hostCount, ninstalled, rset, tmembers, localMembers);

    // Updates group flags
    msgsStable = true;
    viewStable = false;
    status = S_IDLE;

    // Compute the partial view id
    pvid       = Types.computePartialVid(cvid, pid);

    // Create the new view
    view = new ViewImpl(key, pvid, memberIds);
    if (Debug.DEBUG) {
      Debug.println(20, "Stable view: " + stable);
      Debug.print(20, " Vid: " + pvid);
      Debug.println(20, " nviewInst: " + ninstalled);
    }

    // Deliver the view to non-leaving members
    for (int i=0; i < pview.msize; i++)
      if ((md = pview.mdata[i]) != null && !md.leaving)
        md.layer.notifyView(key, view, ninstalled, hpos, i);

    // Remove hosts and members that have request to leave
    removeLeavingHosts(key, ninstalled);
    removeLeavingMembers(key, ninstalled);

    if (!tmembers.isEmpty()) {            // The group has not to be removed

      // Updates HostData information
      hd = (HostData) thosts.getNext();
      while (hd != null) {
        hd.ctbl = null;
        hd.hack = new int[pview.hsize];
        hd.mack = new int[pview.msize];
        hd.delivered = 0;
        hd.hmin = 0;
        hd.buffer.clear();
        hd = (HostData) hd.getNext();
      }

     // Updates information about messages
      sbuffer.setView(view, hostCount);
      nproposals = 0;
      MsgMcast scan;
      while ((scan = (MsgMcast) delayedSent.removeFirst()) != null)
        acceptMessage(scan);
      while ((scan = (MsgMcast) delayedRcvd.removeFirst()) != null)
        if (scan.vid == pvid)
          localDeliver(scan);
      if (!sbuffer.isEmpty()) {
        OutMessage mack =
          MsgAck.encode(mfactory, key, pvid, hpos, pview.hdata, pview.hsize);
        mss.msend(R_ACK, mack, pview.hlist, pview.hsize);
      }

      // Checks if the view is stable
      if (!stable)
        synchronizationPhaseInit();

    }
  }

} // END Group

