/*
 * Copyright (c) 1998-2000 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package relacs.daemon;

import java.net.*;
import java.io.*;

import jgroup.*;

/**
 *  Implementation class of <tt>jgroup.MemberId</tt>.
 *
 *  @author Alberto Montresor
 *  @since  Jgroup 0.9
 */
public class  MemberIdImpl
  implements    MemberId, Serializable // FIXME Should be Externalizable
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  private InetAddress   address;        // IP address of the machine hosting the member
  private int           incarnation;    // Value used to distinguish different hosts
  private int           count;          // Counter

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Default constructor
   */
  public MemberIdImpl()
  {
  }

  /**
   *  Builds a new MemberId with the specified arguments.
   */
  MemberIdImpl(InetAddress address, int incarnation, int count)
  {
    this.address = address;
    this.incarnation = incarnation;
    this.count = count;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from MemberId interface
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns the IP identifier of the Jgroup daemon hosting
   *  this member.
   */
  public InetAddress getAddress()
  {
    return address;
  }


  /**
   *  Returns an integer used to distinguish this particular
   *  member from other members of the same group hosted in
   *  the same JVM.
   */
  public int getCounter()
  {
    return count;
  }


  /**
   *  Returns true if this member id and the specified id have been created
   *  in the same Java virtual machine, and this id is more recent than the
   *  the specified id; returns false otherwise. This method can be used
   *  to distinguish hosts that have performed a recovery.
   */
  public boolean isNewer(MemberId obj)
  {
    if (obj == null)
      return false;
    MemberIdImpl id = (MemberIdImpl)obj;
    return (incarnation > id.incarnation && address.equals(id.address));
  }


  /**
   *  Returns true if this member id is a neighbour (i.e., belong to the same JVM) of the
   *  the specified member id.
   */
  public boolean isNeighbour(MemberId obj)
  {
    if (obj == null)
      return false;
    MemberIdImpl id = (MemberIdImpl)obj;
    return (incarnation == id.incarnation && address.equals(id.address));
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Package methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns a long representing this member if. This value is used in the code of the 
   *  daemon and the group manager packages to uniquely identify a member partecipating in
   *  a view.
   */
  public long getLongIdentifier()
  {
    int v1 = address.hashCode();
    int v2 = count;
    return ((long)(v1) << 32) | (v2 & 0xFFFFFFFFL);
  }
  
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Other methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns a string representation of the object.
   */
  public String toString()
  {
    return "["+address.toString()+","+incarnation+","+count+"]";
  }


  /**
   * Returns a hashcode for the <code>MemberId</code>.  Two <code>MemberId</code>s
   * will have the same hashcode if they are equal with respect to their
   * content.
   *
   * @return the hashcode
   */
  public int hashCode() {
    return address.hashCode() + (int) incarnation + (int) count;
  }

  /**
   * Compares two <code>MemberId</code>s for content equality.
   *
   * @param obj the Object to compare with
   * @return  true if these Objects are equal; false otherwise.
   */
  public boolean equals(Object obj)
  {
    MemberIdImpl id;
    if (obj == null)
      return false;
    try {
      id = (MemberIdImpl) obj;
    } catch (ClassCastException e) {
      return false;
    }
    return (incarnation == id.incarnation && count == id.count && address.equals(id.address));
  }

} // END MemberIdImpl

