/*
 * Copyright (c) 1998-2000 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package relacs.daemon;

import relacs.mss.*;
import debug.Debug;

/**
 * <code>MsgEstim</code> messages are used during the ee-phase (estimate exchange) of
 * the view agreement algorithm. They are sent to daemons belonging to the current
 * estimate whenever this estimate changes.
 *
 * @author  Alberto Montresor
 * @since   Jgroup 0.1
 */
final class MsgEstim
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Unique message instance
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * As these messages are never stored by the daemon, a single message instance is
   * sufficient. In this way, we save the costs of allocating and garbage collecting
   * messages.
   */
  private static MsgEstim msg;

  static {
    msg = new MsgEstim();
  }

 ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Agreed version number of the sender */                int       vsend;
  /** Agreed version number table of next view estimate */  int[]     agreed;
  /** Next view estimate */                                 int[]     hosts;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Message to send
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Returns a <code>MsgEstim</code> object that contains the decoding of the m-received input
   * stream.
   *
   * @param stream    the message input stream to decode
   */
  static MsgEstim decode(InMessage stream)
  {
    msg.vsend  = Types.readVer(stream);
    int len = Types.readGrp(stream);            // Read length of agreed and hosts
    msg.agreed = new int[len];
    msg.hosts  = new int[len];
    for (int i=0; i<len; i++) {                   // Read agreed and hosts
      msg.agreed[i] = Types.readVer(stream);
      msg.hosts[i]  = Types.readHid(stream);
    }

    return msg;
  }

  /**
   * Returns a output stream containing the encoding of a <code>MsgEstim</code> message.
   *
   * @param     gid       group identifier
   * @param     vsend     version number of the sender
   * @param     estimate  current estimate
   * @param     elen      actual length of estimate (estimate.length >= elen)
   */
  static OutMessage encode(MsgFactory mfactory, int gid, int vsend, HostData[] estimate, int elen)
  {
    // Allocates data buffer
    int size = Types.GID_SIZE + Types.VER_SIZE + Types.GRP_SIZE +
               (Types.VER_SIZE + Types.HID_SIZE) * elen;
    OutMessage stream = mfactory.get(size);

    Types.writeGid(stream,  gid);
    Types.writeVer(stream, vsend);

    Types.writeGrp(stream, elen);             // Write length of agreed and hosts
    for (int i=0; i < elen; i++) {                // Write agreed and hosts
      Types.writeVer(stream, estimate[i].agreed);
      Types.writeHid(stream, estimate[i].key);
    }

    return stream;
  }

  void print()
  {
    if (Debug.DEBUG) {
      Debug.print(20, "Vsend: " + vsend + " | Hosts: ");
      for (int i=0; i < hosts.length; i++)
        Debug.print(20, (hosts[i] & 0xFF) + " ");
      Debug.print(20, "| Agreed: ");
      for (int i=0; i < agreed.length; i++)
        Debug.print(20, agreed[i] + " ");
      Debug.println(20, "");
    }
  }

} // END MsgEstim
