/*
 * Copyright (c) 1998-2000 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

/*
 *  History:
 *  22/11/2000: Added field sid, used to distinguish messages with
 *    the same id but originated by different senders.
 */


package relacs.daemon;

import java.io.*;

import jgroup.multicast.AckListener;
import relacs.mss.*;
import debug.Debug;

/**
 * The <code> MsgMcast</code> class
 *
 * @author  Alberto Montresor
 * @since   Jgroup 0.1
 */
final class     MsgMcast
{


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Position constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  private static final int START_FIXED = 0;
  private static final int START_VAR   = Types.GID_SIZE + Types.FLG_SIZE*2 + Types.HID_SIZE +
                                         Types.SID_SIZE;
  private static final int START_DATA  = START_VAR + Types.GRP_SIZE + Types.MID_SIZE +
                                         Types.VID_SIZE;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Serialized fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  boolean               flag;           // Single object flag
  boolean               ackr;           // True if ack is required
  int                   hid;            // Indentifier of the sender host
  long                  sid;            // Sender identifier
  int                   hpos;           // Position of the sender in view identified by <vid>
  long                  vid;            // View identifier
  int                   mid;            // Identifier of this message
  AckListener           ackl;           // Ack listener
  OutMessage            outmsg;         // Message to send
  InMessage             inmsg;          // Message m-received

  // Ack management
  private boolean[]     ackset;         // Set of acks
  private int           nfalse;         // Number of false values in ackset

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors and additional methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /*
   *  Constructs a MsgMcast data structure starting from a m-received message
   */
  private MsgMcast(InMessage inmsg)
  {
    inmsg.print(0);
    flag = Types.readFlg(inmsg);
    ackr = Types.readFlg(inmsg);
    hid  = Types.readHid(inmsg);
    sid  = Types.readSid(inmsg);
    hpos = Types.readGrp(inmsg);
    vid  = Types.readVid(inmsg);
    mid  = Types.readMid(inmsg);

    this.inmsg = inmsg;
    inmsg.seek(START_DATA);
    outmsg = new OutMessage(inmsg);
  }


  /**
   * Creates an uncompleted <code>MsgMcast</code> object; other fields will be set when the
   * information will be available.
   *
   * @param     gid     group identifier
   */
  MsgMcast(OutMessage outmsg, int gid, boolean flag, int hid, AckListener ackl, long sid)
  {
    this.outmsg = outmsg;

    // Store info for local message
    this.flag = flag;
    this.ackl = ackl;
    this.ackr = (ackl != null);
    this.hid  = hid;
    this.sid  = sid;

    // Writes fixed part of the header
    outmsg.seek(START_FIXED);
    Types.writeGid(outmsg,  gid);
    Types.writeFlg(outmsg, flag);
    Types.writeFlg(outmsg, ackr);
    Types.writeHid(outmsg,  hid);
    Types.writeSid(outmsg,  sid);

  }


  /**
   * Complete the message structure with the data related to the view in which
   * the message is sent.
   *
   * @param     hpos      host index in the view.
   * @param     vid       view identifier.
   * @param     mid       message identifier.
   * @param     senderPos identifier of the sender of this message
   */
  void complete(int hpos, long vid, int mid)
  {
    // Writes variable part of the header
    outmsg.seek(START_VAR);
    Types.writeGrp(outmsg, hpos);
    Types.writeVid(outmsg,  vid);
    Types.writeMid(outmsg,  mid);

    // Store info for local message
    this.vid  = vid;
    this.hpos = hpos;
    this.mid  = mid;

    inmsg = new InMessage(outmsg);
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /*
   *
   */
  static OutputStream createOutputStream(MsgFactory mfactory)
  {
    OutMessage stream = mfactory.get();
    stream.seek(START_DATA);
    return stream;
  }

  /**
   * Returns a <code>MsgMcast</code> object that contains the decoding of the m-received input
   * stream.
   *
   * @param outmsg    the message input stream to decode
   */
  static MsgMcast decode(InMessage inmsg)
  {
    return new MsgMcast(inmsg);
  }

  /*
   *
   */
  InputStream getInputStream()
  {
    inmsg.seek(START_DATA);
    return inmsg;
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Ack management
  ////////////////////////////////////////////////////////////////////////////////////////////

  void setAckArray(int len)
  {
    ackset = new boolean[len];
    nfalse = len;
  }

  void setAck(int pos)
  {
    if (!ackset[pos])
      nfalse--;
    ackset[pos] = true;
  }

  boolean isComplete()
  {
    return (nfalse == 0);
  }



  void print()
  {
    if (Debug.DEBUG) {
      Debug.print(20, "Msg: ");
      Debug.println(20, "");
    }
  }

} // END MsgMcast

