/*
 * Copyright (c) 1998-2000 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

/*
 *  History:
 *  22/11/2000: Added field sid, used to distinguish messages with
 *    the same id but originated by different senders.
 */

package relacs.daemon;

import java.io.*;

import relacs.mss.*;
import debug.Debug;

/**
 *
 * @author  Alberto Montresor
 * @since   Jgroup 0.8
 */
final class MsgResult
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** View identifier */
  long vid;
  
  /** Message id */
  int mid;
  
  /** Identifier of the sender of the message to which this
      message is a response */                             
  long sid;
  
  /** Number of messages delivered by all local members */
  int dlvr;
  
  /** Host index in the view */
  int hpos;
  
  /** Member index in the view */
  int vpos;
  
  /** Result value */
  Object result;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Returns a <code>MsgResult</code> object that contains the decoding of the m-received input
   * stream.
   *
   * @param stream    the message input stream to decode
   */
  static MsgResult decode(InMessage stream)
  {
    MsgResult msg = new MsgResult();        // FIXME Can we use a singleton?

    msg.vid   = Types.readVid(stream);
    msg.mid   = Types.readMid(stream);
    msg.sid   = Types.readSid(stream);
    msg.dlvr  = Types.readMid(stream);
    msg.hpos  = Types.readGrp(stream);
    msg.vpos  = Types.readGrp(stream);
    if (Types.readFlg(stream)) {
      try {
        ObjectInputStream objectstream = new ObjectInputStream(stream);
        msg.result = objectstream.readObject();
      } catch (Exception e) {
        /*
         * An exception has been thrown when unmarshaling result value; we just print a
         * warning message on the standard error.
         */
        System.err.println("Warning! Exception caught when unmarshaling result value: " + e);
        if (Debug.DEBUG)
          e.printStackTrace();
      }
    }
    return msg;
  }

  /**
   * Returns a output stream containing the encoding of a <code>MsgResult</code> message.
   *
   * @param     gid     group identifier
   * @param     vid     current view identifier
   * @param     mid     identifier of the original message to which this message is a response
   * @param     sid     identifier of the sender of the original message to which this message is
   *   a response
   * @param     hpos    host index in the view
   * @param     mpos    member index in the view
   * @param     result  result value
   */
  static OutMessage encode(MsgFactory mfactory, int gid, long vid, int mid, long sid, int dlvr,
                       int hpos, int vpos, Object result)
  {
    OutMessage stream;

    // Allocates data buffer
    if (result == null) {
      int size = Types.GID_SIZE + Types.VID_SIZE + Types.MID_SIZE * 2 +
                 Types.GRP_SIZE*2 + Types.SID_SIZE + Types.FLG_SIZE;
      stream = mfactory.get(size);
    } else {
      stream = mfactory.get();
    }

    // Writes single data
    Types.writeGid(stream, gid);
    Types.writeVid(stream, vid);
    Types.writeMid(stream, mid);
    Types.writeSid(stream, sid);
    Types.writeMid(stream, dlvr);
    Types.writeGrp(stream, hpos);
    Types.writeGrp(stream, vpos);
    Types.writeFlg(stream, (result != null));
    if (result != null) {
      try {
        ObjectOutputStream objectstream = new ObjectOutputStream(stream);
        objectstream.writeObject(result);
      } catch (Exception e) {
        /*
         * An exception has been thrown when marshaling result value; we just print a
         * warning message on the standard error.
         */
        System.err.println("Warning! Exception caught when marshaling result value: " + e);
        if (Debug.DEBUG)
          e.printStackTrace();
      }
    }

    return stream;
  }

  void print()
  {
    if (Debug.DEBUG) {
    }
  }

} // END MsgResult
