/*
 * Copyright (c) 1998-2000 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package relacs.daemon;

import relacs.mss.MssHost;
import relacs.mss.Mss;
import relacs.general.Util;
import debug.Debug;

/**
 * The <code> PartialView</code> class
 *
 * @author  Alberto Montresor
 * @since   Jgroup 0.3
 */
final class     PartialView
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  HostData[]            hdata;
  Host[]                hlist;
  int                   hsize;
  MemberData[]          mdata;
  int                   msize;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  PartialView(HostData hd, MemberData md)
  {
    hdata       = new HostData[1];
    hdata[0]    = hd;
    hlist       = new Host[1];
    hlist[0]    = hd.host;
    hsize       = 1;

    // Current view initialization (members)
    mdata       = new MemberData[1];
    mdata[0]    = md;
    msize       = 1;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Initializes the current partial view composition by extracting the HostData objects
   * from the  HostData list <thosts>; used to re-initializing groups after the receipt
   * of a view. Returns true if the host section of the view is stable (i.e., if
   * the host composition of the view corresponds to the current set of hosts
   * both reachable and interested in this group.
   *
   * @param     thosts          complete list of hosts interested in this group
   * @param     addrs           array of host identifier composing the new view
   * @param     length          actual size of addrs(length <= addrs.length)
   * @param     last            number of last installed view
   */
  boolean init(HostData thosts, int[] addrs, int length, int last,
                MssHost[] rset, MemberData tmembers, int[] lids)
  {
    boolean             stable = true;  // True if the current view is stable
    HostData            hscan;          // Used to scan thosts
    MemberData          mscan;          // Used to scan <tmembers>
    int                 i=0;            // Used to scan "addrs"

    // Reallocates host and member arrays if they aren't big enough
    if (hdata.length < length) {
      hdata = new HostData[length];
      hlist = new Host[length];
    }
    if (mdata.length < lids.length)
      mdata = new MemberData[lids.length];

    // Build host composition
    hscan = (HostData) thosts.getFirst();
    hsize = 0;
    while (hscan != null)
      if (i == length || hscan.key < addrs[i]) {
        if (!hscan.leaving && Util.in(rset, hscan.host)) {
          if (Debug.DEBUG)
            Debug.println(20, "host unstable because host reachable and not leaving");
          stable = false;
        }
        hscan = (HostData) hscan.getNext();
      }
      else if (hscan.key == addrs[i]) {
        hdata[hsize] = hscan;
        hlist[hsize] = hscan.host;
        if (hscan.version > hscan.agreed) {
          if (Debug.DEBUG)
            Debug.println(20, "hosts unstable: " + hscan.version + " "  + hscan.agreed);
          stable = false;
        }
        hscan.last = last;
        hscan = (HostData) hscan.getNext();
        i++;
        hsize++;
      }
      else {
        i++;
      }

    // Build member composition
    mscan = (MemberData) tmembers.getFirst();
    msize=0;
    while (mscan != null && msize < lids.length)
      if (mscan.key == lids[msize]) {
        mdata[msize] = mscan;
        mscan.last = last;
        mscan = (MemberData) mscan.getNext();
        msize++;
      }
      else if (mscan.key < lids[msize]) {
        if (!mscan.leaving)
          stable = false;
        mscan = (MemberData) mscan.getNext();
      }
      else {
        if (Debug.DEBUG)
          Util.panic("Member unknown");
      }

    return stable;
  }

}


