/*
 * Copyright (c) 1998-2000 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package relacs.general;

import java.io.*;

/**
 * The <code>ByteVectorOutputStream</code> class is an output stream especially designed
 * to be used with a transport layer which fragments messages. Messages are
 * stored in fragments, each fragment being composed by an header, a payload
 * and a trailer. The length of these fields may be configured when creating
 * a new <code>OutMessage</code>. Data written in the message are stored in
 * the payload field of the fragments.
 *
 * @author      Alberto Montresor
 * @since       Jgroup 0.8
 */
public final class      ByteVectorOutputStream
extends                 OutputStream
implements              Externalizable
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Initial number of fragments composing a message; can grow */
  private static final int NUM_BUFFER = 16;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  private int       payload;            // Payload size in bytes

  private int       position;           // Current position in the stream [0..byteCount]
  private int       fragment;           // Current fragment number
  private int       offset;             // Current offset in the fragment

  private int       fragmentCount;      // Total number of fragments used
  private int       byteCount;          // Total number of bytes written

  private byte[]    buffer;             // Current fragment
  private byte[][]  fragments;          // Byte array fragments



  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////


  /**
   * Construct a message with header and trailer of length zero. In this way, OutMessage may
   * be used as a more efficient version of ByteArrayOutputStream.
   */
  public ByteVectorOutputStream(int payload)
  {
    if (payload <= 0)
      throw new IndexOutOfBoundsException();

    // Store and compute fixed limits
    this.payload = payload;

    // Initialize variables regarding message size
    byteCount = 0;
    fragmentCount = 1;

    // Initialize variables regarding message position in the stream
    position = 0;
    fragment = 0;
    offset = 0;

    // Create fragments
    fragments = new byte[NUM_BUFFER][];
    fragments[0] = new byte[payload];
    buffer = fragments[0];
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods inherited from OutputStream
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Writes the specified byte to this byte array output stream.
   *
   * @param   b   the byte to be written.
   */
  public void write(int b)
  {
    /* Increase the number of stored bytes */
    position++;
    if (position > byteCount) {
      byteCount = position;
      if (offset == payload) {
        /*
         *  We have reached the end of the last fragment; we add a further fragment
         *  to store the new byte; fragment, offset and buffer are updated to point
         *  to the fist byte of the new fragment.
         */
        fragment++;
        offset = 0;
        enlargeSpace(byteCount);
        buffer = fragments[fragment];
      }
    }
    /* We write the byte and increment offset */
    buffer[offset++] = (byte) b;
  }


  /**
   * Writes <code>len</code> bytes from the specified byte array
   * starting at offset <code>off</code> to this msg output stream.
   * <p>
   * If <code>b</code> is <code>null</code>, a
   * <code>NullPointerException</code> is thrown.
   * <p>
   * If <code>off</code> is negative, or <code>len</code> is negative, or
   * <code>off+len</code> is greater than the length of the array
   * <code>b</code>, then an <tt>IndexOutOfBoundsException</tt> is thrown.
   *
   * @param   b     the data.
   * @param   off   the start offset in the data.
   * @param   len   the number of bytes to write.
   */
  public void write(byte b[], int off, int len)
  {
    if ((off<0) || (off>b.length) || (len<0) || ((off+len) > b.length)) {
      throw new IndexOutOfBoundsException();
    } else if (len == 0) {
      return;
    } else {
      position += len;
      if (position > byteCount)
        byteCount = position;
      if (offset + len <= payload) {
        System.arraycopy(b, off, buffer, offset, len);
        offset += len;
      } else {
        enlargeSpace(byteCount);
        int todo = payload - offset;
        System.arraycopy(b, off, fragments[fragment], offset, todo);
        int remaining = len - todo;
        off+=todo;
        offset = payload;
        while (remaining > payload) {
          System.arraycopy(b, off, fragments[++fragment], 0, payload);
          remaining -= payload;
          off += payload;
        }
        if (remaining > 0) {
          System.arraycopy(b, off, fragments[++fragment], 0, remaining);
          offset = remaining;
        }
        buffer = fragments[fragment];
      }
    }
  }

  /**
   * Writes <code>b.length</code> bytes from the specified byte array
   * to this output stream. The general contract for <code>write(b)</code>
   * is that it should have exactly the same effect as the call
   * <code>write(b, 0, b.length)</code>.
   * <p>
   * If the total number of bytes written in the message exceeds the maximum
   * length, a <tt>ArrayIndexOutOfBoundsException</tt> is thrown.
   *
   * @param      b   the data.
   *
   */
  public void write(byte b[])
  {
    write(b, 0, b.length);
  }


  /**
   *  Returns an array of bytes containing the contents of the unique fragment
   *  of the system, if the system is
   */
  public byte[] getByteArray()
  {
    if (fragmentCount == 1)
      return fragments[0];
    else {
      // Never used
      byte[] temp = new byte[byteCount];
      int off = 0;
      for (int i=0; i < fragmentCount-1; i++) {
        System.arraycopy(fragments[i], 0, temp, off, payload);
        off += payload;
      }
      System.arraycopy(fragments[fragmentCount-1], 0, temp, off, byteCount - off);
      return temp;
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Externalizable methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Used only by the externalization service; users cannot create a OutMessage
   *  to initialize the message.
   */
  public ByteVectorOutputStream()
  {
  }

  /**
   *
   */
  public void writeExternal(ObjectOutput out)
        throws IOException
  {
    out.writeInt(byteCount);

    for (int i=0; i < fragmentCount-1; i++)
      out.write(fragments[i], 0, payload);
    out.write(fragments[fragmentCount-1], 0, byteCount - (payload * (fragmentCount-1)) );
    // System.out.println("Fragments " + payload);
    // System.out.println("Last fragments " + (byteCount - (payload * (fragmentCount-1) )));
  }

  /**
   *
   */
  public void readExternal(ObjectInput in)
        throws IOException, ClassNotFoundException
  {
    // Initialize variables regarding message size
    fragmentCount = 1;
    byteCount = in.readInt();
    this.payload = byteCount;


    // Initialize variables regarding message position in the stream
    position = 0;
    fragment = 0;
    offset   = 0;

    // Create fragments
    fragments = new byte[1][];
    fragments[0] = new byte[payload];
    int remaining = payload;
    int read;
    while (remaining > 0) {
      read = in.read(fragments[0], payload-remaining, remaining);
      // System.out.println("Read " + read);
      if (read < 0)
        throw new IOException("Not enough bytes when unmarshaling OutMessage");
      remaining -= read;
    }
    buffer = fragments[0];
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Private methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Alloc new byte fragments to guarantee that there is enough space to store
   *  <code> newsize </code> bytes.
   */
  private void enlargeSpace(int newsize)
  {
    int nfragment = (newsize-1) / payload + 1;  // Total number of fragments to store <size> bytes
    if (nfragment > fragments.length) {
      // Double the fragment array if needed
      byte[][] tempbuffer = new byte[fragments.length*2][];
      System.arraycopy(fragments, 0, tempbuffer, 0, fragments.length);
      fragments = tempbuffer;
    }
    for (int i = fragmentCount; i < nfragment; i++)
      fragments[i] = new byte[payload];
    fragmentCount = nfragment;
  }

} // END ByteVectorOutputStream

