/*
 * Copyright (c) 1998-2000 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package relacs.gm;

import java.util.*;

import relacs.mss.Upcall;
import relacs.general.Queue;
import debug.Debug;


/**
 * The <code> DispatcherLayer</code> class handles
 *
 * @author  Alberto Montresor
 * @since   Jgroup 0.1
 */
public final class  DispatcherLayer
extends             Thread
implements          DispatcherService
{

  //////////////////////////////////////////////////////////////////////////////////////////
  // Private fields
  //////////////////////////////////////////////////////////////////////////////////////////

  private Queue                 upcalls;        // Upcalls queue
  private BaseLayer             layer;          // Associated layer
  private Hashtable             monitors;       // List of objects on which add has been performed
  private boolean               halted;         // True if the thread must halt


  //////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  //////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Initializes a DispatcherLayer
   */
  DispatcherLayer(Object[] configuration)
  {
    super();
    setPriority(MAX_PRIORITY);
    upcalls = new Queue();
    monitors = new Hashtable();
    halted = false;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static factory
  ////////////////////////////////////////////////////////////////////////////////////////////

  public static DispatcherLayer getLayer(Object[] configuration)
  {
    return new DispatcherLayer(configuration);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // setListener methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Add a listener
   */
  public void addListener(BaseLayer listener)
  {
    this.layer = listener;
  }


  //////////////////////////////////////////////////////////////////////////////////////////
  // Methods from DispatcherService
  //////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Receives and stores a message from daemon
   */
  public void insert(Upcall upcall)
  {
    /* Insert the upcall in the upcall queue; synchronize to guarantee consistency */
    synchronized (upcalls) {
      upcalls.insert(upcall);
    }
    
    /* Wake up all objects which are blocked waiting for some event and 
     * have been added to the list of objects 
     */
    for (Enumeration e = monitors.keys(); e.hasMoreElements(); ) {
      Object obj = e.nextElement();
      synchronized (obj) {
        obj.notify();
      }
    }

  }

  /**
   * Halts the thread
   */
  public void halt()
  {
    halted = true;
    layer = null;

    /* Wake up all objects which are blocked waiting for some event and 
     * have been added to the list of objects 
     */
    for (Enumeration e = monitors.keys(); e.hasMoreElements(); ) {
      Object obj = e.nextElement();
      synchronized (obj) {
        obj.notify();
      }
    }

  }

  public void dispatch(Object obj)
  {
    Upcall upcall;
    
    /** Get the first upcall in the queue */
    synchronized (upcalls) {
      upcall = (Upcall) upcalls.removeFirst();
    }
    if (upcall != null) {
      layer.dispatch(upcall);
      yield();
    } else {
      monitors.put(obj, obj);
      synchronized (obj) { 
        try { obj.wait(); } catch (Exception e) { }
      }
      monitors.remove(obj);
    }
  }


  //////////////////////////////////////////////////////////////////////////////////////////
  // Run method of the thread
  //////////////////////////////////////////////////////////////////////////////////////////


  /* RUN
   * Run method of the thread
   */
  public void run()
  {
    try {
      while (!halted)
        dispatch(this);
    } catch (Exception e) {
      // Should never be thrown
      Debug.println("Dispatcher Exception " + e);
      e.printStackTrace();
    }
  }

} // END DispatcherLayer



