/*
 * Copyright (c) 1998-2000 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package relacs.gm;

import java.net.InetAddress;
import java.io.*;
import java.util.*;

import jgroup.*;
import jgroup.multicast.*;
import jgroup.reference.*;

import relacs.general.*;
import relacs.*;
import relacs.mss.*;
import relacs.daemon.*;
import relacs.rmi.*;

import debug.Debug;

/**
 *
 * @author  Alberto Montresor
 * @since   0.7
 */
public class    MergingLayer
implements      MembershipService, MembershipListener, MulticastListener
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** */
  public static final String PROTOCOL_NAME = "relacs.SMS";

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  private MembershipService     membershipService;    // Underlying membership service
  private MulticastService      multicastService;     // Underlying multicast service
  private ArrayList             membershipArray;      // Dynamic array of membership listeners
  private ArrayList             mergingArray;         // Dynamic array of merging listeners
  private MembershipListener[]  membershipListeners;  // Layers listening for membership events
  private MergingListener[]     mergingListeners;     // Layers listening for merging events

  private MemberTable           table;                // Member table used by this layer
  private MemberTable           providedTable;        // Member table provided by this layer

  private MemberId              me;                   // The MemberId of this member
  private MemberId[]            members;              // The composition of the last view


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *
   * @exception DSException
   *            raised when the distributed system configuration object contains an error
   * @exception IOException
   *            if an I/O error occurs when creating communication sockets.
   */
  protected MergingLayer(Object[] configuration, MembershipService membershipService,
    MulticastService multicastService)
    throws JgroupException
  {
    this.membershipService = membershipService;
    this.multicastService = multicastService;

    membershipArray = new ArrayList();
    membershipListeners = new MembershipListener[0];

    mergingArray = new ArrayList();
    mergingListeners = new MergingListener[0];

    table = membershipService.getMemberTable();
    providedTable = new MemberTable();
    me = membershipService.getMyIdentifier();
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static factory
  ////////////////////////////////////////////////////////////////////////////////////////////

  public static MergingLayer getLayer(Object[] configuration,
      MembershipService membershipService,
      MulticastService multicastService)
    throws JgroupException
  {
    return new MergingLayer(configuration, membershipService, multicastService);
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods of ExtensibleMembershipService
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Add a listener
   */
  public void addListener(MembershipListener listener)
  {
    membershipArray.add(listener);
    membershipListeners = (MembershipListener[])
      membershipArray.toArray(new MembershipListener[membershipArray.size()]);
  }

  /**
   *  Add a listener
   */
  public void addListener(MergingListener listener)
  {
    mergingArray.add(listener);
    mergingListeners = (MergingListener[])
      mergingArray.toArray(new MergingListener[mergingArray.size()]);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods of MembershipService
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Implements <tt>void join(int gid)</tt> and stores
   * some data needed in the merging protocol.
   *
   * @exception JgroupException
   *            if the group manager receives a request to join a group
   *            when it is already member of a group.
   */
  public void join(int gid)
    throws JgroupException
  {
    membershipService.join(gid);
  }

  /**
   * Implements <tt>void join()</tt> and stores
   * some data needed in the merging protocol.
   *
   * @exception JgroupException
   *            if the group manager receives a request to join a group
   *            when it is already member of a group.
   */
  public synchronized void join()
    throws JgroupException
  {
    membershipService.join();
  }


  /**
   *
   *
   * @exception JgroupException   If the group manager is not member of any group, or is
   *            already leaving the current group.
   */
  public synchronized void leave()
    throws JgroupException
  {
    membershipService.leave();
  }


  /**
   *
   */
  public MemberId getMyIdentifier()
  {
    return me;
  }

  /**
   *
   */
  public MemberTable getMemberTable()
  {
    return providedTable;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from MembershipListener
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Upcall that is invoked by Jgroup when a view change occur in the
   *  group identified by <code> gid</code>.
   *
   *  @param group    user structure containing information associated to the group.
   *  @param view     the new view object.
   */
  public void viewChange(View view)
  {
    if (Debug.DEBUG)
      Debug.println(50, "MergingLayer: viewChange");
    // Update the member table associated to this group
    providedTable.viewChange(view);

    for (int j = 0; j < membershipListeners.length; j++)
      membershipListeners[j].viewChange(view);

    // Associate new (empty) information to new members
    // Associate new information also to members that have recovered
    members = view.getMembers();
    for (int i=0; i < members.length; i++) {
      int status = table.getState(members[i]);
      if (status == MemberTable.NEWMEMBER || status == MemberTable.RECOVERING)
        table.put(members[i], new MemberData(members[i], members[i].equals(me)));
    }

    // Given the new view, get the total list of members registered
    // in table
    Object[] objs = table.elements();

    MemberData[] mdata = new MemberData[objs.length];
    for (int i=0; i < objs.length; i++)
      mdata[i] = (MemberData) objs[i];


    // Update the mset and lset values, and compute if an update
    // message is necessary.
    boolean needupdate = false;
    for (int i=0; i < mdata.length; i++) {
      MemberData md = (MemberData) mdata[i];
      if (!table.isMember(mdata[i].id)) {
        mdata[i].mset = false;
        mdata[i].lset = false;
        mdata[i].tset = false;
      } else {
        mdata[i].tset = true;
        if (!mdata[i].mset && !mdata[i].id.equals(me))
          needupdate = true;
      }
      if (Debug.DEBUG) {
        Debug.print(50, "Table: ");
        Debug.print(50, mdata[i].id + " ");
        Debug.print(50, table.getState(mdata[i].id) + " ");
        Debug.println(50, mdata[i].mset + " " + mdata[i].lset);
      }
    }

    if (needupdate) {
      // Search the first member in mdata with mset equal to true
      int min = 0;
      for (int i=0; i < mdata.length; i++) {
        if (mdata[i].mset && table.getState(mdata[i].id) != MemberTable.CRASHED) {
          min = i;
          break;
        }
      }
      if (mdata[min].id.equals(me)) {
        try {
          OutputStream stream = multicastService.getMessage();
          MarshalOutputStream msg = new MarshalOutputStream(stream);
          int[] owners = getOwners(mdata);
          int[] dests = getDests(mdata);
          MemberId[] destIds = new MemberId[dests.length];
          for (int j = 0; j < dests.length; j++)
            destIds[j] = members[dests[j]];
          msg.writeUTF(PROTOCOL_NAME);
          msg.writeInt(dests.length);
          for (int j=0; j < dests.length; j++)
            msg.writeInt(dests[j]);
          msg.writeInt(owners.length);
          for (int j=0; j < owners.length; j++)
            msg.writeInt(owners[j]);
          try {
            for (int j=0; j < mergingListeners.length; j++)
              msg.writeObject(mergingListeners[j].getState(destIds));
          } catch (Exception e) {
            System.err.println("Exception caught when getting state from members");
            e.printStackTrace(System.err);
          }
          msg.flush();
          multicastService.mcast(stream, null);
        } catch(Exception e) {
          e.printStackTrace();
          Util.panic("Exception during view change events of MergingData");
        }
      }
    } // END if (needupdate)
  }


  /**
   *  Upcall that is invoked by Jgroup to deliver a message <code> msg</code>
   *  related to the group identified by <code> group</code>. In this case,
   *  the message is a sequence of bytes and must be interpreted by the
   *  application.
   *
   *  @param group    user structure containing information associated to the group.
   *  @param msg      the sequence of bytes containing the message.
   */
  public Object deliverStream(InputStream stream)
  {
    if (Debug.DEBUG)
      Debug.println(50, "MergingLayer: deliverStream");
    try {
      MarshalInputStream in = new MarshalInputStream(stream);
      if (!PROTOCOL_NAME.equals(in.readUTF()))
        throw new UnsupportedOperationException();

      MemberId[] dests = new MemberId[in.readInt()];
      for (int i=0; i < dests.length; i++)
        dests[i] = members[in.readInt()];

      MemberId[] owners = new MemberId[in.readInt()];
      for (int i=0; i < owners.length; i++)
        owners[i] = members[in.readInt()];

      // Questo e' da pensare; avere piu' merging listener?
      Object[] objs = new Object[mergingListeners.length];
      for (int j=0; j < mergingListeners.length; j++)
        objs[j] = in.readObject();

      for (int j=0; j<owners.length; j++) {

        // If I belong to <owners>, the <mset> contains all the members
        // in the current view.
        if (owners[j].equals(me)) {
          for (int k=0; k < members.length; k++) {
            MemberData md = (MemberData) table.get(members[k]);
            md.mset = true;
          }
        }

        // Each of the members in pset are contained in <lset>
        MemberData md = (MemberData) table.get(owners[j]);
        md.lset = true;
      }
      for (int j=0; j < mergingListeners.length; j++)
        mergingListeners[j].putState(objs[j], dests);

    } catch (Exception e) {
      if (Debug.DEBUG) {
        Debug.println(50, "MergingLayer: message not for me");
        Debug.printStackTrace(50, e);
      }
      throw new UnsupportedOperationException();
    }
    return null;
  }

  /**
   *  Upcall that is invoked by Jgroup to deliver a message <code> obj</code>
   *  related to the group identified by <code> group</code>. In this case,
   *  the message is an object marshalled and unmarshalled through java
   *  serialization.
   *
   *  @param userinfo   user structure containing information associated to the group.
   *  @param obj        the object containing the message.
   */
  public Object deliverObject(Object msg)
  {
    Debug.println(50, "MergingLayer: deliverObject");
    throw new UnsupportedOperationException();
  }

  /**
   *  Upcall that is invoked by Jgroup to acknowledge the fact that an object
   *  is not member of a group any more more:
   *
   *  @param userinfo   user structure containing information associated to the group.
   */
  public void leaved()
  {
    for (int i = 0; i < membershipListeners.length; i++)
      membershipListeners[i].leaved();
  }

  /**
   *  Currently not implemented.
   *
   *  @param gdesc    user structure containing information associated to the group.
   */
  public void prepareChange()
  {
    for (int i = 0; i < membershipListeners.length; i++)
      membershipListeners[i].prepareChange();
  }

  /**
   *
   */
  private int[] getOwners(MemberData[] mdata)
  {
    // Count owners
    int len = 0;
    for (int i=0; i < mdata.length; i++)
      if (mdata[i].lset) len++;

    // Write owners
    int[] owners = new int[len];
    len = 0;
    for (int i=0; i < mdata.length; i++)
      if (mdata[i].lset) {
        owners[len] = table.getIndex(mdata[i].id);
        len++;
      }

    return owners;
  }

  /**
   *
   */
  private int[] getDests(MemberData[] mdata)
  {
    // Count dests
    int len = 0;
    for (int i=0; i < mdata.length; i++)
      if (!mdata[i].mset && mdata[i].tset) len++;

    // Write dests
    int[] dests = new int[len];
    len = 0;
    for (int i=0; i < mdata.length; i++)
      if (!mdata[i].mset && mdata[i].tset) {
        dests[len] = table.getIndex(mdata[i].id);
        len++;
      }

    return dests;
  }

  class MemberData
  {

    boolean   lset;             // True if the member may potentially know LESS updates
    boolean   mset;             // True if the member may potentially know MORE updates
    boolean   tset;

    MemberId  id;

    MemberData(MemberId id, boolean set)
    {
      this.id = id;
      lset = set;
      mset = set;
      tset = set;
    }

  } // END MemberData


}
