/*
 * Copyright (c) 1998-2000 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package relacs.mss;

import relacs.general.Util;

/**
 * The <code> Header </code> class is used to encode and decode the header
 * of each Mss message.
 *
 * @author  Alberto Montresor
 * @since   Jgroup 0.1
 */
final class Header
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Size constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Size of pair (Host identifier, incarnation number) (in bytes). */
  static final int HOSTDATA_SIZE = Types.HID_SIZE + Types.INC_SIZE;

  /** Size of header in bytes */
  static final int SIZE = 2*HOSTDATA_SIZE + Types.MID_SIZE + Types.TAG_SIZE + Types.CHK_SIZE;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Position constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  private static final int SRC_POS  = 0;                            // Source host address
  private static final int SINC_POS = SRC_POS  + Types.HID_SIZE; // Source incarnation
  private static final int DEST_POS = SINC_POS + Types.INC_SIZE; // Dest. host address
  private static final int DINC_POS = DEST_POS + Types.HID_SIZE; // Dest incarnation
  private static final int MID_POS  = DINC_POS + Types.INC_SIZE; // Message identifier
  private static final int TAG_POS  = MID_POS  + Types.MID_SIZE; // Type of message
  private static final int CHK_POS  = TAG_POS  + Types.TAG_SIZE; // Header checksum


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  int           src;              // Source host indentifier
  int           sinc;             // Source incarnation number
  int           dest;             // Destination host identifier
  int           dinc;             // Destination incarnation number
  int           mid;              // Message identifier
  byte          tag;              // Type of message (ack, normal, routing)


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Allocates an empty header
   */
  Header()
  {
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Decodes an array of bytes into an header.
   * @param     buffer          buffer containing the header.
   */
  void read(byte[] buffer)
  {
    src   = Types.byte2hid(buffer, SRC_POS);
    sinc  = Types.byte2inc(buffer, SINC_POS);
    dest  = Types.byte2hid(buffer, DEST_POS);
    dinc  = Types.byte2inc(buffer, DINC_POS);
    mid   = Types.byte2mid(buffer, MID_POS);
    tag   = buffer[TAG_POS];
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Returns true if the checksum header is correct.
   */
  static boolean checkHeader(byte[] buffer)
  {
    return (Types.computeChecksum(buffer, 0, CHK_POS) == Types.byte2chk(buffer, CHK_POS));
  }

  /**
   * Encodes a header starting from raw data.
   *
   * @param     buffer          buffer where the header is written.
   * @param     tag             type of header.
   * @param     srcdata         source message.
   * @param     destdata        destination data.
   * @param     mid             message identifier.
   */
  static void write(byte[] buffer, byte tag, byte[] srcdata, byte[] destdata, int mid)
  {
    int i, j;           // Counters

    for (i=0, j=SRC_POS; i < HOSTDATA_SIZE; i++, j++)
      buffer[j] = srcdata[i];
    for (i=0, j=DEST_POS; i < HOSTDATA_SIZE; i++, j++)
      buffer[j] = destdata[i];
    buffer[TAG_POS] = tag;
    Types.mid2byte(buffer, MID_POS,  mid);
    Types.chk2byte(buffer, CHK_POS, Types.computeChecksum(buffer, 0, CHK_POS));
  }

  /**
   * Encodes the source-fixed section of a header starting from raw data.
   *
   * @param     buffer          buffer where the header is written.
   * @param     srcdata         source message.
   * @param     tag             type of header.
   */
  static void writeSrc(byte[] buffer, byte[] srcdata, byte tag)
  {
    int i, j;           // Counters

    for (i=0, j=SRC_POS; i < HOSTDATA_SIZE; i++, j++)
      buffer[j] = srcdata[i];
    buffer[TAG_POS] = tag;
  }

  /**
   * Encodes the destination section of a header starting from raw data.
   *
   * @param     buffer          buffer where the header is written.
   * @param     destdata        destination data.
   * @param     mid             message identifier.
   */
  static void writeDest(byte[] buffer, byte[] destdata, int mid)
  {
    int i, j;           // Counters

    for (i=0, j=DEST_POS; i < HOSTDATA_SIZE; i++, j++)
      buffer[j] = destdata[i];
    Types.mid2byte(buffer, MID_POS,  mid);
    Types.chk2byte(buffer, CHK_POS, Types.computeChecksum(buffer, 0, CHK_POS));
  }

  /**
   * Encodes the pair (Host identifier, incarnation number).
   *
   * @param     buffer          buffer where the header is written.
   * @param     hid             host identifier.
   * @param     inc             incarnation number.
   */
  static void writeHostData(byte[] buffer, int hid, int inc)
  {
    Types.hid2byte(buffer, 0,  hid);
    Types.inc2byte(buffer, Types.HID_SIZE, inc);
  }

} // END Header
