/*
 * Copyright (c) 1998-2000 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package relacs.mss;

import java.net.*;
import java.io.*;
import java.util.*;

import jgroup.DSException;
import relacs.RelacsDS;
import relacs.general.HashTable;

import debug.Debug;

/**
 * The <code> HostTable</code> class
 *
 * @author  Alberto Montresor
 * @since   Jgroup 0.1
 */
public final class      HostTable
extends                 HashTable
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  private ArrayList     array;          // Dynamic array containing the hosts
  private MssHost[]     hosts;          // Host array
  private MssHost[]     htable;         // Hash table of the hosts
  private int           size;           // Number of elements inserted
  private MssHost       me;             // This host

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  public HostTable(MssHost[] list)
        throws DSException
  {
    array = new ArrayList(list.length+16);
    htable = new MssHost[HASHTABLE_SIZE];
    size   = list.length;
    for (int i=0; i< size; i++)
      insertHost(list[i]);
    hosts = new MssHost[list.length];
    array.toArray(hosts);
    for (int i=0; i< size; i++) {
      hosts[i].index = i;
      hosts[i].tnext = (i+1 == size ? null : hosts[i+1]);
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Public methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Inserts host <code>host</code> in the table, unless an host
   * with the same 32-bit address is already in the table.
   *
   * @param      host     the host to be inserted.
   * @exception  DSException   if an host with the same address is already in the table, or
   *                      the address is unknown.
   */
  public int[] insert(MssHost host)
        throws DSException
  {
    int[]       ret;

    size++;
    insertHost(host);
    hosts = new MssHost[size];
    ret = new int[size];
    array.toArray(hosts);
    for (int i=0; i < size; i++) {
      ret[i] = hosts[i].index;
      hosts[i].index = i;
    }
    return ret;
  }


  /**
   * Returns the host reference to which this address is mapped.
   *
   * @param   address   32-bit address.
   * @return            the host reference to which this address is mapped.
   */
  public MssHost lookup(int address)
  {
    MssHost        host;                           // MssHost reference

    host = htable[computeHash(address)];
    while (host != null && host.address != address)
      host = host.hnext;
    return host;
  }

  /**
   * Returns the first host in the host table.
   *
   * @return            the first host in the host table.
   */
  public MssHost getFirst()
  {
    return hosts[0];
  }

  public int size()
  {
    return size;
  }

  public MssHost getMe()
  {
    return me;
  }

  public InetAddress[] getHostAddressArray()
  {
    InetAddress[]       ret;
    MssHost             scan;
    int                 j;

    ret = new InetAddress[size];
    for (int i=0; i < size; i++)
      ret[i] = hosts[i].inetaddr;
    return ret;
  }

  public void printRoutingTable()
  {
    if (Debug.DEBUG) {
      MssHost scan;

      for (int i=0; i < size; i++)
        Debug.println(10, "Host: " + hosts[i].name + " L= " + hosts[i].lambda +
                  " D= " + hosts[i].delta + " route= " + hosts[i].route.name);
    }
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Private methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  private void insertHost(MssHost host)
        throws DSException
  {
    int         index;          // Bucket index in host table
    int         i = 0;          // Counter

    try {
      if (Debug.DEBUG)
        Debug.println(10, "Host inserted: " + host.name);
      while (i<array.size() && ((MssHost) array.get(i)).address < host.address)
        i++;
      if (i < array.size() && ((MssHost) array.get(i)).address == host.address )
        throw new DSException("Duplicated host: " + host.name);
      array.add(i, host);

      // Update hash table
      index = computeHash(host.address);
      host.hnext = htable[index];
      htable[index] = host;

      // Check if the inserted host is the local host
      if (host.inetaddr.equals(RelacsDS.getLocalHost()))
        me = host;
    } catch (UnknownHostException e) {
      throw new DSException(e.getMessage());
    }
  }


} // END HostTable


