/*
 * Copyright (c) 1998-2000 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package relacs.mss;

import java.io.*;

import debug.Debug;

/**
 *
 * @author      Alberto Montresor
 * @since       Jgroup 0.8
 */
public final class      InMessage
extends                 InputStream
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Initial number of fragments composing a message; can grow */
  private static final int NUM_BUFFER = 16;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static section
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Temporary buffer used in writeTemp */
  private static byte[] temp = new byte[8];


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  private int       payload;            // Payload size in bytes
  private int       header;             // Header size in bytes
  private int       trailer;            // Trailer size in bytes
  private int       totlen;             // Total length of a fragment
  private int       endpayload;         // Last byte used for payload

  private int       position;           // Current position in the stream [0..byteCount]
  private int       fragment;           // Current fragment number
  private int       index;              // Current index in the fragment

  private int       fragmentCount;      // Total number of fragments used
  private int       byteCount;          // Total number of bytes written

  private byte[]    buffer;             // Current buffer
  private byte[][]  fragments;          // Byte array fragments

  private int       marked;             // Mark position


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////


  /**
   *  This constructor builds an empty message subdivided in fragments.
   *  Each fragment has a length of <code> header + payload + trailer</code>.
   *  <p>
   *  A <tt>IndexOutOfBoundsException</tt> is thrown if: <tt>payload</tt> is
   *  negative or equal to zero; <tt>header</tt> is negative; <tt>trailer</tt>
   *  is negative; <tt>size</tt> is negative or equal to zero.
   */
  InMessage(int payload, int header, int trailer)
  {
    if (payload <= 0 || header < 0 || trailer < 0)
      throw new IndexOutOfBoundsException();

    // Store and compute fixed limits
    this.payload = payload;
    this.header  = header;
    this.trailer = trailer;
    endpayload = payload + header;
    totlen = endpayload + trailer;

    // Initialize variables regarding message size
    byteCount = 0;
    fragmentCount = 0;

    // Initialize variables regarding message position in the stream
    position = 0;
    fragment = 0;
    index = header;

    // Create fragments
    fragments = new byte[NUM_BUFFER][];
  }


  /**
   *  This constructor builds a message input stream subdivided in
   *  fragments with length <code> header + payload + trailer</code>.
   *  <p>
   *  A <tt>IndexOutOfBoundsException</tt> is thrown if: <tt>payload</tt> is
   *  negative or equal to zero; <tt>header</tt> is negative; <tt>trailer</tt>
   *  is negative; <tt>size</tt> is negative or equal to zero.
   */
  InMessage(int payload, int header, int trailer, int size)
  {
    if (payload <= 0 || header < 0 || trailer < 0 || size <= 0)
      throw new IndexOutOfBoundsException();

    // Store and compute fixed limits
    this.payload = payload;
    this.header  = header;
    this.trailer = trailer;
    endpayload = payload + header;
    totlen = endpayload + trailer;

    // Initialize variables regarding message size
    byteCount = 0;
    fragmentCount = 0;

    // Initialize variables regarding message position in the stream
    position = 0;
    fragment = 0;
    index = header;

    // Create fragments
    fragments = new byte[size % payload + 1][];
  }

  /**
   * Construct a InMessage starting from an OutMessage. In this way, it possible
   * to local deliver messages without copying them.
   */
  public InMessage(OutMessage msg)
  {
    // Store and compute fixed limits
    this.payload = msg.getPayload();
    this.header  = msg.getHeader();
    this.trailer = msg.getTrailer();
    endpayload = payload + header;
    totlen = endpayload + trailer;

    // Initialize variables regarding message size
    byteCount = msg.getByteCount();
    fragmentCount = msg.getFragmentCount();

    // Initialize variables regarding message position in the stream
    position = 0;
    fragment = 0;
    index = header;

    // Create fragments
    fragments = msg.getFragments();
    buffer = fragments[0];
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods inherited from InputStream
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Reads the next byte of data from the input stream. The value byte is
   * returned as an <code>int</code> in the range <code>0</code> to
   * <code>255</code>. If no byte is available because the end of the stream
   * has been reached, the value <code>-1</code> is returned. This method
   * blocks until input data is available, the end of the stream is detected,
   * or an exception is thrown.
   *
   * @return     the next byte of data, or <code>-1</code> if the end of the
   *             stream is reached.
   * @exception  IOException  if an I/O error occurs.
   */
  public int read()
  {
    if (position == byteCount) {
      /*
       *  The end of the stream has been reached
       */
      return -1;
    }

    if (index == endpayload) {
      /*
       *  The end of the current fragment has been reached; we move to the next
       *  fragment and we reset the index.
       */
      fragment++;
      buffer = fragments[fragment];
      index = header;
    }
    position++;
    return buffer[index++] & 0xFF;
  }


  /**
   * Reads some number of bytes from the input stream and stores them into
   * the buffer array <code>b</code>. The number of bytes actually read is
   * returned as an integer.  This method blocks until input data is
   * available, end of file is detected, or an exception is thrown.
   *
   * <p> If <code>b</code> is <code>null</code>, a
   * <code>NullPointerException</code> is thrown.  If the length of
   * <code>b</code> is zero, then no bytes are read and <code>0</code> is
   * returned; otherwise, there is an attempt to read at least one byte. If
   * no byte is available because the stream is at end of file, the value
   * <code>-1</code> is returned; otherwise, at least one byte is read and
   * stored into <code>b</code>.
   *
   * <p> The first byte read is stored into element <code>b[0]</code>, the
   * next one into <code>b[1]</code>, and so on. The number of bytes read is,
   * at most, equal to the length of <code>b</code>. Let <i>k</i> be the
   * number of bytes actually read; these bytes will be stored in elements
   * <code>b[0]</code> through <code>b[</code><i>k</i><code>-1]</code>,
   * leaving elements <code>b[</code><i>k</i><code>]</code> through
   * <code>b[b.length-1]</code> unaffected.
   *
   * <p> The <code>read(b)</code> method for class <code>InputStream</code>
   * has the same effect as: <pre><code> read(b, 0, b.length) </code></pre>
   *
   * @param      b   the buffer into which the data is read.
   * @return     the total number of bytes read into the buffer, or
   *             <code>-1</code> is there is no more data because the end of
   *             the stream has been reached.
   */
  public int read(byte b[])
  {
    return read(b, 0, b.length);
  }

  /**
   * Reads up to <code>len</code> bytes of data from the input stream into
   * an array of bytes.  An attempt is made to read as many as
   * <code>len</code> bytes, but a smaller number may be read, possibly
   * zero. The number of bytes actually read is returned as an integer.
   *
   * <p> If <code>b</code> is <code>null</code>, a
   * <code>NullPointerException</code> is thrown.
   *
   * <p> If <code>off</code> is negative, or <code>len</code> is negative, or
   * <code>off+len</code> is greater than the length of the array
   * <code>b</code>, then an <code>IndexOutOfBoundsException</code> is
   * thrown.
   *
   * <p> If <code>len</code> is zero, then no bytes are read and
   * <code>0</code> is returned; otherwise, there is an attempt to read at
   * least one byte. If no byte is available because the stream is at end of
   * file, the value <code>-1</code> is returned; otherwise, at least one
   * byte is read and stored into <code>b</code>.
   *
   * <p> The first byte read is stored into element <code>b[off]</code>, the
   * next one into <code>b[off+1]</code>, and so on. The number of bytes read
   * is, at most, equal to <code>len</code>. Let <i>k</i> be the number of
   * bytes actually read; these bytes will be stored in elements
   * <code>b[off]</code> through <code>b[off+</code><i>k</i><code>-1]</code>,
   * leaving elements <code>b[off+</code><i>k</i><code>]</code> through
   * <code>b[off+len-1]</code> unaffected.
   *
   * <p> In every case, elements <code>b[0]</code> through
   * <code>b[off]</code> and elements <code>b[off+len]</code> through
   * <code>b[b.length-1]</code> are unaffected.
   *
   * @param      b     the buffer into which the data is read.
   * @param      off   the start offset in array <code>b</code>
   *                   at which the data is written.
   * @param      len   the maximum number of bytes to read.
   * @return     the total number of bytes read into the buffer, or
   *             <code>-1</code> if there is no more data because the end of
   *             the stream has been reached.
   */
  public int read(byte b[], int off, int len)
  {
    int   tobecopied;
    int   remained;

    if (b == null)
      throw new NullPointerException();
    if (off < 0 || off > b.length || len < 0 || (off + len > b.length) || (off + len < 0))
      throw new IndexOutOfBoundsException();
    if (len == 0)
      return 0;

    if (position + len > byteCount)
      len = byteCount - position;
    if (b.length < len)
      len = b.length;
    position += len;
    remained = len;
    tobecopied = endpayload - index;
    while (remained > tobecopied) {
      System.arraycopy(fragments[fragment], index, b, off, tobecopied);
      off += tobecopied;
      remained -= tobecopied;
      fragment++;
      index = header;
      tobecopied = payload;
    }
    buffer = fragments[fragment];
    System.arraycopy(buffer, index, b, off, remained);
    index += remained;

    return len;
  }

  /**
   * Skips over and discards <code>n</code> bytes of data from this input
   * stream. The <code>skip</code> method may, for a variety of reasons, end
   * up skipping over some smaller number of bytes, possibly <code>0</code>.
   * This may result from any of a number of conditions; reaching end of file
   * before <code>n</code> bytes have been skipped is only one possibility.
   * The actual number of bytes skipped is returned.  If <code>n</code> is
   * negative, no bytes are skipped.
   *
   * <p> The <code>skip</code> method of <code>InputStream</code> creates a
   * byte array and then repeatedly reads into it until <code>n</code> bytes
   * have been read or the end of the stream has been reached. Subclasses are
   * encouraged to provide a more efficient implementation of this method.
   *
   * @param      n   the number of bytes to be skipped.
   * @return     the actual number of bytes skipped.
   * @exception  IOException  if an I/O error occurs.
   */
  public long skip(long n)
  {
    long value = n;

    position += n;
    if (position > byteCount) {
      value -= (position - byteCount);
      position = byteCount;
    }
    return value;
  }

  /**
   * Marks the current position in this input stream. A subsequent call to
   * the <code>reset</code> method repositions this stream at the last marked
   * position so that subsequent reads re-read the same bytes.
   *
   * <p> The <code>readlimit</code> arguments tells this input stream to
   * allow that many bytes to be read before the mark position gets
   * invalidated.
   *
   * <p> The general contract of <code>mark</code> is that, if the method
   * <code>markSupported</code> returns <code>true</code>, the stream somehow
   * remembers all the bytes read after the call to <code>mark</code> and
   * stands ready to supply those same bytes again if and whenever the method
   * <code>reset</code> is called.  However, the stream is not required to
   * remember any data at all if more than <code>readlimit</code> bytes are
   * read from the stream before <code>reset</code> is called.
   *
   *
   * @param   readlimit   not used
   */
  public synchronized void mark(int readlimit)
  {
    marked = position;
  }


  /**
   * Repositions this stream to the position at the time the
   * <code>mark</code> method was last called on this input stream.
   *
   * <p> The general contract of <code>reset</code> is:
   *
   * <p><ul>
   *
   * <li> If the method <code>markSupported</code> returns
   * <code>true</code>, then:
   *
   *     <ul><li> If the method <code>mark</code> has not been called since
   *     the stream was created, or the number of bytes read from the stream
   *     since <code>mark</code> was last called is larger than the argument
   *     to <code>mark</code> at that last call, then an
   *     <code>IOException</code> might be thrown.
   *
   *     <li> If such an <code>IOException</code> is not thrown, then the
   *     stream is reset to a state such that all the bytes read since the
   *     most recent call to <code>mark</code> (or since the start of the
   *     file, if <code>mark</code> has not been called) will be resupplied
   *     to subsequent callers of the <code>read</code> method, followed by
   *     any bytes that otherwise would have been the next input data as of
   *     the time of the call to <code>reset</code>. </ul>
   *
   * <li> If the method <code>markSupported</code> returns
   * <code>false</code>, then:
   *
   *     <ul><li> The call to <code>reset</code> may throw an
   *     <code>IOException</code>.
   *
   *     <li> If an <code>IOException</code> is not thrown, then the stream
   *     is reset to a fixed state that depends on the particular type of the
   *     input stream and how it was created. The bytes that will be supplied
   *     to subsequent callers of the <code>read</code> method depend on the
   *     particular type of the input stream. </ul></ul>
   *
   * <p> The method <code>reset</code> for class <code>InputStream</code>
   * does nothing and always throws an <code>IOException</code>.
   *
   * @exception  IOException  if this stream has not been marked or if the
   *               mark has been invalidated.
   */
  public synchronized void reset()
    throws IOException
  {
    fragment = marked / payload;
    index = header + marked % payload;
    position = marked;
    buffer = fragments[fragment];
  }

  /**
   * Tests if this input stream supports the <code>mark</code> and
   * <code>reset</code> methods. The <code>markSupported</code> method of
   * <code>InputStream</code> returns <code>false</code>.
   *
   * @return  <code>true</code> if this true type supports the mark and reset
   *          method; <code>false</code> otherwise.
   */
  public boolean markSupported()
  {
    return true;
  }



  ////////////////////////////////////////////////////////////////////////////////////////////
  // Public methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Move the current position in the stream to <code>position</code>.
   * <p>
   * If the current position in the message exceeds the
   * length, a <tt>ArrayIndexOutOfBoundsException</tt> is thrown.
   */
  public void seek(int newpos)
  {
    if (newpos > byteCount) {
      newpos = byteCount;
      throw new IndexOutOfBoundsException();
    }
    fragment = newpos / payload;
    index = header + newpos % payload;
    position = newpos;
    buffer = fragments[fragment];
  }

  /**
   * Read a short integer from the stream at the current index.
   * <p>
   * If the message ends before having the possibility to read the
   * value, a <tt>ArrayIndexOutOfBoundsException</tt> is thrown.
   */
  public int read2()
  {
    position += 2;
    if (position > byteCount) {
      position = byteCount;
      throw new IndexOutOfBoundsException();
    }
    if (index + 2 > endpayload) {
      readTemp(2);
      return ((temp[0] << 8) & 0xFF00) | (temp[1] & 0xFF);
    } else {
      return ((buffer[index++] << 8) & 0xFF00) | (buffer[index++] & 0xFF);
    }
  }


  /**
   * Read an integer from the stream at the current index.
   * <p>
   * If the message ends before having the possibility to read the
   * value, a <tt>ArrayIndexOutOfBoundsException</tt> is thrown.
   */
  public int read4()
  {
    position += 4;
    if (position > byteCount) {
      position = byteCount;
      throw new IndexOutOfBoundsException();
    }
    if (index + 4 > endpayload) {
      readTemp(4);
      return ((temp[0] << 24) & 0xFF000000) | ((temp[1] << 16)
                & 0xFF0000) | ((temp[2] << 8) & 0xFF00) | (temp[3] & 0xFF);
    } else {
      return ((buffer[index++] << 24) & 0xFF000000) | ((buffer[index++] << 16)
                & 0xFF0000) | ((buffer[index++] << 8) & 0xFF00) | (buffer[index++] & 0xFF);
    }
  }


  /**
   * Read a long integer from the stream at the current index.
   * <p>
   * If the message ends before having the possibility to read the
   * value, a <tt>ArrayIndexOutOfBoundsException</tt> is thrown.
   */
  public long read8()
  {
    int v1, v2;

    position += 8;
    if (position > byteCount) {
      position = byteCount;
      throw new IndexOutOfBoundsException();
    }
    if (index + 8 > endpayload) {
      readTemp(8);
      v1 = ((temp[0] << 24) & 0xFF000000) | ((temp[1] << 16) & 0xFF0000) |
           ((temp[2] << 8) & 0xFF00) | (temp[3] & 0xFF);
      v2 = ((temp[4] << 24) & 0xFF000000) | ((temp[5] << 16) & 0xFF0000) |
           ((temp[6] << 8) & 0xFF00) | (temp[7] & 0xFF);
    } else {
      v1 = ((buffer[index++] << 24) & 0xFF000000) | ((buffer[index++] << 16)
           & 0xFF0000) | ((buffer[index++] << 8) & 0xFF00) | (buffer[index++] & 0xFF);
      v2 = ((buffer[index++] << 24) & 0xFF000000) | ((buffer[index++] << 16)
           & 0xFF0000) | ((buffer[index++] << 8) & 0xFF00) | (buffer[index++] & 0xFF);
    }
    return ((long)(v1) << 32) | (v2 & 0xFFFFFFFFL);
  }

  /**
   *  Compares the bytes contained in this InMessage with those
   *  contained in another InMessage. Starting from position
   *  <t>pos1</t> in this message and from position <t>pos2</t>
   *  from the other message, compares <t>len</t> bytes and
   *  returns false if they differ.
   *
   *  @param pos1 starting position in this message
   *  @param msg the meessage to be compared
   *  @param pos2 starting position in <t>msg</t>
   *  @param len the number of bytes to be compared
   */
  public boolean compare(int pos1, InMessage msg, int pos2, int len)
  {
    /* 
     *  Verify that comparison arguments are regular
		 *  FIX 16/11/2000: Fixed the check by removing unnecessary (and wrong)
		 *  comparison between pos2+len and byteCount
     */
    if (pos1 >= byteCount || pos2 >= byteCount)
      throw new IllegalArgumentException("Pos1: " + pos1 + "Pos2: " + pos2 + " byteCount: " + byteCount + " len " + len);

    /*
     *  Compute start positions for comparison
     */
    int frag1 = pos1 / payload;
    int frag2 = pos2 / msg.payload;
    int index1 = header + pos1 % payload;
    int index2 = msg.header + pos2 % payload;


    for (int i=0; i < len; i++) {

      /*
       *  Stop comparison if bytes in byte arrays differ
       */
      if (fragments[frag1][index1] != msg.fragments[frag2][index2]) 
        return false;

      /*
       *  Go to the next byte, moving to the next fragment if necessary
       */
      index1++;
      index2++;
      if (index1 == endpayload) {
        index1 = header;
        frag1++;
      }
      if (index2 == msg.endpayload) {
        index1 = header;
        frag1++;
      }
      
    }

    /*
     *  The byte arrays are equal.
     */
    return true;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Package methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Insert a new fragment in the message whose payload field is
   * <tt>blen</tt> bytes long.
   */
  void insert(byte[] data, int blen)
  {
    byteCount += blen;
    if (fragmentCount == fragments.length) {
      // Double the space
      byte[][] btemp = new byte[fragments.length*2][];
      System.arraycopy(fragments, 0, btemp, 0, fragments.length);
      fragments = btemp;
    }
    fragments[fragmentCount] = data;
    if (fragmentCount == 0)
      buffer = fragments[0];
    fragmentCount++;
  }


  /**
   *  Returns the total number of bytes stored in this message output stream.
   */
  public int getByteCount()
  {
    return byteCount;
  }

  public int getPosition()
  {
    return position;
  }


  /**
   *  Returns the number of fragments composing the message.
   */
  int getFragmentCount()
  {
    return fragmentCount;
  }


  byte[][] getFragments()
  {
    return fragments;
  }


  /**
   *
   */
  int getPayload()
  {
    return payload;
  }

  /**
   *
   */
  int getHeader()
  {
    return header;
  }

  /**
   *
   */
  int getTrailer()
  {
    return trailer;
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Private methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Copy <tt>size</tt> bytes in the temp array; this function is invoked when
   *  the bytes read are part in fragment and part in the following one.
   */
  private void readTemp(int size)
  {
    for (int i = 0; i < size; i++) {
      if (index == endpayload) {
        fragment++;
        buffer = fragments[fragment];
        index = header;
      }
      temp[i] = buffer[index++];
    }
  }

  public void print(int level)
  {
    if (Debug.DEBUG) {
      Debug.print(level, "InMessage: ");
      for (int i=0; i < fragmentCount-1; i++)
        for (int j=header; j < endpayload; j++)
          Debug.print(level, " " + fragments[i][j]);
      for (int j=header; j < byteCount % payload; j++)
        Debug.print(level, " " + fragments[fragmentCount-1][j]);
      Debug.println(level, "");
    }
  }

} // END InMessage

