/*
 * Copyright (c) 1998-2000 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package relacs.mss;

import java.net.*;
import java.io.*;

import jgroup.DSException;
import relacs.RelacsConfig;
import relacs.general.*;
import debug.Debug;


/**
 * The <code>Mss</code> class implements a Multi-send Service
 *
 * @author  Salvatore Cammarrata
 * @author  Alberto Montresor
 * @since   Jgroup 0.5
 */

public class Mss
implements EhandlerUser, MssConstants, MssTag
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  private MssUser       mssuser;                // Upper layer
  private HostTable     hosttable;              // Hosttable
  private Ehandler      ehandler;               // Event handler
  private MssHost       me;                     // This host
  private int           myaddress;              // This host 32-bit address
  private Header        header;                 // Header (allocated only once)
  private int           rtsent;                 // Id. of the last routing table sent

  // Configuration
  private int           payload;                // Payload size
  private int           lambda_warning;         // N. of rounds before host warning
  private int           lambda_suspect;         // N. of rounds before host suspect
  private int           alfan;                  // alfa = alfan / alfad
  private int           alfad;                  // alfa = alfan / alfad
  private int           routingtimeout;         // time between two routing table send
  private MsgFactory    mfactory;               // Message factory to obtain OutMessage


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  public Mss(String name, MssUser mssuser, HostTable hosttable, RelacsConfig conf, int port)
        throws DSException, IOException
  {
    this.mssuser   = mssuser;
    this.hosttable = hosttable;
    this.me        = hosttable.getMe();
    this.myaddress = me.address;
    this.header    = new Header();

    // Configuration initialization
    this.payload        = conf.getPayload();
    this.lambda_suspect = conf.getLambda();
    this.lambda_warning = conf.getLambda()/2;
    this.alfan          = conf.getAlfan();
    this.alfad          = conf.getAlfad();
    this.routingtimeout = conf.getRtimeout();
    this.mfactory       = new MsgFactory(payload, HEADER_SIZE, TRAILER_SIZE);

    this.ehandler  = new Ehandler(name, this, port, conf.getPayload() + OVERHEAD_SIZE);
    rtTimeout(false);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Implementation of abstract methods inherithed from EhandlerUser
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Handles a message received by a remote host
   */
  public void rreceive(InetAddress fwd, byte[] buffer, int blen)
  {
    MssHost             dest;           // MssHost reference
    MssHost             src;            // MssHost reference

    if (Debug.DEBUG) {
      Debug.print(10, "HEADER: ");
      for (int k=0; k<21; k++)
        Debug.print(10, " " + buffer[k]);
      Debug.println(10, "");
    }
    header.read(buffer);
    if (Debug.DEBUG)
      Debug.print((header.tag==MSS_NACK || header.tag==MSS_RT ? 5 :10), "Recv| Tag:" + header.tag); //sax
    dest = hosttable.lookup(header.dest);
    if (dest != null && header.dest != myaddress) {
      if (Debug.DEBUG)
        Debug.println((header.tag==MSS_NACK || header.tag==MSS_RT ? 5 :10), "Forward:" + header.dest); //sax
      ehandler.rsend(dest.iroute, buffer, blen);                // Route message
    } else {
      if (Debug.DEBUG)
        Debug.println(10, "--------> Header.src " + header.src);
      src  = hosttable.lookup(header.src);
      if (src != null) {
        if (Debug.DEBUG) {
          Debug.print((header.tag == MSS_NACK || header.tag==MSS_RT ? 5 :10), " Src: " + src.name);
          Debug.print((header.tag == MSS_NACK || header.tag==MSS_RT ? 5 :10), " Mid: " + header.mid);
          Debug.println((header.tag == MSS_NACK || header.tag==MSS_RT ? 5 :10), " Leng: " + blen);
        }
        if (header.sinc >= src.incarnation) {
          if (header.sinc > src.incarnation) {
            if (Debug.DEBUG)
              Debug.println(10, "rreceive: handling new incarnation");
            handleNewIncarnation(src, header.sinc);
          }
          if (Debug.DEBUG) {
            if (header.dinc == me.incarnation)
              Debug.print((header.tag == MSS_NACK || header.tag==MSS_RT ? 5 :10)," >"); //sax
            Debug.println((header.tag == MSS_NACK || header.tag==MSS_RT ? 5 :10),"");  //sax
          }
          switch (header.tag) {
            case MSS_NACK:                                  //sax
              receiveNack(src, header);
              break;
            case MSS_RT:
              if (header.mid > src.rtrcvd) {
                src.rtrcvd = header.mid;
                rtUpdate(src, new RoutingData(buffer));
                swUpdate(src,buffer);   // sax
              }
              break;
            case MSS_SYNCH:
              if (header.mid > src.mrcvd) {
                src.mrcvd = header.mid;
                receiveSynch(src, header.mid);
              }
              break;
            default:
              msgReceive(src, header, buffer, blen);
              break;
          }
        }
      }
    }
  }

  /**
   *  Handles a message received by a local member
   */
  public void lreceive(Upcall upcall)
  {
    mssuser.lreceive(upcall);
  }


  /**
   *  Handles a timeout expiring
   */
  public void treceive (EventList event)
  {
    if (event.tag == RT_TIMEOUT)
      rtTimeout(false);
    else if (event.tag == NACK_TIMEOUT) //sax
      nackTimeout(event);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Public methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Reliable m-sends message <msg> to the destination array <dests>.
   * ??? Attualmente non e' reliable
   */
  public void rmsend(byte tag, OutMessage msg, MssHost[] dests)
  {
    for (int i=0; i<dests.length; i++)
      //      if (dests[i] != me)
        completeSend(tag, msg, dests[i]);
  }

  /**
   * M-sends message <msg> to the destination array <dests>.
   */
  public void msend(byte tag, OutMessage msg, MssHost[] dests) //sax
  {
        String s;

    for (int i=0; i<dests.length; i++) {
      s = dests[i].getName();
      if (s.compareTo(me.getName()) != 0)
        completeSend(tag, msg, dests[i]);
    }
  }

  public void msend(byte tag, OutMessage msg, MssHost[] dests, int length)
  {
    for (int i=0; i < length; i++)
      if (dests[i] != me)
        completeSend(tag, msg, dests[i]);
  }

  /**
   * M-sends message <msg> to host <dest>
   */
  public void msend(byte tag, OutMessage msg, MssHost dest)
  {
    completeSend(tag, msg, dest);
  }

  /**
   * M-sends message <msg> to all the hosts in the hosttable
   */
  public void msend(byte tag, OutMessage msg)
  {
    MssHost             scan;           // Used to scan hosttable

    scan = hosttable.getFirst();
    while (scan != null) {
      if (scan != me)
        completeSend(tag, msg, scan);
      scan = scan.getNext();
    }
  }

  public void lsend(Object msg)
  {
    if (Debug.DEBUG)
      Debug.println(10, "mss.lsend");
    ehandler.lnotify(msg);
  }

  /**
   *
   */
  public MsgFactory getMsgFactory()
  {
    return mfactory;
  }


  /////////////////////////////////////////////////////////////////////////////
  // Flow CTRL infos handling  //sax
  /////////////////////////////////////////////////////////////////////////////

  /**
   * Updates the flow control infos about <src>
   */

  private void swUpdate(MssHost src, byte[] buff) //sax
  {
    MssMessage msg;

    if (Debug.DEBUG)
      Debug.println(10,"swUpdate from "+src.name);
    src.sentSW.byte2win(buff);
    if (Debug.DEBUG)
      src.sentSW.printSW(10);

    // look if there are frozen msgs that can be sent
    msg = (MssMessage) src.frozen.getFirst();
    while (msg != null && msg.key <= (src.sentSW.leftedge + SWindow.BIT_SIZE)) {
      src.frozen.removeFirst();
      msg.start = System.currentTimeMillis();
      if (msg.dest != me)
        msg.dest.sent.insert(msg);
      if (Debug.DEBUG)
        Debug.println(10,"swUpdate : sending frozen msg "+msg.key+" to "+src.name);
      Header.writeDest(msg.buffer, msg.dest.header, msg.key);
      ehandler.rsend(msg.dest.iroute, msg.buffer, msg.blen);
      msg = (MssMessage) src.frozen.getFirst();
    }

    // remove the msgs that have been received from the sent queue
    msg = (MssMessage) src.sent.getFirst();
    while (msg != null && msg.key <= src.sentSW.leftedge) {
      if (Debug.DEBUG)
        Debug.println(10,"swUpdate : Removing "+msg.key+" msg from "+src.name+" sent msg queue");
      src.sent.removeFirst();
      msg = (MssMessage) src.sent.getFirst();
    }
    if (Debug.DEBUG) {
      for (int i=0; i < SWindow.BIT_SIZE; i++)
        if (src.sentSW.lookup(i))
          if ((msg = (MssMessage) src.sent.remove(src.sentSW.leftedge+i+1)) != null)
            Debug.println(10,"MSS : Removing "+msg.key+" msg from "+src.name+" sent msg queue");
    }

    // flush the sent queue if necessary
    if (src.sentSW.alfa >= SWindow.SWALFA) {
      msg = (MssMessage) src.sent.getFirst();
      while (msg != null) {
        if (Debug.DEBUG)
          Debug.println(10,"swUpdate Flush : Resending  msg "+msg.key+" to "+src.name);
        Header.writeDest(msg.buffer, msg.dest.header, msg.key);
        ehandler.rsend(msg.dest.iroute, msg.buffer, msg.blen);
        msg = (MssMessage) msg.getNext();
      }
    }
    if (Debug.DEBUG)
      Debug.println(10,"swUpdate end");
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Routing table handling
  ////////////////////////////////////////////////////////////////////////////////////////////

  private void rtUpdate(MssHost src, RoutingData rt)
  {
    MssHost             scan;           // Used to scan host table
    int                 i;              // Used to scan rt
    MssHost             route;          // Temporary var

    // PATCH
    boolean             reply = false;  // True if necessary



    src.lambda = 0;
    src.delta  = 1;
    src.route = src;
    src.iroute = src.inetaddr;
    scan = hosttable.getFirst();
    i = 0;
    while (scan != null && i<rt.len)
      if (scan.address == rt.dest[i]) {
        if (scan != me) {
          if (rt.delta[i] == DELTA_INFINITE) {
            if (scan.route == src) {
              scan.route = scan;
              scan.iroute = scan.inetaddr;
              scan.delta = DELTA_INFINITE;
            }
            // PATCH
            if (scan.delta < DELTA_INFINITE)
              reply = true;
          }
          else if (rt.delta[i] + src.delta < scan.delta /* || scan.route == src*/) {
            // route = hosttable.lookup(rt.route[i]);
            // if (route != null) {
              // scan.route = route;
              // scan.iroute = route.inetaddr;
              scan.route = src;
              scan.iroute = src.inetaddr;
              // scan.incarnation = rt.incarns[i];
              if (rt.incarns[i] > scan.incarnation) {
                if (Debug.DEBUG)
                  Debug.println(10, "rtUpdate: handling new incarnation");
                handleNewIncarnation(scan, rt.incarns[i]);
              }
              scan.delta = rt.delta[i] + src.delta;
              // }
          }
        }
        scan = scan.getNext();
        i++;
      }
      else if (scan.address < rt.dest[i])
        scan = scan.getNext();
      else
        i++;

    // PATCH
    if (reply)
      rtTimeout(true);

    /*
      Il codice marcato con PATCH e' stato introdotto per evitare uno
      stupido problema dell'algoritmo di routing; quando un processo
      ne sospetta un'altro, e' possibile che altri processi non si
      accorgano del sospetto per lungo tempo;
    */

    if (Debug.DEBUG)
      Debug.println(10, "rtUpdate: generating suspect");
    generateSuspect(null);
  }

  private void rtTimeout(boolean async)
  {
    MssHost             scan1;          // Used to scan host table
    MssHost             scan2;          // Used to scan host table
    byte[]              buffer;         // Buffer


    if (Debug.DEBUG)
      Debug.println(10, "rtTimeout");

    // Updates the host table
    scan1 = hosttable.getFirst();
    while (scan1 != null) {
      scan1.lambda++;
      if (scan1.route == scan1 && scan1 != me && scan1.reachable)
        if (scan1.lambda >= lambda_suspect) {
          scan1.delta = DELTA_INFINITE;
          scan1.route = scan1;
          scan1.iroute = scan1.inetaddr;
          scan2 = hosttable.getFirst();
          while (scan2 != null) {
            if (scan2.route == scan1) {
              scan2.delta = DELTA_INFINITE;
              scan2.route = scan2;
              scan2.iroute = scan2.inetaddr;
            }
            scan2 = scan2.getNext();
          }
          /*
        } else if (scan1.lambda >= lambda_warning) {
          scan1.delta = DELTA_MAX;
          scan1.route = scan1;
          scan1.iroute = scan1.inetaddr;
          scan2 = hosttable.getFirst();
          while (scan2 != null) {
            if (scan2.route == scan1) {
              scan2.delta = DELTA_MAX;
              scan2.route = scan2;
              scan2.iroute = scan2.inetaddr;
            }
            scan2 = scan2.getNext();
          }
          */
        }
      scan1 = scan1.getNext();
    }

    // hosttable.printRoutingTable();

    // Sends routing table to each other host
    buffer = RoutingData.message(hosttable.size(), hosttable.getFirst());
    Header.writeSrc(buffer, me.header, MSS_RT);
    scan1 = hosttable.getFirst();
    rtsent++;
    while (scan1 != null) {
      if (scan1 != me) {
        Header.writeDest(buffer, scan1.header, rtsent);
        scan1.rcvdSW.win2byte(buffer); // write in the buffer the Flow CRT infos sax
        ehandler.rsend(scan1.inetaddr, buffer, buffer.length);
      }
      scan1 = scan1.getNext();
    }
    if (Debug.DEBUG)
      Debug.println(10, "Sent all");

    // Generate suspect information
    if (Debug.DEBUG)
      Debug.println(10, "rtTimeout: generating suspect");
    generateSuspect(null);

    // Set next timeout
    if (!async)
      ehandler.setTimeout(routingtimeout, new EventList(RT_TIMEOUT, null));
  }

  /**
   *
   */
  void generateSuspect(MssHost newinc)
  {
    MssHost             scan;           // Used to scan host table
    MssHost[]           trset;          // Total set of reachable host
    MssHost[]           nrset;          // Host that have just been added to trset
    MssHost[]           nuset;          // Host that have just been removed from trset
    int                 trsetlen;       // Number of elements in trset
    int                 nrsetlen;       // Number of elements in nrset
    int                 nusetlen;       // Number of elements in nuset
    int                 tr;             // Counter to scan trset
    int                 nr;             // Counter to scan nrset
    int                 nu;             // Counter to scan nuset
    byte[]              buffer;         // Buffer
    MssMessage          msg;


    trsetlen = 0;
    nrsetlen = 0;
    nusetlen = 0;
    scan = hosttable.getFirst();
    while (scan != null) {
      if (scan.delta == DELTA_INFINITE) {
        if (scan.reachable)
          nusetlen++;
      } else {
        trsetlen++;
        if (!scan.reachable)
          nrsetlen++;
      }
      scan = scan.getNext();
    }

    // Checks if the perceveived reachability relation has changed.
    if (nrsetlen > 0 || nusetlen > 0) {
      trset = new MssHost[trsetlen];
      nrset = new MssHost[nrsetlen];
      nuset = new MssHost[nusetlen];
      tr = nr = nu = 0;
      scan = hosttable.getFirst();
      while (scan != null) {
        if (scan.delta == DELTA_INFINITE) {
          if (scan.reachable) {
            scan.reachable = false;
            nuset[nu++] = scan;
          }
        } else {
          trset[tr++] = scan;
          if (!scan.reachable) {
            scan.reachable = true;
            nrset[nr++] = scan;
          }
        }
        scan = scan.getNext();
      }

      // Sending synch messages
      for (int i = 0; i< nrsetlen; i++) {
        MssHost dest = nrset[i];
        dest.msent++;
        if (Debug.DEBUG)
          Debug.println(10,"generateSuspect : sending Synch MSG "+dest.msent+" to "+dest.getName());
        buffer = new byte[Header.SIZE];
        Header.write(buffer, MSS_SYNCH, me.header, dest.header, dest.msent);
        msg = new MssMessage(dest.msent, buffer, Header.SIZE, dest, System.currentTimeMillis(),MSS_SYNCH);
        dest.sent.insert(msg);
        ehandler.rsend(dest.iroute, buffer, Header.SIZE);
      }

      // Generating suspects
      if (Debug.DEBUG) {
        Debug.print(10, "New unreachable daemons: ");
        if (nuset.length > 0)
          for (int i=0; i< nuset.length; i++)
            Debug.print(10, nuset[i].name + " ");
        Debug.println(10, "");
      }
      mssuser.suspect(trset, nrset, nuset, newinc);
    }

  }

  /**
   * Updates the host table by considering the last timeout.
   */
  private void updateTimeout(MssHost host, int t)
  {
    host.rtt = (host.rtt * alfan + t*(alfad - alfan)) / alfad;
    host.dev = (host.dev * alfan + (host.rtt>t ? host.rtt-t : t - host.rtt)* (alfad - alfan))
                / alfad;
    host.timeout = host.rtt + host.dev * 4;
  }

  /**
   * Handles the timeout expiration of a nack msg, resending the nack if the
   * host is reachable
   */
  private void nackTimeout(EventList event)  //sax
  {
    MssMessage              msg;            // Message

    msg = (MssMessage) event.obj;
    if (Debug.DEBUG)
      Debug.println(10, "nackTimeout Expired for stream msg from host :" + msg.dest.name);
    if (msg.dest.reachable) {
      Header.writeDest(me.nackbuffer, msg.dest.header, msg.key);
      ehandler.setTimeout(msg.dest.timeout, event);
      ehandler.rsend(msg.dest.iroute, me.nackbuffer, HEADER_SIZE);
      if (Debug.DEBUG)
        Debug.println(10, "Resending Nack to host "+ msg.dest.name);
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Message receiving section
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Handles a new incarnation.
   */
  private void handleNewIncarnation(MssHost src, int sinc)
  {
    if (Debug.DEBUG)
      Debug.print(10, " handleNewIncarnation ");
    src.setNewIncarnation(sinc);
    generateSuspect(src);
  }



  /**
   * Handles the receiving of a normal message from the net. There are three cases:
   * (i) the message is delivered immediately; (ii) the message is enqueued because is
   * not complete; (iii) the message is enqueued because it is not in FIFO order.
   */
  private void msgReceive(MssHost src, Header header, byte[] buffer, int blen)  //sax
  {
    MssMessage              msg;

    if (Debug.DEBUG) {
      Debug.println(10, "--- Enter: msgReceive --------------------------------------------------");

      // Fatal error :  Only for Debugging
      if (header.mid-src.rcvdSW.leftedge > SWindow.BIT_SIZE) {
        Debug.println(10, "Mss PANIC: Flow CTRL overflow");
        Debug.println(10, "mid:"+header.mid);
        Debug.println(10, "le:"+src.rcvdSW.leftedge);
        Debug.println(10, "sw:"+SWindow.BIT_SIZE );
        System.exit(1);
      }
    }

    // check if there is a nack timeout not yet expired for this msg
    if (header.mid < src.maxmsg && header.mid > src.rcvdSW.leftedge)
      if ((msg = (MssMessage) src.nacks.remove(header.mid)) != null && msg.event != null) {
        if (Debug.DEBUG)
          Debug.println(10, "msgReceive: remove nackTimeout event associated with msg "+msg.key+" : " + msg.event.value);
        msg.event.remove();
        updateTimeout(src, (int) (System.currentTimeMillis()-msg.start));
      }

    // if the msg can be delivered in spite of the FIFO order
    if (header.mid == src.rcvdSW.leftedge+1) {
      if (Debug.DEBUG)
        Debug.println(10,"msgReceive : received msg "+header.mid);
      src.rcvdSW.shift();
      src.maxmsg = (src.maxmsg < src.rcvdSW.leftedge ? src.rcvdSW.leftedge : src.maxmsg);
      queueAdd(src, header, buffer, blen);
      checkReceived(src);
    }

    // if the msg cannot be delivered
    else if (header.mid < src.maxmsg && header.mid > src.rcvdSW.leftedge + 1) {
      if (Debug.DEBUG)
        Debug.println(10,"msgReceive : Insert msg "+header.mid+" in received queue because not respecting the FIFO order");
      if (!src.rcvdSW.lookup(header.mid - src.rcvdSW.leftedge - 1)) {
        src.rcvdSW.set(header.mid - src.rcvdSW.leftedge - 1);
        src.rcvd.insert(new MssMessage(header.mid, buffer, blen, header.tag));
      }
    }
    // if msg cannot be delivered and there is an hop in the sequence not notified yet
    else if (header.mid > src.maxmsg) {
      if (Debug.DEBUG)
        Debug.println(10,"msgReceive : "+src.name+" has received msg "+header.mid+" and has maxmsg "+src.maxmsg);
      int pos = header.mid - src.rcvdSW.leftedge - 1;
      src.rcvdSW.set(pos--);
      src.rcvd.insert(new MssMessage(header.mid, buffer, blen, header.tag));
      src.maxmsg = header.mid;
      // sending nack msgs
      int i=0; // ALB
      if (src != me)
        // while (pos >=0 && !src.rcvdSW.lookup(pos) ) { // OLD SAX
        while (i <= pos && !src.rcvdSW.lookup(pos) ) { // ALB
          if (Debug.DEBUG)
            Debug.println(10,"msgReceived : sending nack to "+src.name+" for msg "+(src.rcvdSW.leftedge+i+1));
          msg = new MssMessage(src.rcvdSW.leftedge+i+1, null, 0, src, System.currentTimeMillis(),
                               NACK_TIMEOUT);
          ehandler.setTimeout(src.timeout, msg.event);
          src.nacks.insert(msg);
          Header.writeDest(me.nackbuffer, src.header, src.rcvdSW.leftedge+i+1);
          ehandler.rsend(src.iroute, me.nackbuffer, HEADER_SIZE);
          i++;
        }
    }

    if (Debug.DEBUG) {
      src.rcvdSW.printSW(10);
      Debug.println(10, "--- Exit: msgReceive -------------------------------");
    }
  }


  /**
   * Checks if in the FIFO-order queue of  <src> there are messages that can
   * be delivered.
   */
  private void checkReceivedOrig(MssHost src)
  {
    MssMessage          msg;            // Messages to deliver

    msg = (MssMessage) src.rcvd.getFirst();
    while (msg != null && msg.key == src.mrcvd +1) {
      src.mrcvd++;
      queueAdd(src, msg);
      msg = (MssMessage) src.rcvd.removeFirst();
    }
  }

  /**
   * Checks if in the FIFO-order queue of  <src> there are messages that can
   * be delivered.
   */
  private void checkReceived(MssHost src)  //sax
  {
    MssMessage          msg;            // Messages to deliver
    int pos = 0;

    while (src.rcvdSW.lookup(pos))
      pos++;
    src.rcvdSW.shift(pos);
    for (int i=0; i < pos; i++) {
      msg = (MssMessage) src.rcvd.removeFirst();
      queueAdd(src,msg);
      if (Debug.DEBUG)
        Debug.println(10,"checkReceived : deliver msg "+msg.key);
    }
  }



  /**
   *
   */
  private void queueAdd(MssHost src, Header header, byte[] buffer, int blen)
  {
    if (header.tag == MSS_MESSAGE) {
      src.totlen += payload;
      src.fragments.insert(new MssMessage(header.mid, buffer, blen, header.tag));
    } else {
      src.totlen += blen - OVERHEAD_SIZE;
      compose(src, header.tag, buffer, blen);
    }
  }

  /**
   *
   */
  private void queueAdd(MssHost src, MssMessage msg)
  {
    if (msg.tag == MSS_MESSAGE) {
      src.totlen += payload;
      src.fragments.insert(msg);
    } else {
      src.totlen += (msg.blen - OVERHEAD_SIZE);
      compose(src, msg.tag, msg.buffer, msg.blen);
    }
  }

  private void compose(MssHost src, byte tag, byte[] buffer, int blen)
  {
    MssMessage          msg;
    InMessage           msgstream;
    int                 pos;

    msgstream = new InMessage(payload, Header.SIZE, 0, src.totlen);
    pos = 0;
    while ((msg = (MssMessage) src.fragments.removeFirst()) != null)
      msgstream.insert(msg.buffer, payload);
    msgstream.insert(buffer, blen - Header.SIZE);
    mssuser.mreceive(tag, msgstream, src);
    src.totlen = 0;
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Message sending section
  ////////////////////////////////////////////////////////////////////////////////////////////

  private void completeSend(byte tag, OutMessage msgstream, MssHost dest)  //sax
  {
    MssMessage    msg;
    int           len;
    byte          type;
    byte[][]      fragments;
    int           nfragments;

    if (dest.reachable) {

      fragments = msgstream.getFragments();
      nfragments = msgstream.getFragmentCount();
      for (int i=0; i < nfragments; i++) {
        if (i == nfragments-1) {
          len = msgstream.getLastPayload() + Header.SIZE;
          type = tag;
        } else {
          len = payload + Header.SIZE;
          type = MSS_MESSAGE;
        }
        dest.msent++;
        Header.write(fragments[i], type, me.header, dest.header, dest.msent);
        if (Debug.DEBUG) {
          Debug.print(10, "HEADER: ");
          for (int k=0; k<21; k++)
            Debug.print(10, " " + fragments[i][k]);
          Debug.println(10, "");
        }

        /*
         * This is required because
         */
        byte buf[] = new byte[len];
        System.arraycopy(fragments[i], 0, buf, 0, len);

        // if the msg can be sent in spite of the flow control infos
        if (dest.msent < dest.sentSW.leftedge + SWindow.BIT_SIZE || dest == me) { // ALB
          if (dest != me) {
            msg = new MssMessage(dest.msent, buf, len, dest, System.currentTimeMillis(),type);
            dest.sent.insert(msg);
          }
          if (Debug.DEBUG)
            Debug.println(10,"completeSend : sending msg "+dest.msent+" to "+dest.getName());
          ehandler.rsend(dest.iroute, buf, len);
        }
        // the msg is frozen because out of the window
        else {
          msg = new MssMessage(dest.msent, buf, len, dest, System.currentTimeMillis(),type);
          dest.frozen.insert(msg);
          if (Debug.DEBUG)
            Debug.println(10,"completeSend : frozen msg "+dest.msent+" to "+dest.getName());
        }
      }

    } else {
      if (Debug.DEBUG)
        Debug.println(10,"completeSend : "+dest.getName()+" is unreachable:" + dest.msent);
    }

  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Nack handling section
  ////////////////////////////////////////////////////////////////////////////////////////////

  private void receiveNack(MssHost src, Header header) //sax
  {
    MssMessage msg;

    if (Debug.DEBUG)
      Debug.println(10,"receiveNack: ReceiveNack for  msg "+header.mid+" from "+src.name);
    // if the source host is reachable resend the msg
    if (src.reachable)
      if ((msg = (MssMessage) src.sent.lookup(header.mid)) != null && msg.event != null) {
        Header.writeDest(msg.buffer, msg.dest.header, msg.key);
        ehandler.rsend(msg.dest.iroute, msg.buffer, msg.blen);
        if (Debug.DEBUG)
          Debug.println(10,"receiveNack : Resending msg "+msg.key+" to "+src.name);
      }
  }

  private void receiveSynch(MssHost src, int mid)
  {
    MssMessage        msg;
    int               pos;

    if (Debug.DEBUG) {
      Debug.println(5,"--- Enter :receive synch -------------------");
      src.rcvdSW.printSW(10);
    }

    // Shift the window if necessary
    if ((pos = (mid - src.rcvdSW.leftedge)) > SWindow.BIT_SIZE) {
      src.rcvdSW.clear();
      src.rcvdSW.setLeftEdge(mid);
    }
    else
      src.rcvdSW.shift(pos);
      src.maxmsg = src.maxmsg < src.rcvdSW.leftedge ? src.rcvdSW.leftedge : src.maxmsg;

    // clear the queue
    msg = (MssMessage) src.rcvd.getFirst();
    while (msg != null && msg.key < mid) {
      src.rcvd.removeFirst();
      msg = (MssMessage) src.rcvd.getFirst();
    }
    src.fragments.clear();
    src.totlen = 0;
    checkReceived(src);
    if (Debug.DEBUG)
      Debug.println(5,"--- Exit :receive synch -------------------");
  }

} // END Mss
