/*
 * Copyright (c) 1998-2000 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package relacs.mss;

import java.net.*;
import java.io.*;

import relacs.RelacsDS;

/**
 * The <code> MssHost </code> class
 *
 * @author  Salvatore Cammarrata
 * @author  Alberto Montresor
 * @since   Jgroup 0.3
 */
public class    MssHost
implements      MssConstants, MssTag
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  // Keys
  int                   address;        // 32-bit address used to identify hosts
  int                   incarnation;    // Incarnation number
  String                name;           // Name of the host
  int                   index;          // Index of the host in the host table

  // Incarnation info
  byte[]                header;         // Header (used to avoid encoding)
  byte[]                nackbuffer;     // Buffer used to send nack //sax

  // Routing info
  InetAddress           inetaddr;       // InetAddress to send data
  boolean               reachable;      // True if the host is reachable
  int                   lambda;         // Number of ping interval without hearing from host
  int                   delta;          // Distance (number of host) from host
  InetAddress           iroute;         // Internet address of route host
  MssHost               route;          // Route host

  // Adaptive timeout
  int                   rtt;            // Roundtrip timeout
  int                   dev;            // Standard deviation
  int                   timeout;        // Next timeout

  // Messages information
  MssMessage            rcvd;           // Received messages
  MssMessage            sent;           // Sent messages
  MssMessage            fragments;      // Fragments of the next message to deliver
  int                   totlen;         // Total length of the next message to deliver
  int                   rtrcvd;         // Id. of the last routing table received from host
  int                   msent;          // Id. of the last message m-sent to host
  int                   mrcvd;          // Id. of the last message m-received from host


  // Flow Control information           // sax
  SWindow               sentSW;         // Sliding Window for msgs sent to MssHost and have been received from it.
  SWindow               rcvdSW;         // Sliding Window for msgs received from MssHost
  MssMessage            nacks;          // List of nack sent to host
  MssMessage            frozen;         // List of msgs that connot be sent and have to be frozen
  int                   maxmsg;         // Biggest Id. m-received

  // MssHost table link
  MssHost               hnext;          // Hash chain reference
  MssHost               tnext;          // Total list chain reference



  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Builds an empty host, used in host table for initialization.
   */
  public MssHost()
  {
  }

  /**
   * Builds an host identified by <code> name <\code>.
   */
  public MssHost(String name)
  throws IOException
  {
    // Initializes address information
    this.name = name;
    inetaddr  = InetAddress.getByName(name);
    address   = Types.byte2hid(inetaddr.getAddress(), 0);
    index = -1;

    // Initializes fixed header and host
    // reachable = (myaddress == address);
    reachable = (inetaddr.equals(RelacsDS.getLocalHost()));
    incarnation = (reachable ? (int) (System.currentTimeMillis()/1000) : 0);
    header = new byte[Header.HOSTDATA_SIZE];
    Header.writeHostData(header, address, incarnation);

    nackbuffer = new byte[Header.SIZE];                         //sax
    Header.writeSrc(nackbuffer, header, MSS_NACK);              //sax

    // Initializes routing information
    // this.lambda      = lambda;
    this.lambda       = Integer.MAX_VALUE;
    delta       = (reachable ? 0 : DELTA_INFINITE);
    route       = this;
    iroute      = inetaddr;
    rtt         = START_RTT;
    dev         = START_DEV;
    timeout     = START_TIMEOUT;

    // Initializes data structures
    msent       = 0;
    mrcvd       = 0;
    rtrcvd      = 0;
    totlen      = 0;
    sent        = new MssMessage();
    rcvd        = new MssMessage();
    fragments   = new MssMessage();
    sentSW      = new SWindow();                        // sax
    rcvdSW      = new SWindow();                        // sax
    maxmsg      = 0;                                    // sax
    nacks       = new MssMessage();                     // sax
    frozen      = new MssMessage();                     // sax
   }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /*
   *
   */
  public MssHost getNext()
  {
    return tnext;
  }

  /*
   *
   */
  public int getAddress()
  {
    return address;
  }

  /*
   *
   */
  public InetAddress getInetAddress()
  {
    return inetaddr;
  }

  /*
   *
   */
  public String getName()
  {
    return name;
  }

  /*
   *
   */
  public int getIndex()
  {
    return index;
  }

  /*
   *
   */
  public int getIncarnation()
  {
    return incarnation;
  }


  /**
   * Handles a new incarnation.
   */
  void setNewIncarnation(int newIncarnation)
  {
    incarnation = newIncarnation;
    rtrcvd = 0;
    msent  = 0;     // ALB
    mrcvd  = 0;
    totlen = 0;
    sent.clear();
    rcvd.clear();
    fragments.clear();
    sentSW.clear();   // sax
    rcvdSW.clear();   // sax
    nacks.clear();    // sax
    maxmsg = 0;       // sax
    frozen.clear();   // sax
    Header.writeHostData(header, address, newIncarnation);
    delta = DELTA_INFINITE;
    reachable = true;
  }


} //END MssHost





