/*
 * Copyright (c) 1998-2000 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package relacs.mss;

import java.io.*;

import debug.Debug;

/**
 * The <code>OutMessage</code> class is an output stream especially designed
 * to be used with a transport layer which fragments messages. Messages are
 * stored in fragments, each fragment being composed by an header, a payload
 * and a trailer. The length of these fields may be configured when creating
 * a new <code>OutMessage</code>. Data written in the message are stored in
 * the payload field of the fragments.
 *
 * @author      Alberto Montresor
 * @since       Jgroup 0.8
 */
public final class      OutMessage
extends                 OutputStream
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Initial number of fragments composing a message; can grow */
  private static final int NUM_BUFFER = 16;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static section
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Temporary buffer used in writeTemp */
  private static byte[] temp = new byte[8];

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  private int       payload;            // Payload size in bytes
  private int       header;             // Header size in bytes
  private int       trailer;            // Trailer size in bytes
  private int       totlen;             // Total length of a fragment
  private int       endpayload;         // Last byte used for payload

  private int       position;           // Current position in the stream [0..byteCount]
  private int       fragment;           // Current fragment number
  private int       offset;             // Current offset in the fragment

  private int       fragmentCount;      // Total number of fragments used
  private int       byteCount;          // Total number of bytes written
  private int       maxsize;            // Max size in bytes

  private byte[]    buffer;             // Current fragment
  private byte[][]  fragments;          // Byte array fragments



  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////


  /**
   * Construct a message with header and trailer of length zero. In this way, OutMessage may
   * be used as a more efficient version of ByteArrayOutputStream.
   */
  public OutMessage(int payload)
  {
    this(payload, 0, 0);
  }

  /**
   * Construct a OutMessage starting from an InMessage. In this way, it possible to forward
   * messages received from other servers without copying them.
   */
  public OutMessage(InMessage msg)
  {
    // Store and compute fixed limits
    this.payload = msg.getPayload();
    this.header  = msg.getHeader();
    this.trailer = msg.getTrailer();
    endpayload = payload + header;
    totlen = endpayload + trailer;

    // Initialize variables regarding message size
    byteCount = msg.getByteCount();
    fragmentCount = msg.getFragmentCount();
    maxsize = byteCount;

    // Initialize variables regarding message position in the stream
    position = 0;
    fragment = 0;
    offset = header;

    // Create fragments
    fragments = msg.getFragments();
    buffer = fragments[0];
  }

  /**
   *  This constructor builds a message output stream subdivided in
   *  fragments with length <code> header + payload + trailer</code>.
   *  Used for messages whose size is initially unknown.
   *  A <tt>IndexOutOfBoundsException</tt> is thrown if: <tt>payload</tt> is
   *  negative or equal to zero; <tt>header</tt> is negative; <tt>trailer</tt>
   *  is negative.
   */
  OutMessage(int payload, int header, int trailer)
  {
    if (payload <= 0 || header < 0 || trailer < 0)
      throw new IndexOutOfBoundsException();

    // Store and compute fixed limits
    this.payload = payload;
    this.header  = header;
    this.trailer = trailer;
    endpayload = payload + header;
    totlen = endpayload + trailer;

    // Initialize variables regarding message size
    byteCount = 0;
    fragmentCount = 1;
    maxsize = Integer.MAX_VALUE;

    // Initialize variables regarding message position in the stream
    position = 0;
    fragment = 0;
    offset = header;

    // Create fragments
    fragments = new byte[NUM_BUFFER][];
    fragments[0] = new byte[totlen];
    buffer = fragments[0];
  }

  /**
   *  This constructor builds a message output stream subdivided in
   *  fragments. Used for messages whose size is known.
   *  If <code>maxsize</code> is less than <code> payload</code>, the
   *  payload of the message is limited to <code>maxsize</code>.
   *  The total length of each fragments is <code> header + payload +
   *  trailer</code>
   *  <p>
   *  A <tt>IndexOutOfBoundsException</tt> is thrown if: <tt>payload</tt> is
   *  negative or equal to zero; <tt>header</tt> is negative; <tt>trailer</tt>
   *  is negative; <tt>maxsize</tt> is negative or equal to zero.
   */
  OutMessage(int payload, int header, int trailer, int maxsize)
  {
    if (payload <= 0 || header < 0 || trailer < 0 || maxsize <= 0)
      throw new IndexOutOfBoundsException();

    // Store and compute fixed limits
    this.payload = (maxsize < payload ? maxsize : payload);
    this.header = header;
    this.trailer = trailer;
    endpayload = this.payload + header;
    totlen = endpayload + trailer;

    /*
     * Initialize variables regarding message size; the total number of byte is
     * already set to maxsize; in this way, in the following we'll skip to check
     * whether the size exceed maxsize or not.
     *
     */
    byteCount = maxsize;
    fragmentCount = ((maxsize-1) / payload) + 1;
    this.maxsize = maxsize;

    // Initialize variables regarding message position in the stream
    position = 0;
    fragment = 0;
    offset = header;

    // Create fragments
    fragments = new byte[fragmentCount][];
    for (int i=0; i< fragmentCount; i++)
      fragments[i] = new byte[totlen];
    buffer = fragments[0];
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods inherited from OutputStream
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Writes the specified byte to this byte array output stream.
   * <p>
   * If the total number of bytes written in the message exceeds the maximum
   * length, a <tt>ArrayIndexOutOfBoundsException</tt> is thrown.
   *
   * @param   b   the byte to be written.
   */
  public void write(int b)
  {
    /* Increase the number of stored bytes */
    position++;
    if (position > byteCount) {
      byteCount = position;
      if (position > maxsize)
        throw new ArrayIndexOutOfBoundsException();
      if (offset == endpayload) {
        /*
         *  We have reached the end of the last fragment; we add a further fragment
         *  to store the new byte; fragment, offset and buffer are updated to point
         *  to the fist byte of the new fragment.
         */
        fragment++;
        offset = header;
        enlargeSpace(byteCount);
        buffer = fragments[fragment];
      }
    }
    /* We write the byte and increment offset */
    buffer[offset++] = (byte) b;
  }


  /**
   * Writes <code>len</code> bytes from the specified byte array
   * starting at offset <code>off</code> to this msg output stream.
   * <p>
   * If <code>b</code> is <code>null</code>, a
   * <code>NullPointerException</code> is thrown.
   * <p>
   * If <code>off</code> is negative, or <code>len</code> is negative, or
   * <code>off+len</code> is greater than the length of the array
   * <code>b</code>, then an <tt>IndexOutOfBoundsException</tt> is thrown.
   * <p>
   * If the total number of bytes written in the message exceeds the maximum
   * length, a <tt>ArrayIndexOutOfBoundsException</tt> is thrown.
   *
   * @param   b     the data.
   * @param   off   the start offset in the data.
   * @param   len   the number of bytes to write.
   */
  public void write(byte b[], int off, int len)
  {
    if ((off<0) || (off>b.length) || (len<0) || ((off+len) > b.length)) {
      throw new IndexOutOfBoundsException();
    } else if (len == 0) {
      return;
    } else {
      position += len;
      if (position > byteCount) {
        byteCount = position;
        if (position > maxsize)
          throw new ArrayIndexOutOfBoundsException();
      }
      if (offset + len <= endpayload) {
        System.arraycopy(b, off, buffer, offset, len);
        offset += len;
      } else {
        enlargeSpace(byteCount);
        int todo = endpayload - offset;
        System.arraycopy(b, off, buffer, offset, todo);
        int remaining = len-todo;
        off += todo;
        while (remaining > payload) {
          System.arraycopy(b, off, fragments[fragment], header, payload);
          remaining -= payload;
          off += payload;
          fragment++;
        }
        System.arraycopy(b, off, fragments[++fragment], header, remaining);
        offset = remaining + header;
        buffer = fragments[fragment];
      }
    }
  }

  /**
   * Writes <code>b.length</code> bytes from the specified byte array
   * to this output stream. The general contract for <code>write(b)</code>
   * is that it should have exactly the same effect as the call
   * <code>write(b, 0, b.length)</code>.
   * <p>
   * If the total number of bytes written in the message exceeds the maximum
   * length, a <tt>ArrayIndexOutOfBoundsException</tt> is thrown.
   *
   * @param      b   the data.
   *
   */
  public void write(byte b[])
  {
    write(b, 0, b.length);
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Public methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Move the current position in the stream to <code>position</code>.
   * <p>
   * If the current position in the message exceeds the maximum
   * length, a <tt>ArrayIndexOutOfBoundsException</tt> is thrown.
   */
  public void seek(int position)
  {
    fragment = position / payload;
    offset = header + position % payload;
    this.position = position;
    if (position > byteCount) {
      byteCount = position;
      if (position > maxsize)
        throw new ArrayIndexOutOfBoundsException();
      enlargeSpace(position);
    }
    buffer = fragments[fragment];
  }


  /**
   * Write a short integer <code>v</code> on the stream at the current position.
   * <p>
   * If the total number of bytes written in the message exceeds the maximum
   * length, a <tt>ArrayIndexOutOfBoundsException</tt> is thrown.
   */
  public void write2(int v)
  {
    position += 2;
    if (position > byteCount) {
      byteCount = position;
      if (position > maxsize)
        throw new ArrayIndexOutOfBoundsException();
    }
    if (offset + 2 > endpayload) {
      temp[0] = (byte) (v >>> 8);
      temp[1] = (byte) (v >>> 0);
      writeTemp(temp, 2);
    } else {
      buffer[offset++] = (byte) (v >>>  8);
      buffer[offset++] = (byte) (v >>>  0);
    }
  }

  /**
   * Write a integer <code>v</code> on the stream at the current position.
   * <p>
   * If the total number of bytes written in the message exceeds the maximum
   * length, a <tt>ArrayIndexOutOfBoundsException</tt> is thrown.
   */
  public void write4(int v)
  {
    position += 4;
    if (position > byteCount) {
      byteCount = position;
      if (position > maxsize)
        throw new ArrayIndexOutOfBoundsException();
    }
    if (offset + 4 > endpayload) {
      temp[0] = (byte) (v >>> 24);
      temp[1] = (byte) (v >>> 16);
      temp[2] = (byte) (v >>>  8);
      temp[3] = (byte) (v >>>  0);
      writeTemp(temp, 4);
    } else {
      buffer[offset++] = (byte) (v >>>  24);
      buffer[offset++] = (byte) (v >>>  16);
      buffer[offset++] = (byte) (v >>>   8);
      buffer[offset++] = (byte) (v >>>   0);
    }
  }

  /**
   * Write a long <code>v</code> on the stream at the current position.
   * <p>
   * If the total number of bytes written in the message exceeds the maximum
   * length, a <tt>ArrayIndexOutOfBoundsException</tt> is thrown.
   */
  public void write8(long v)
  {
    position += 8;
    if (position > byteCount) {
      byteCount = position;
      if (position > maxsize)
        throw new ArrayIndexOutOfBoundsException();
    }
    int v1 = (int) (v >>> 32);
    int v2 = (int) (v & 0xFFFFFFFFL);
    if (offset + 8 > endpayload) {
      temp[0] = (byte) (v1 >>> 24);
      temp[1] = (byte) (v1 >>> 16);
      temp[2] = (byte) (v1 >>>  8);
      temp[3] = (byte) (v1 >>>  0);
      temp[4] = (byte) (v2 >>> 24);
      temp[5] = (byte) (v2 >>> 16);
      temp[6] = (byte) (v2 >>>  8);
      temp[7] = (byte) (v2 >>>  0);
      writeTemp(temp, 8);
    } else {
      buffer[offset++] = (byte) (v1 >>> 24);
      buffer[offset++] = (byte) (v1 >>> 16);
      buffer[offset++] = (byte) (v1 >>> 8);
      buffer[offset++] = (byte) (v1 >>> 0);
      buffer[offset++] = (byte) (v2 >>> 24);
      buffer[offset++] = (byte) (v2 >>> 16);
      buffer[offset++] = (byte) (v2 >>> 8);
      buffer[offset++] = (byte) (v2 >>> 0);
    }
  }

  /**
   * Write the contents of another outmessage in the current message.
   * <p>
   * If the total number of bytes written in the message exceeds the maximum
   * length, a <tt>ArrayIndexOutOfBoundsException</tt> is thrown.
   */
  public void write(OutMessage msg)
  {
    byte[][] fragments = msg.getFragments();
    int length = msg.getFragmentCount();
    for (int i=0; i < length-1; i++)
      write(fragments[i], msg.getHeader(), msg.getPayload());
    write(fragments[length-1], msg.getHeader(), msg.getLastPayload());
  }

  /**
   *  Reset the byte array
   */
  public void reset()
  {
    // Reinitialize fragments
    buffer = fragments[0];

    // Initialize variables regarding message size
    byteCount = 0;
    fragmentCount = 1;
    maxsize = Integer.MAX_VALUE;

    // Initialize variables regarding message position in the stream
    position = 0;
    fragment = 0;
    offset = header;
  }

  /**
   * Compute the checksum of the bytes stored in the message
   *
   * @return the computed checksum.
   */
  public int computeChecksum()
  {
    byte        check = 0;                      // Return value

    for (int j = 0; j < fragmentCount - 1; j++)
      for (int i=header; i < endpayload; i++)
        check ^= fragments[j][i];
    byte[] buffer = fragments[fragmentCount -1];
    int end = getLastPayload()+header;
    for (int i=header; i<end; i++)
      check ^= buffer[i];

    return check;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Package methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Return the length of the payload for this message.
   */
  int getPayload()
  {
    return payload;
  }

  /**
   * Return the length of the header for this message.
   */
  int getHeader()
  {
    return header;
  }

  /**
   * Return the length of the trailer for this message.
   */
  int getTrailer()
  {
    return trailer;
  }

  /**
   *  Returns the total number of bytes stored in this message.
   */
  public int getByteCount()
  {
    return byteCount;
  }

  /**
   *  Returns the number of fragments composing the message.
   */
  int getFragmentCount()
  {
    return fragmentCount;
  }

  /**
   *  Returns the length of the payload of the last fragment composing the message.
   */
  int getLastPayload()
  {
    if (byteCount <= payload)
      return byteCount;
    else
      return byteCount % payload;
  }

  /**
   *  Returns an array of byte arrays which contains the fragments composing the
   *  message. The length of fragment array may be greater of the number of fragments
   *  actually allocated; the actual number of fragments is obtained by invoking
   *  method <code>getFragmentCount</code>.
   */
  byte[][] getFragments()
  {
    return fragments;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Private methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Copy <code>length</code> bytes from array <code>temp</code>; this method
   * is invoked when methods <code>write2</code>, <code>write4</code>, or
   * <code>write8</code> should write part of their data in a fragment and
   * part in the following one.
   */
  private void writeTemp(byte[] temp, int length)
  {
    enlargeSpace(byteCount);
    for (int i = 0; i < length; i++) {
      if (offset == endpayload) {
        fragment++;
        buffer = fragments[fragment];
        offset = header;
      }
      buffer[offset++] = temp[i];
    }
  }


  /**
   *  Alloc new byte fragments to guarantee that there is enough space to store
   *  <code> newsize </code> bytes.
   */
  private void enlargeSpace(int newsize)
  {
    int nfragment = (newsize-1) / payload + 1;  // Total number of fragments to store <size> bytes
    if (nfragment > fragments.length) {
      // Double the fragment array if needed
      byte[][] tempbuffer = new byte[fragments.length*2][];
      System.arraycopy(fragments, 0, tempbuffer, 0, fragments.length);
      fragments = tempbuffer;
    }
    for (int i = fragmentCount; i < nfragment; i++)
      fragments[i] = new byte[totlen];
    if (fragmentCount < nfragment)
      fragmentCount = nfragment;
  }

  public void print(int level)
  {
    if (Debug.DEBUG) {
      Debug.print(level,"OutMessage: ");
      for (int i=0; i < fragmentCount-1; i++)
        for (int j=header; j < endpayload; j++)
          Debug.print(level, " " + fragments[i][j]);
      for (int j=header; j < byteCount % payload; j++)
        Debug.print(level, " " + fragments[fragmentCount-1][j]);
      Debug.println(level, "");
    }
  }



} // END OutMessage


