/*
 * Copyright (c) 1998-2000 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package relacs.mss;

import debug.Debug;

/**
 * The <code> SWindow </code> class
 *
 * @author  Salvatore Cammarrata
 * @since   Jgroup 0.5
 */
class       SWindow
implements  MssConstants
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  static final int SWIN_SIZE = 4;               // byte size of the window
  static final int BIT_SIZE = SWIN_SIZE * 8;    // bit size of the window
  static final int SWALFA = 4;                  // parameter used to flush the sent msg queue
                                                // if necessary


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  int leftedge;                                 // left edge of the window
  int prevLeftedge;                             // previous leftedge
  int alfa;                                     // incremented if the last update is the same of the current window

  private boolean[] win;  // window of bit

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  public SWindow()
  {
    this.win = new boolean[BIT_SIZE];
  }

  public void printSW(int debLevel)
  {
    if (Debug.DEBUG) {
      Debug.print(debLevel,"Flow CTRL : ");
      for (int i=0; i<BIT_SIZE; i++)
        if (win[i])
          Debug.print(debLevel,"1");
      else
        Debug.print(debLevel,"0");
      Debug.println(debLevel,"\nLeftEdge : "+leftedge);
    }
  }


  /**
   *  Test if all the bit of the window are set to false
   */
  public boolean isAllZero()
  {
    int i;

    for (i = 0; i< BIT_SIZE && !win[i]; i++);
    return (i == BIT_SIZE);
  }


  /**
   * Shift the window to left of <pos> bit
   */
  public int shift(int pos)
  {
    int i;
    for (i=0; i < BIT_SIZE - pos ; i++)
      win[i]=win[i+pos];
    for (i= BIT_SIZE - pos; i < BIT_SIZE ; i++)
      win[i]=false;
    return (leftedge += pos);
  }


  /**
   * Shift the window to left of one bit
   */
  public int shift()
  {
    return shift(1);
  }

  /**
   *
   */
  public int setLeftEdge(int le)
  {
    int ple = leftedge;
    leftedge = le;
    return ple;
  }

  /**
   *
   */
  public int getLeftEdge()
  {
    return leftedge;
  }

  /**
   *
   */
  public int incLeftEdge()
  {
    return (leftedge += 1);
  }

  /**
   * Copy the flow control information in the buffer <data> starting from
   * <pos> byte
   */
  public int win2byte(byte[] data, int pos)
  {
    byte    tmp;
    int     nbyte;

    data[pos++] = (byte) (leftedge >>> 8);
    data[pos++] = (byte) leftedge;

    for (nbyte=0; nbyte < SWIN_SIZE ; nbyte++) {
      tmp= 0x00;
      for (int nbit=0 ; nbit < 8; nbit++)
      {
        if (win[8*nbyte+nbit])
          tmp |= 0x01;
        if (nbit != 7)
          tmp <<= 1;
      }
      data[nbyte + pos] = tmp;
    }
    return pos + nbyte + 1;
  }



  /**
   * Copy the flow control information in the buffer <data> starting from
   * the first byte
   */
  public int win2byte(byte[] data)
  {
    byte tmp;
    int nbyte;
    int pos = HEADER_SIZE;

    data[pos++] = (byte) (leftedge >>> 8);
    data[pos++] = (byte) leftedge;

    for (nbyte=0; nbyte < SWIN_SIZE ; nbyte++) {
      tmp= 0x00;
      for (int nbit=0 ; nbit < 8; nbit++) {
        if (win[8*nbyte+nbit])
          tmp |= 0x01;
        if (nbit != 7)
          tmp <<= 1;
      }
      data[nbyte + pos] = tmp;
    }
    return pos + nbyte + 1;
  }


  /**
   * Copy the flow control information from the buffer <data> to the window
   */
  public int byte2win(byte[] data, int pos)
  {
    byte tmp;

    prevLeftedge = leftedge;
    leftedge = ((data[pos++] << 8) & 0xFF00) | (data[pos++] & 0xFF);

    alfa = ((leftedge == prevLeftedge) && isAllZero() ? alfa + 1 : 0);
    for (int nbyte=0; nbyte < SWIN_SIZE  ; nbyte++) {
      tmp= data[pos+nbyte];
      for (int nbit=0 ; nbit < 8; nbit++) {
        if ((tmp & 0x80) != 0)
          win[8*nbyte+nbit] = true;
        else
          win[8*nbyte+nbit] = false;
        tmp <<= 1;
      }
    }
    return leftedge;
  }


  /**
   *
   */
  public int byte2win(byte[] data)
  {
    byte tmp;
    int pos = HEADER_SIZE;

    prevLeftedge = leftedge;
    leftedge = ((data[pos++] << 8) & 0xFF00) | (data[pos++] & 0xFF);
    alfa = (leftedge == prevLeftedge ? alfa + 1 : 0);

    for (int nbyte=0; nbyte < SWIN_SIZE  ; nbyte++) {
      tmp= data[pos+nbyte];
      for (int nbit=0 ; nbit < 8; nbit++) {
        if ((tmp & 0x80) != 0)
          win[8*nbyte+nbit] = true;
        else
          win[8*nbyte+nbit] = false;
        tmp <<= 1;
      }
    }
    return leftedge;
  }

  /**
   *
   */
  public boolean set(int pos)
  {
    return (win[pos] = true);
  }

  public boolean unset(int pos)
  {
    win[pos] = false;
    return true;
  }

  /**
   *
   */
  public boolean lookup(int pos)
  {
    return win[pos];
  }

  /**
   *
   */
  public void clear()
  {
    leftedge = 0;
    prevLeftedge = 0;
    alfa = 0;
    for (int i=0 ; i < BIT_SIZE ; i++)
      win[i]=false;
  }

} // END SWindow
