/*
 * Copyright (c) 1998-2000 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package relacs.rmi;

import java.io.*;
import java.net.*;
import java.rmi.*;
import java.rmi.server.*;
import java.lang.reflect.*;
import java.util.*;

import jgroup.*;
import jgroup.multicast.*;
import jgroup.registry.*;

import relacs.*;
import relacs.gm.*;
import relacs.mss.Upcall;
import relacs.general.*;

import debug.Debug;

/**
 *
 * @author  Alberto Montresor
 * @since   0.8
 */
public class  ExtGMILayer
extends       UnicastRemoteObject
implements    ExternalGMIService, MulticastListener, GroupDispatcher
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static section
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Map between ExtGMILayers and associated servers */
  private static HashMap map = new HashMap();
  
  /**
   *  Return the map between a server object and key
   */
  public static Object get(Object key)
  {
    return map.get(key);
  }
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /* Requested services */
  private MembershipService   membershipService;  // Underlying membership service
  private MulticastService    multicastService;   // Underlying multicast service

  /* Provided services */
  private ExternalGMIListener server;             // Layer listening for external invocation

  /* Internal variables */
  private IIDTable            table;              // Table containing invocations
  private HashMap             ids;                // Table containing current ids
  private RelacsDS            rds;                // Distributed system description
  private ServerMethodTable   methodTable;        // Method table for converting hash to method

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *
   * @exception DSException
   *            raised when the distributed system configuration object contains an error
   */
  public ExtGMILayer(Object[] configuration,
                  MembershipService membershipService,
                  MulticastService multicastService)
    throws JgroupException, RemoteException
  {
    for (int i=0; i < configuration.length; i++) {
      if (configuration[i] instanceof RelacsDS)
        rds = (RelacsDS) configuration[i];
      else if (configuration[i] instanceof DistributedSystem)
        rds = new RelacsDS((DistributedSystem) configuration[i]);
    }
    if (rds == null)
      throw new JgroupException("The external GMI service requires a DistributedSystem object");

    this.membershipService = membershipService;
    this.multicastService = multicastService;
    table = new IIDTable(membershipService.getMyIdentifier());
    ids = new HashMap(7);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static factory
  ////////////////////////////////////////////////////////////////////////////////////////////

  public static ExtGMILayer getLayer(Object[] configuration,
                  MembershipService membershipService,
                  MulticastService multicastService)
    throws JgroupException, RemoteException
  {
    return new ExtGMILayer(configuration, membershipService, multicastService);
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // setListener methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /*
   */
  public void addListener(ExternalGMIListener server)
    throws JgroupException
  {
    if (this.server != null)
      throw new JgroupException("Error composing group managers layers; multiple local " +
        " listeners for group method invocations");

    /* Store the server reference for future invocations */
    this.server = server;
    
    /** Save the association between this ExtGMILayer and the server */
    map.put(server, this);

    /* Compute the external interfaces implemented by this object */
    Class[] interfaces = server.getClass().getInterfaces();
    ArrayList array = new ArrayList();
    for (int i=0; i < interfaces.length; i++) {
      if (ExternalGMIListener.class.isAssignableFrom(interfaces[i]))
        array.add(interfaces[i]);
    }
    interfaces = (Class[]) array.toArray(new Class[array.size()]);

    /* Compute the method table for these interfaces */
    methodTable = new ServerMethodTable(interfaces);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Dispatching methods (from GroupDispatcher)
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *
   */
  public Object invoke(IID iid, long hash, Object[] args)
    throws Exception
  {
    if (Debug.DEBUG)
      Debug.println(40, "ExtGMILayer: anycast invoke non-void");
    try {
      IIDEntry entry = table.insert((IIDImpl) iid); // throws RemoteException if removed
      Method m = methodTable.get(hash);
      entry.anycastInvoke(server, m, args);
      return entry.anycastResult();
    } catch (Exception e) {
      e.printStackTrace();
      throw e;
    }
  }

  /**
   *
   *  Implementation note: this method is not synchronized; if concurrent invocations
   *  of this method are executed, the synchronized methods invoked in the method
   *  body provide safety with respect to concurrency. First of all, the invocation
   *  of method <tt>insert</tt> on the iid table (first line) is synchronized; this
   *  is to avoid concurrent modification of the table. Then, invocation of
   *  <tt>multicastInvoke</tt> and of </tt>multicastResult</tt> on the obtained
   *  entry are also synchronized.
   */
  public Object invoke(IID iid, long hash, ByteVectorOutputStream msg)
    throws Exception
  {
    if (Debug.DEBUG)
      Debug.println(40, "ExtGMILayer: multicast invoke non-void");
    try {
      IIDImpl iidimpl = (IIDImpl) iid;
      IIDEntry entry = table.insert(iidimpl); // throws RemoteException if removed
      try {
        OutputStream message = multicastService.getMessage();
        ObjectOutputStream out = new MarshalOutputStream(message);
        out.writeLong(hash);
        out.writeBoolean(iidimpl != null);
        if (iidimpl != null)
          iidimpl.writeExternal(out);
        msg.writeExternal(out);
        out.flush();
        entry.multicastInvoke(multicastService, message);
      } catch (Exception e) {
        throw new McastRemoteException(e.getMessage());
      }
      Object res = entry.multicastResult();
      return res;
    } catch (Exception e) {
      e.printStackTrace();
      throw e;
    }
  }


  /**
   *
   */
  public void vinvoke(IID iid, long hash, Object[] args)
    throws Exception
  {
    if (Debug.DEBUG)
      Debug.println(40, "ExtGMILayer: anycast invoke void");
    try {
      IIDEntry entry = table.insert((IIDImpl) iid); // throws RemoteException if removed
      Method m = methodTable.get(hash);
      entry.anycastInvoke(server, m, args);
    } catch (Exception e) {
      e.printStackTrace();
      throw e;
    }
  }

  /**
   *
   */
  public void vinvoke(IID iid, long hash, ByteVectorOutputStream msg)
    throws Exception
  {
			//	System.out.println(iid);
    if (Debug.DEBUG)
      Debug.println(40, "ExtGMILayer: multicast invoke void");
    try {
      IIDImpl iidimpl = (IIDImpl) iid;
      IIDEntry entry = table.insert(iidimpl); // throws RemoteException if removed
      try {
        OutputStream message = multicastService.getMessage();
        ObjectOutputStream out = new ObjectOutputStream(message);
        out.writeLong(hash);
        out.writeBoolean(iidimpl != null);
        if (iidimpl != null)
          iidimpl.writeExternal(out);
        msg.writeExternal(out);
        out.flush();
        entry.multicastInvoke(multicastService, message);
      } catch (Exception e) {
        throw new McastRemoteException(e.getMessage());
      }
      /* We call multicastResult to be sure that the method invocation has been completed */
			//System.out.println("Before multicast result");
      entry.multicastResult();
    } catch (Exception e) {
      e.printStackTrace();
      throw e;
    }
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from MulticastListener
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Upcall that is invoked by Jgroup to deliver a message <code> msg</code>
   *  related to the group identified by <code> group</code>.
   *
   *  @param msg      the stream from which the message may be read.
   */
  public Object deliverStream(InputStream msg)
  {
    if (Debug.DEBUG)
      Debug.println(40, "ExtGMILayer: received stream message");
    try {
      ObjectInputStream in = new MarshalInputStream(msg);
      long hash = in.readLong();
      boolean hasIid = in.readBoolean();
      if (hasIid) {               // External invocation
        IIDImpl iid = new IIDImpl();
        iid.readExternal(in);

        IIDEntry entry = null;

        try { entry = table.insert((IIDImpl) iid); }
        catch (Exception e) {
          /* An exception here means that the method has already been computed,
           * and the result has been discarded after some time; there is no way
           * to access the result again
           */
          if (Debug.DEBUG)
            Debug.printStackTrace(40, e);
          return null;
        }

        /* Obtain the arguments from the message */

        ByteVectorOutputStream bvector = new ByteVectorOutputStream();
        bvector.readExternal(in);
        ObjectInputStream serializedArgs =
          new MarshalInputStream(new ByteArrayInputStream(bvector.getByteArray()));
        Object[] args = (Object[]) serializedArgs.readObject();

        /* Obtain the method from the method hash */
        Method m = methodTable.get(hash);

        /* Execute the invocation */
        ids.put(Thread.currentThread(), iid);
        entry.executeMulticastInvocation(server, m, args);
        ids.remove(Thread.currentThread());

      } else {
        /*
         * The invocation is internal
         */

        /*
         * FIXME 20/11/2000: Raised UnsupportedOperationException
         * in order that this message is handled by the internal
         * invocation handler.
         */
        throw new UnsupportedOperationException("ExtGMILayer: Internal GMI");

      }
      return null;

    } catch (Exception e) {
      if (Debug.DEBUG) {
        Debug.println(40, "ExtGMILayer: I'm not the recipient");
        Debug.printStackTrace(40, e);
      }
      throw new UnsupportedOperationException();
    }
  }


  /**
   *  Upcall that is invoked by Jgroup to deliver a message <code> obj</code>
   *  related to the group identified by <code> group</code>. In this case,
   *  the message is an object marshalled and unmarshalled through java
   *  serialization.
   *
   *  @param obj        the object containing the message.
   */
  public Object deliverObject(Object obj)
  {
    throw new UnsupportedOperationException();
  }



  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods of ExternalGMIService
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *
   */
  public IID getIdentifier()
  {
    return (IID) ids.get(Thread.currentThread());
  }

  /**
   *
   */
  public IID bind(String name)
    throws RemoteException, AccessException
  {
    DependableRegistry registry = RegistryFactory.getRegistry(rds);
    return registry.bind(name, this, server.getClass());
  }

  /**
   *
   */
  public IID bind(String name, DependableRegistry registry)
    throws RemoteException, AccessException
  {
    return registry.bind(name, this, server.getClass());
  }

} // END ExtGMILayer






