/*
 * Copyright (c) 1998-2000 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package relacs.rmi;

import java.io.*;
import java.rmi.*;
import java.util.*;
import java.lang.reflect.*;

import jgroup.*;
import relacs.rmi.proxy.*;
import relacs.general.*;

import debug.Debug;


/**
 *
 *
 *  @author   Alberto Montresor
 *  @since    Jgroup 0.9
 */
public class ExtGroupHandler2
implements JgroupHandler, Externalizable
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static section
  ////////////////////////////////////////////////////////////////////////////////////////////

  Random random = new Random();


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  List of remote references contained in the system
   */
  private ArrayList   refs;

  /**
   *  Factory used to generate invocation identifiers
   */
  private IIDFactory  generator;

  /**
   *  Initially empty method table containing info related to methods.
   *  When methods are called, information are cached by this table.
   */
  private ClientMethodTable methodTable;

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Constructs an empty <tt>GroupProxy</tt> with no remote references.
   */
  public ExtGroupHandler2()
  {
    refs = new ArrayList();
    generator = new IIDFactory();
    methodTable = new ClientMethodTable();
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from InvocationHandler interface
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Processes a method invocation on a proxy instance and returns
   * the result.  This method will be invoked on an invocation handler
   * when a method is invoked on a proxy instance that it is
   * associated with.
   */
  public Object invoke(Object proxy, Method m, Object[] args)
    throws Throwable
  {
    /* Verifies that there is at least one server in the group */
    int size = refs.size();
    if (size == 0)
      throw new RemoteException("No server in the group");
    
    /* Generate an invocation identifier for this invocation */
    IID iid = generator.getId();

    /* Obtain a method descriptor for the method */
    MethodInfo method = methodTable.get(m);

    if (method.isMulticast()) {

      /*
       * The method has multicast semantics; the data sent to the dispatcher
       * will not be immediately interpreted, but multicast to the other
       * members of the group. So, we send them as a byte stream which can
       * be immediately multicast without being unmarshaled.
       */

      /* Create a ByteVectorOutputStream containing the arguments */
      ByteVectorOutputStream bout = new ByteVectorOutputStream(256);
      ObjectOutputStream out = new MarshalOutputStream(bout);
      out.writeObject(args);

      /* Try to invoke the method on one of the remote references */
      int start = random.nextInt(size); // start = 0;
      for (int i=0; i < size; i++) {
        try {
          GroupDispatcher dispatcher = (GroupDispatcher) refs.get((i+start)%size);
					// System.out.println("Attempt: " + i + " host " + dispatcher);
					// System.out.println("IID: " + iid);
          if (method.isVoid()) {
            dispatcher.vinvoke(iid, method.getHash(), bout);
            return null;
          } else {
            return dispatcher.invoke(iid, method.getHash(), bout);
          }
        } catch (ClassCastException e) {
          throw new RemoteException("Invalid remote reference");
        } catch (MarshalException e) {
          throw e;
					//} catch (UnmarshalException e) {
          // throw e;
        } catch (RemoteException e) {
          // e.printStackTrace();
          // Do nothing
        } finally {
          generator.ackId(iid);
        }
      }


    } else {

      /*
       * The method has anycast semantics; the data sent to the dispatcher
       * will be immediately interpreted, so we can send them as an object
       * array that is unmarshaled at the receipt.
       */

      /* Try to invoke the method on one of the remote references */
      int start = random.nextInt(size);
      for (int i=0; i < size; i++) {
        try {
          GroupDispatcher dispatcher = (GroupDispatcher) refs.get((i+start)%size);
          if (method.isVoid()) {
            dispatcher.vinvoke(iid, method.getHash(), args);
            return null;
          } else {
            return dispatcher.invoke(iid, method.getHash(), args);
          }
        } catch (ClassCastException e) {
          throw new RemoteException("Invalid remote reference");
        } catch (MarshalException e) {
          throw e;
        } catch (UnmarshalException e) {
          throw e;
        } catch (RemoteException e) {
          // Do nothing
        } finally {
          generator.ackId(iid);
        }
      }

    }
    throw new RemoteException("Group unreachable");
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Inserts a new remote reference in the group. Note that a group
   *  handler does not check that remote references inserted in it
   *  have all the same interfaces. This has to be performed at an
   *  higher layer.
   */
  public void insert(Remote remote)
  {
    /* Add the remote ref to the reference set */
    refs.add(remote);
  }


  /**
   *
   */
  public int size()
  {
    return refs.size();
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Externalization
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Write the content of this ExtGroupHandler2 on stream <tt>out</tt>.
   */
  public void writeExternal(ObjectOutput out)
    throws IOException
  {
    int size = refs.size();
    out.writeInt(size);
    for (int i=0; i < size; i++)
      out.writeObject(refs.get(i));
  }

  /**
   *  Read the content of this ExtGroupHandler2 from stream <tt>in</tt>.
   */
  public void readExternal(ObjectInput in)
    throws IOException, ClassNotFoundException
  {
    int size = in.readInt();
    for (int i=0; i < size; i++)
      insert((Remote) in.readObject());
  }

} // END ExtGroupHandler2
