/*
 * Copyright (c) 1998-2000 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */


package relacs.rmi;


import java.io.*;
import java.net.*;
import java.lang.reflect.*;
import java.util.*;

import jgroup.*;
import jgroup.multicast.*;


import relacs.general.*;


final class     IIDEntry
extends         OrderedList
implements      AckListener
{


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////


  private static final int START      = 0;
  private static final int WAITING    = 1;
  private static final int EXECUTED   = 2;
  private static final int COMPLETED  = 3;



  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Boolean array for results */
  private boolean[] completed;
  
  /** Counter */
  private int missing;

  /** Unicast result to be returned */
  private Object result;
  
  /** Identifier of this server */
  private MemberId    me;

  /** Status of the invocation */
  private int         status;
  
  /** True if the result is exceptional */
  private boolean     isException;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////


  /**
   *  Build an empty list
   */
  IIDEntry()
  {
    super();
  }


  /**
   *  Build an entry
   */
  IIDEntry(IID id, MemberId me)
  {
    key         = id.getCounter();
    this.me     = me;
    status      = START;
    result      = null;
    isException = false;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////


  /**
   *  Invoke a method (if the method has not already completed)
   *  following the anycast invocation semantics.
   */
  synchronized void anycastInvoke(Object server, Method m, Object[] args)
    throws Exception
  {
    if (status == START) {
      try {
        result = m.invoke(server, args);
      } catch (InvocationTargetException e) {
        try {
          throw (Exception) e.getTargetException();
        } catch (ClassCastException ex) {
          throw e;
        }
      }
      status = COMPLETED;
    }
  }


  /**
   *  Returns the result of an anycast invocation; this method must
   *  be invoked after the invocation of <tt>anycastInvoke</tt>. The
   *  result is returned immediately, as method <tt>anycastInvoke</tt>
   *  has already obtained a result. This method does not need to
   *  be synchronized.
   */
  Object anycastResult()
  {
    return result;
  }


  /**
   *  Multicasts a message containing the message (if this has not already
   *  been performed) following the multicast invocation semantics.
   */
  synchronized void multicastInvoke(MulticastService multicastService, OutputStream message)
    throws Exception
  {
    if (status == START) {
      multicastService.mcast(message, this);
      status = WAITING;
    }
  }


  /**
   *  Returns the result of a multicast invocation; this method must
   *  be invoked after the invocation of <tt>multicastInvoke</tt>. The
   *  result may not be immediately available; in this case, we need
   *  to wait.
   */
  synchronized Object multicastResult()
    throws Exception
  {
    while (status != COMPLETED) {
      try { wait(); } catch (InterruptedException e) {}
    }
    if (isException)
      throw (Exception) result;
    return result;
  }


  /**
   */
  synchronized void executeMulticastInvocation(Object server, Method m, Object[] args)
  {
    if (status != EXECUTED && status != COMPLETED) {
      try {
        result = m.invoke(server, args);
      } catch (InvocationTargetException e) {
        isException = true;
        try {
          result = e.getTargetException();
        } catch (Exception ex) {
          result = e;
        }
      } catch (Exception e) {
        isException = true;
        result = e;
      }
      status = EXECUTED;
      if (missing == 0) {
        // Special case: viewChange occurred
        status = COMPLETED;
        notifyAll();
      }
    }
  }


  /**
   * Upcalls invoked by Jgroup to notify the ack listener
   * about the view in which the multicast message is
   * delivered. Note that following the Jgroup specification,
   * a message may be delivered in a view different from the
   * one in which the message has been multicast; this is
   * motivated by the fact that during agreement protocols,
   * no new messages can be delivered; in fact, cascading
   * sequences of delivered messages could prevent the
   * completion of agreement protocols.
   *
   * @param view the view in which the multicast message is
   *   delivered
   */
  synchronized public void notifyView(View view)
  {
    MemberId[] members = view.getMembers();
    missing = members.length;
    completed = new boolean[missing];
    for (int i=0; i < missing; i++) 
      completed[i] = false;
  }



  /**
   * Upcalls invoked by Jgroup to notify the ack listener
   * that the member identified by <tt>id</tt> has responded
   * with <tt>obj</tt> to the multicast message associated with
   * this ack listener. <p>
   *
   * Member <tt>id</tt> is contained in the view notified by
   * <tt>notifyView</tt>.
   *
   * @param id the identifier of the member responding to the
   *   multicast message
   * @param obj the response message
   */
  synchronized public void ack(MemberId id, int vpos, Object obj)
  {
    // System.out.println("Ack: Id " + id + " missing " + missing + " status " + status);
    if (!completed[vpos]) {
      completed[vpos] = true;
      missing--;
    }
    if (missing == 0 && (status == EXECUTED || status == COMPLETED)) {
      status = COMPLETED;
      notifyAll();
    }
  }


  /**
   * Upcalls invoked by Jgroup to notify the ack listener
   * that a new view has been installed after the view
   * in which the message has been delivered. After this
   * event, no missing response messages for the multicast
   * message associated to this ack listener will ever
   * be received.
   */
  synchronized public void viewChange()
  {
    for (int i=0; i < completed.length; i++) 
      completed[i] = true;
    missing = 0;
    if (status == EXECUTED || status == COMPLETED) {
      status = COMPLETED;
      notifyAll();
    }
  }
  
} // END IIDEntry
