/*
 * Copyright (c) 1998-2000 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package relacs.rmi;

import java.io.*;
import java.lang.reflect.*;
import java.util.*;
import java.security.*;

import jgroup.*;

/**
 * Helper class usueful to associate a "method hash" to each of the methods
 * contained in a set of interfaces. The method hash is a long containing
 * the first 64 bits of the SHA digest from the UTF encoded string of the
 * method name and descriptor. The table maintains two hash table, one
 * for methods and another for hash numbers.
 *
 * @author  Alberto Montresor (Partly copied from Java 2 SDK 1.3)
 * @since   Jgroup 0.9
 */
final class ServerMethodTable
{
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  private HashMap   table;
  private HashMap   invertedTable;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Create a method table that contains, for each remote method in the
   * given interfaces, a mapping from its java.lang.reflect.Method object
   * to its corresponding MethodInfo object.
   */
  ServerMethodTable(Class[] interfaces)
  {
    table = new HashMap();
    invertedTable = new HashMap();
    for (int i = 0; i < interfaces.length; i++) {
      Method methods[] = interfaces[i].getMethods();
      for (int j = 0; j < methods.length; j++) {
        String name = getMethodNameAndDescriptor(methods[j]);
        Long hashLong = new Long(computeInfo(methods[j]));
        table.put(hashLong, methods[j]);
        invertedTable.put(name, hashLong);
      }
    }
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *
   */
  Method get(long hash)
  {
    Long hashLong = new Long(hash);
    return (Method) table.get(hashLong);
  }

  /**
   *
   */
  long get(Method m)
  {
    Long hashLong = (Long) invertedTable.get(getMethodNameAndDescriptor(m));
    return hashLong.longValue();
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  static long computeInfo(Method method)
  {
    final Method m = method;

    /*
     * Set this Method object to override language
     * access checks so that the dispatcher can invoke
     * methods from non-public remote interfaces.
     */
    AccessController.doPrivileged(new PrivilegedAction()
      {
        public Object run() {
          m.setAccessible(true);
          return null;
      }
    });

    return computeHash(m);
  }

  /**
   * Compute the "method hash" of a remote method.  The method hash
   * is a long containing the first 64 bits of the SHA digest from
   * the UTF encoded string of the method name and descriptor.
   */
  static long computeHash(Method m)
  {
    long hash = 0;
    ByteArrayOutputStream sink = new ByteArrayOutputStream(127);
    try {
      MessageDigest md = MessageDigest.getInstance("SHA");
      DataOutputStream out = new DataOutputStream(
        new DigestOutputStream(sink, md));

      String s = getMethodNameAndDescriptor(m);

      out.writeUTF(s);

      // use only the first 64 bits of the digest for the hash
      out.flush();
      byte hasharray[] = md.digest();
      for (int i = 0; i < Math.min(8, hasharray.length); i++) {
        hash += ((long) (hasharray[i] & 0xFF)) << (i * 8);
      }
    } catch (IOException ignore) {
      /* can't happen, but be deterministic anyway. */
      hash = -1;
    } catch (NoSuchAlgorithmException complain) {
      throw new SecurityException(complain.getMessage());
    }
    return hash;
  }


  /**
   * Return a string consisting of the given method's name followed by
   * its "method descriptor", as appropriate for use in the computation
   * of the "method hash".
   *
   * See section 4.3.3 of The Java Virtual Machine Specification for
   * the definition of a "method descriptor".
   */
  static String getMethodNameAndDescriptor(Method m) {
    StringBuffer desc = new StringBuffer(m.getName());
    desc.append('(');
    Class[] paramTypes = m.getParameterTypes();
    for (int i = 0; i < paramTypes.length; i++) {
      desc.append(getTypeDescriptor(paramTypes[i]));
    }
    desc.append(')');
    Class returnType = m.getReturnType();
    if (returnType == void.class) { // optimization: handle void here
      desc.append('V');
    } else {
      desc.append(getTypeDescriptor(returnType));
    }
    return desc.toString();
  }

  /**
   * Get the descriptor of a particular type, as appropriate for either
   * a parameter or return type in a method descriptor.
   */
  static String getTypeDescriptor(Class type)
  {
    if (type.isPrimitive()) {
      if (type == int.class) {
        return "I";
      } else if (type == boolean.class) {
        return "Z";
      } else if (type == byte.class) {
        return "B";
      } else if (type == char.class) {
        return "C";
      } else if (type == short.class) {
        return "S";
      } else if (type == long.class) {
        return "J";
      } else if (type == float.class) {
        return "F";
      } else if (type == double.class) {
        return "D";
      } else if (type == void.class) {
        return "V";
      } else {
        throw new Error("unrecognized primitive type: " + type);
      }
    } else if (type.isArray()) {
      /*
       * According to JLS 20.3.2, the getName() method on Class does
       * return the VM type descriptor format for array classes (only);
       * using that should be quicker than the otherwise obvious code:
       *
       *     return "[" + getTypeDescriptor(type.getComponentType());
       */
      return type.getName().replace('.', '/');
    } else {
      return "L" + type.getName().replace('.', '/') + ";";
    }
  }

} // END ServerMethodTable

