/*
 * Copyright (c) 1998-2000 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package relacs.rmi;

import java.io.*;
import java.net.*;

import jgroup.*;
import relacs.*;

/**
 *  <code>VMIDImpl</code> is the implementation of the <code>VMID</code> class.
 *
 *  @author Alberto Montresor
 *  @since    Jgroup 0.7
 */
public final class  VMIDImpl
implements          VMID, Externalizable
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  private InetAddress     inetAddress;    // IP address of the generating host
  private int             address;        // Hash value
  private Integer         wrappedAddress; // Wrapped address
  private int             hash;           // "Random" number to distinguish concurrent JVM
  private long            time;           // Time when this <code>VMIDImpl</code> has been
                                          // generated


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static section
  ////////////////////////////////////////////////////////////////////////////////////////////

  private static VMIDImpl vmid;         // Unique <code>VMIDImpl</code> instantiated

  /*
   *  Create a singleton instance of <code>VMIDImpl</code>.
   */
  static {
    try {
      InetAddress inetAddress = RelacsDS.getLocalHost();
      long time = System.currentTimeMillis();
      int hash = (new Object()).hashCode();
      vmid = new VMIDImpl(inetAddress, time, hash);
    } catch (Exception e) {
      e.printStackTrace();
      System.exit(0);
    }
  }


  /**
   *  Returns a <code>VMIDImpl</code> for the local Java virtual machine.
   */
  public static VMIDImpl getLocalVMID()
  {
    return vmid;
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Private default constructor.
   */
  public VMIDImpl()
  {
  }

  /**
   *
   */
  VMIDImpl(InetAddress address, long time, int hash)
  {
    this.inetAddress = address;
    this.time = time;
    this.hash = hash;
    byte[] temp = inetAddress.getAddress();
    if (temp.length != 4)
      throw new IllegalArgumentException("Jgroup accepts only 32-bit IP addresses");
    this.address = ((temp[0] << 24) & 0xFF000000) | ((temp[1] << 16) & 0xFF0000) |
      ((temp[2] << 8) & 0xFF00) | (temp[3] & 0xFF);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns the address of the host in which the virtual machine
   *  identified by this <code>VMIDImpl</code> is executed.
   */
  public InetAddress getAddress()
    throws UnknownHostException
  {
    if (inetAddress == null) {
      byte[] temp = new byte[4];
      temp[0] = (byte) (address >>> 24);
      temp[1] = (byte) (address >>> 16);
      temp[2] = (byte) (address >>>  8);
      temp[3] = (byte) (address >>>  0);
      inetAddress = InetAddress.getByName(temp[0] + "." + temp[1]  + "." +
                                          temp[2] + "." + temp[3] );
    }

    return inetAddress;
  }


  /**
   *  Returns the hashed address of the host in which the virtual machine
   *  identified by this <code>VMIDImpl</code> is executed.
   */
  public Object getHashAddress()
  {
    if (wrappedAddress == null)
      wrappedAddress = new Integer(address);
    return wrappedAddress;
  }


  /**
   *  Returns the time at which this <code>VMIDImpl</code> has been created.
   */
  public long getTime()
  {
    return time;
  }


  /**
   *  Returns a string representation of the object.
   */
  public String toString()
  {
    return "["+address+","+time+","+hash+"]";
  }


  /**
   * Returns a hashcode for the <code>VMIDImpl</code>.  Two <code>VMIDImpl</code>s
   * will have the same hashcode if they are equal with respect to their
   * content.
   *
   * @return the hashcode
   */
  public int hashCode() {
    return address + (int) time + (int) hash;
  }

  /**
   * Compares two <code>VMIDImpl</code>s for content equality.
   *
   * @param obj the Object to compare with
   * @return  true if these Objects are equal; false otherwise.
   */
  public boolean equals(Object obj) {
    if ((obj != null) && (obj instanceof VMIDImpl)) {
      VMIDImpl vmid = (VMIDImpl)obj;
      return (hash == vmid.hash && address == address && time == vmid.time);
    } else {
      return false;
    }
  }


  /**
   *  Write the content of this VMID on stream <tt>out</tt>.
   *  Note that the InetAddress is not serialized, because too
   *  inefficient. If requested, the InetAddress is retrieved
   *  through DNS lookup in method <tt>getAddress</tt>.
   */
  public void writeExternal(ObjectOutput out)
    throws IOException
  {
    out.writeInt(hash);
    out.writeLong(time);
    out.writeInt(address);
  }

  /**
   *  Read the content of this VMID from stream <tt>in</tt>.
   */
  public void readExternal(ObjectInput in)
    throws IOException, ClassNotFoundException
  {
    hash = in.readInt();
    time = in.readLong();
    address = in.readInt();
  }

} // END VMIDImpl
