/*
 * Copyright (c) 1998-2000 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package relacs.rmi.registry;

import java.io.*;
import java.rmi.*;
import java.rmi.server.*;
import java.rmi.registry.*;
import java.net.*;

import jgroup.*;
import jgroup.registry.*;
import relacs.gm.*;
import relacs.rmi.*;

import debug.Debug;

/**
 *
 *
 * @author  Alberto Montresor
 * @since   Jgroup 0.4
 */
public final class  RegistryImpl
implements          DependableRegistry, MergingListener, MembershipListener
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  private static final int DRGID = 30000;                       // Group identifier


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  private ExternalGMIService  ExternalGMIService;
  private MembershipService   membershipService;
  private RegistryTable       registrytable;
  private MemberId            me;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Creates and exports a dependable registry replica on the local host that
   * listen to incoming requests on an anonymous communication port.
   * Objects contained <tt>conf</tt> are used by Jgroup to configure the
   * group communication; the distributed system description must be
   * inclued in this array.
   *
   * @param conf        configuration information
   * @return            the new dependable registry replica.
   * @exception RemoteException if the registry could not be exported
   * @exception JgroupException
   *            Generic exception raised by Jgroup.
   * @exception DSException
   *            raised when the distributed system configuration object contains an error
   */
  public RegistryImpl(Object[] conf)
      throws RemoteException, JgroupException
  {
    initRegistry(conf, 0);
  }


  /**
   * Creates and exports a dependable registry replica on the local host that
   * listen to incoming requests on  the given <code>port</code> communication
   * port.
   * Objects contained <tt>conf</tt> are used by Jgroup to configure the
   * group communication; the distributed system description must be
   * inclued in this array.
   *
   * @param conf        configuration information
   * @param port        port used for RMI communications
   * @return            the new dependable registry replica.
   * @exception RemoteException if the registry could not be exported
   * @exception JgroupException
   *            Generic exception raised by Jgroup.
   * @exception DSException
   *            raised when the distributed system configuration object contains an error
   */
  public RegistryImpl(Object[] conf, int port)
      throws RemoteException, JgroupException
  {
    initRegistry(conf, port);
  }


  /**
   * Creates and exports a dependable registry replica on the local host that
   * listen to incoming requests on  the given <code>port</code> communication
   * port.
   * Objects contained <tt>conf</tt> are used by Jgroup to configure the
   * group communication; the distributed system description must be
   * inclued in this array.
   * The new instance listens to incoming requests using a
   * <code>ServerSocket</code> created from the supplied
   * <code>RMIServerSocketFactory</code>.  A client
   * that receives a reference to this registry will use a
   * <code>Socket</code> created from the supplied
   * <code>RMIClientSocketFactory</code>.
   *
   * @param conf        configuration information
   * @param port        port used for RMI communications
   * @param csf         client-side <code>Socket</code> factory used to
   *                    make connections to the registry
   * @param ssf         server-side <code>ServerSocket</code> factory
   *                    used to accept connections to the registry
   * @return            the new dependable registry replica.
   * @exception RemoteException if the registry could not be exported
   * @exception JgroupException
   *            Generic exception raised by Jgroup.
   * @exception DSException
   *            raised when the distributed system configuration object contains an error
   */
  public RegistryImpl(Object[] conf, int port, RMIClientSocketFactory csf, RMIServerSocketFactory ssf)
      throws RemoteException, JgroupException
  {
    initRegistry(conf, port, csf, ssf);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Main
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Main method of the class
   */
  public static void main(String argv[])
  {
    // Default values
    String hostfile = null;
    int jport = DREGISTRY_JGROUP_PORT;
    int port = 0;

    // Read arguments
    for (int i = 0 ; i < argv.length ; i++) {
      if  (argv[i].equals("-jport")) {
        if ((i + 1) == argv.length) {
          System.err.println("Options -jport must be followed by a port number");
          usage();
          return;
        }
        try {
          jport = Integer.parseInt(argv[++i]);
        } catch (Exception e) {
          System.err.println("Value " + argv[i] + " is not a number");
          usage();
          return;
        }
      } else if (argv[i].equals("-rport")) {
        if ((i + 1) == argv.length) {
          System.err.println("Options -rport must be followed by a port number");
          usage();
          return;
        }
        try {
          port = Integer.parseInt(argv[++i]);
        } catch (Exception e) {
          System.err.println("Value " + argv[i] + " is not a number");
          usage();
          return;
        }
      } else if (argv[i].startsWith("-")) {
        System.err.println("Unknown option " + argv[i]);
        usage();
      } else {
        if (hostfile != null) {
          System.err.println("Hostfile already specified");
          usage();
          return;
        }
        hostfile = argv[i];
      }
    }
    if (hostfile == null) {
      System.err.println("You must specify an hostfile");
      usage();
      return;
    }

    // Start registry
    try {
      DistributedSystem ds = new DistributedSystem(hostfile);
      Object[] conf = new Object[] { ds, new TransportConfig(jport) };
      DependableRegistry registry = new RegistryImpl(conf, port);
      while (true) {
        try { Thread.sleep(0x7fffffffffffffffL); } catch(InterruptedException ex) {}
      }
    } catch(Exception exception) {
      Debug.println("RegistryImpl.main: an exception occurred: " + exception.getMessage());
      exception.printStackTrace();
    }
  }

  static void usage()
  {
    System.err.println("dregistry <hostfile> [-jport <port number>] [-port <port number>]");
    System.exit(0);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from DependableRegistry
  ////////////////////////////////////////////////////////////////////////////////////////////

  /*
   *  Comments inherited from jgroup.DependableRegistry
   */
  public IID bind(String service, Remote remote, Class serverClass)
    throws RemoteException, McastRemoteException, AccessException
  {
    if (Debug.DEBUG)
      Debug.println(60, "Registry.bind(): " + service + " " + remote);
    checkAccess("Registry.bind");
    IID iid = ExternalGMIService.getIdentifier();
    BindingList bl = new BindingList(iid.getCounter(), service, remote, serverClass);
    registrytable.add((VMIDImpl) iid.getVmid(), bl);
    return iid;
  }


  /*
   *  Comments inherited from jgroup.DependableRegistry
   */
  public void unbind(IID iid)
    throws RemoteException, McastRemoteException, NotBoundException, AccessException
  {
    if (Debug.DEBUG)
      Debug.println(60, "Registry.unbind "+ iid);
    checkAccess("Registry.unbind");
    registrytable.remove((VMIDImpl) iid.getVmid(), iid.getCounter());
  }


  /*
   *  Comments inherited from jgroup.DependableRegistry
   */
  public String[] list()
    throws RemoteException
  {
    if (Debug.DEBUG)
      Debug.println(60, "Registry.list()");
    return registrytable.list();
  }

  /*
   *  Comments inherited from jgroup.DependableRegistry
   */
  public Remote lookup(String service)
    throws RemoteException, NotBoundException
  {
    ServiceData         sdata;          // Service data

    if (Debug.DEBUG)
      Debug.println(60, "Registry.lookup: " + service);
    return registrytable.getStub(service);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Upcalls
  ////////////////////////////////////////////////////////////////////////////////////////////

  /*
   *  Comments inherited from jgroup.MembershipListener
   */
  public void viewChange(View view)
  {
    if (Debug.DEBUG) {
      MemberId[] members = view.getMembers();
      Debug.println(60, "View: Vid" + view.getVid());
      for (int i=0; i<members.length; i++)
        Debug.println(60, "View: member[" + i + "]" + members[i] + " ");
    }
    registrytable.viewChange(view);
  }


  /*
   *  Comments inherited from jgroup.MembershipListener
   */
  public Object getState(MemberId[] dests)
  {
    if (Debug.DEBUG) {
      for (int i=0; i < dests.length; i++)
        Debug.println(60, "getState dests["+i+"]: "+dests[i]);
    }
    return registrytable.getState(dests, me);
  }


  /*
   *  Comments inherited from jgroup.MergingListener
   */
  public void putState(Object obj, MemberId[] sources)
  {
    if (Debug.DEBUG) {
      for (int i=0; i < sources.length; i++)
        Debug.println(60, "putState sources["+i+"]: "+sources[i]);
    }
    try {
      registrytable.putState((Status) obj, sources);
    } catch (Exception e) {
      System.err.println("Error when receiving a putState invocation");
      e.printStackTrace();
    }
  }


  /*
   *  Comments inherited from jgroup.MergingListener
   */
  public void leaved()
  {
  }


  /*
   *  Comments inherited from jgroup.MembershipListener
   */
  public void prepareChange()
  {
  }



  ////////////////////////////////////////////////////////////////////////////////////////////
  // Private methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  private static void checkAccess(String s)
    throws AccessException
  {
    // TODO: Currently, bind is always permitted; we need to think if it is necessary to
    // restrict the access to the hosts in the configuration file.
    return;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // RMI related methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *
   */
  private void initRegistry(Object[] conf, int port)
    throws RemoteException, JgroupException
  {
    Registry    reg = null;

    registrytable = new RegistryTable();
    Object[] listeners = new Object[] { this };
    GroupManager gm = GroupManager.getGroupManager(conf, listeners);
    ExternalGMIService = (ExternalGMIService) gm.getService(ExternalGMIService.class);
    membershipService = (MembershipService) gm.getService(MembershipService.class);
    me = membershipService.getMyIdentifier();
    if (Debug.DEBUG) {
      Debug.println(60, "Obtained identifier: " + me);
    }
    membershipService.join(DRGID);
    if (port == 0)
      port = Registry.REGISTRY_PORT+1;
    try {
      reg = LocateRegistry.getRegistry(port);
      reg.rebind(DEPENDABLE_REGISTRY_NAME, (GroupDispatcher) ExternalGMIService);
    } catch (Exception e) {
      reg = LocateRegistry.createRegistry(port);
      reg.rebind(DEPENDABLE_REGISTRY_NAME, (GroupDispatcher) ExternalGMIService);
    }

  }

  /**
   *
   */
  private void initRegistry(Object[] conf, int port, RMIClientSocketFactory csf, RMIServerSocketFactory ssf)
    throws RemoteException, JgroupException
  {
    Registry    reg = null;

    registrytable = new RegistryTable();
    Object[] listeners = new Object[] { this };
    GroupManager gm = GroupManager.getGroupManager(conf, listeners);
    ExternalGMIService = (ExternalGMIService) gm.getService(ExternalGMIService.class);
    membershipService = (MembershipService) gm.getService(MembershipService.class);
    membershipService.join(DRGID);
    me = membershipService.getMyIdentifier();
    if (Debug.DEBUG)
      Debug.println(60, "Obtained identifier: " + me);
    if (port == 0)
      port = Registry.REGISTRY_PORT+1;
    try {
      reg = LocateRegistry.getRegistry(null, port, csf);
      reg.rebind(DEPENDABLE_REGISTRY_NAME, (GroupDispatcher) ExternalGMIService);
    } catch (RemoteException e) {
      reg = LocateRegistry.createRegistry(port, csf, ssf);
      reg.rebind(DEPENDABLE_REGISTRY_NAME, (GroupDispatcher) ExternalGMIService);
    }
  }

} // END RegistryImpl



