/*
 * Copyright (c) 1998-2000 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package relacs.rmi.registry;

import java.util.*;
import java.rmi.*;
import java.io.*;

import jgroup.*;
import relacs.general.*;
import relacs.rmi.*;
import debug.Debug;

/**
 *
 *
 * @author  Alberto Montresor
 * @since   Jgroup 0.5
 */
final class     RegistryTable
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  private Hashtable     servertable;
  private Hashtable     servicetable;
  private View          view;

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  RegistryTable()
  {
    servertable = new Hashtable(101);
    servicetable = new Hashtable(101);
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////


  /**
   *
   */
  synchronized void add(VMIDImpl vmid, BindingList bl)
    throws RemoteException
  {
    ServiceData       service;
    ServerData        server;

    // Inserts in service table
    service = (ServiceData) servicetable.get(bl.servicename);
    if (service == null) {
      service = new ServiceData(bl.serverClass);
      servicetable.put(bl.servicename, service);
    }
    service.insert(bl);

    // Insert in server table
    server = get(vmid);
    server.add(bl.key, bl);
  }

  /**
   *
   */
  synchronized void remove(VMIDImpl vmid, int key)
  {
    ServerData        server;
    ServiceData       service;
    BindingList       bl;

    server = get(vmid);
    bl = server.remove(key);
    if (bl != null) {
      service = (ServiceData) servicetable.get(bl.servicename);
      if (service != null) {
        service.remove(bl);
        if (service.isEmpty())
          servicetable.remove(bl.servicename);
      }
    }
  }


  /**
   *
   */
  synchronized void viewChange(View view)
  {
    ServerData    server;

    this.view = view;
    for (Enumeration e = servertable.elements(); e.hasMoreElements();) {
      server = (ServerData) e.nextElement();
      server.viewChange(view);
    }
  }


  /**
   *
   */
  synchronized String[] list()
  {
    String[] list = new String[servicetable.size()];
    Enumeration enumeration = servicetable.keys();
    for (int i=0; i < list.length; i++)
      list[i] = (String) enumeration.nextElement();
    return list;
  }


  /**
   *
   */
  synchronized Remote getStub(String servicename)
    throws RemoteException, NotBoundException
  {
    ServiceData service = (ServiceData) servicetable.get(servicename);
    if (service == null)
      throw new NotBoundException(servicename);
    return service.getStub();
  }


  /**
   *
   */
  synchronized Status getState(MemberId[] dests, MemberId me)
  {
    ServerData  server;
    Status      status = new Status();
    int         size = servertable.size();

    status.vmids = new VMIDImpl[size];
    status.removed = new IntList[size];
    status.bindings = new BindingList[size][];

    int i=0;
    for (Enumeration e = servertable.elements() ; e.hasMoreElements() ; i++) {
      server = (ServerData) e.nextElement();
      status.vmids[i] = server.vmid;
      status.removed[i] = server.removed;
      status.bindings[i] = server.getState(dests, me);
    }
    return status;
  }

  synchronized void putState(Status status, MemberId[] dests)
    throws RemoteException, ClassNotFoundException
  {
    ServerData      server;
    for (int i=0; i < status.vmids.length; i++) {
      if (Debug.DEBUG)
        Debug.println(60, "putState: status.vmids["+i+"]: " + status.vmids[i]);
      server = (ServerData) get(status.vmids[i]);
      server.putState(servicetable, status.removed[i], status.bindings[i], dests);
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Private methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  private ServerData get(VMIDImpl vmid)
  {
    ServerData          server;

    server = (ServerData) servertable.get(vmid);
    if (server == null) {
      server = new ServerData(vmid);
      server.viewChange(view);
      servertable.put(vmid, server);
    }
    return server;
  }

}
