/*
 * Copyright (c) 1998-2000 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package test.hello;

import java.io.*;
import java.net.*;
import java.rmi.server.*;
import java.rmi.*;

import jgroup.*;
import jgroup.registry.*;

import debug.Debug;

/**
 * Simple Hello server aimed at demonstrating the capabilities of both
 * internal and external invocations.
 *
 * @author  Alberto Montresor
 * @since   Jgroup 1.0
 */
public class HelloServer
implements Hello, InternalHello, MembershipListener
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Stores the answer for external invocatins */
  private Answer answer;

  /** Service proxies */
  private MembershipService   membershipService;
  private ExternalGMIService  externalGMIService;
  private InternalHello       internalHello;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor and main
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Instantiates an HelloServer object.
   */
  public static void main(String argv[])
    throws Exception
  {
    String hostfile = null;
    int jport = 0;
    int rport = 0;

    for (int i = 0 ; i < argv.length ; i++) {
      if  (argv[i].equals("-jport")) {
        if ((i + 1) == argv.length) {
          System.err.println("Options -jport must be followed by a port number");
          usage();
          return;
        }
        try {
          jport = Integer.parseInt(argv[++i]);
        } catch (Exception e) {
          System.err.println("Value " + argv[i] + " is not a number");
          usage();
          return;
        }
      } else if (argv[i].equals("-rport")) {
        if ((i + 1) == argv.length) {
          System.err.println("Options -rport must be followed by a port number");
          usage();
          return;
        }
        try {
          rport = Integer.parseInt(argv[++i]);
        } catch (Exception e) {
          System.err.println("Value " + argv[i] + " is not a number");
          usage();
          return;
        }
      } else if (argv[i].startsWith("-")) {
        System.err.println("Unknown option " + argv[i]);
        usage();
      } else {
        if (hostfile != null) {
          System.err.println("Hostfile already specified");
          usage();
          return;
        }
        hostfile = argv[i].trim();
      }
    }
    if (hostfile == null) {
      System.err.println("You must specify an hostfile");
      usage();
      return;
    }
    HelloServer obj = new HelloServer("Hello from \""+InetAddress.getLocalHost()+"\"", hostfile, jport, rport);
  }

  private static void usage()
  {
    System.out.println("Usage: helloserver <hostfile>  [-jport <port number>] [-rport <port number>]");
  }

  /**
   *  Constructor invoked by main
   */
  public HelloServer(String message, String hostfile, int jport, int rport)
    throws Exception
  {

    /*
     * Prepare answer
     */
    answer = new Answer();
    answer.message = message;

    /*
     * Obtain the group manager
     */
    Object[] configuration = new Object[] { new DistributedSystem(hostfile), 
      new TransportConfig(jport) };
    Object[] listeners = new Object[] { this };
    GroupManager gm = GroupManager.getGroupManager(configuration, listeners);

    /*
     * Obtain the service proxies
     */
    membershipService = (MembershipService) gm.getService(MembershipService.class);
    externalGMIService = (ExternalGMIService) gm.getService(ExternalGMIService.class);
    internalHello = (InternalHello) gm.getService(InternalHello.class);

    /*
     * Join the group and bind the server in the dependable registry
     */
    membershipService.join(12);
    DistributedSystem ds = new DistributedSystem(hostfile);
    DependableRegistry dr = RegistryFactory.getRegistry(ds, rport);
    externalGMIService.bind("Jgroup/HelloServer", dr);
    Debug.println("Server ready and bound to the reliable registry");
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Hello
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  External method
   */
  public Answer sayHello()
    throws RemoteException
  {
    Debug.println("sayHello: returning " + answer.message);
    return answer;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from InternalHello
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Internal method
   */
  public Object time()
    throws RemoteException
  {
    long tmp = System.currentTimeMillis();
    Debug.println("Generating time: " + tmp);
    return new Long(tmp);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from MembershipListener
  ////////////////////////////////////////////////////////////////////////////////////////////

  public void viewChange(View v)
  {
    Debug.println("View id: " + v.getVid());
    MemberId[] members = v.getMembers();
    for (int i=0; i < members.length; i++)
      Debug.println("Member " + i + ": " + members[i]);
    try {
      Object[] objs = (Object[]) internalHello.time();
      for (int i=0; i< objs.length; i++)
        Debug.println("Time: " + objs[i]);
      answer.time = objs;
    } catch (Exception e) {
      e.printStackTrace();
    }
  }

  public void prepareChange()
  {
  }

  public void leaved()
  {
  }

} // END HelloServer
