/*
 * Copyright (c) 1998-2000 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package test.performance;

import java.rmi.*;
import java.rmi.registry.*;
import java.util.*;


import jgroup.*;
import relacs.*;
import test.performance.*;

/**
 * Console used to execute performance tests.
 *
 * @author  Alberto Montresor
 * @since   Jgroup 1.1
 */
public class Console
{

	/**
   *  Prints the usage of this command
   */
	static void usage()
	{
		System.out.println("Usage: java test.performance.Console [-kill] <ds file>");
		System.exit(0);
	}

  public static void main(String[] args)
    throws Exception
  {
    // Check arguments
    if (args.length == 0) {
			usage();
    }

		if (args[0].equals("-kill")) {
			if (args.length == 2) {
				RelacsDS ds = new RelacsDS(args[1]);
				String[] names = ds.getNames();
				Executor[] executors = getExecutors(names);
				for (int i=0; i < executors.length; i++)
					executors[i].halt();
			} else {
				usage();
			}
		} else {
			RelacsDS ds = new RelacsDS(args[0]);
			String[] names = ds.getNames();
			Executor[] executors = getExecutors(names);
			
			if (executors.length < 3) {
				System.out.println("Test failed: not enough executors available");
				System.exit(0);
			}
			test0(executors, names, ds);
			test1(executors, names, ds);
		}
  }

  public static Executor[] getExecutors(String[] names)
    throws Exception
  {

    ArrayList list = new ArrayList();
    for (int i = 0; i < names.length; i++) {
      try {
        System.out.print("Contacting " + names[i] + ": ");
        Object server = Naming.lookup("rmi://"+names[i]+":"+(Registry.REGISTRY_PORT+1)+
                                      "/ExecServer");
        list.add(server);
        System.out.println("success");
      } catch (Exception e) {

        System.out.println("failed: " + e.getMessage());
      }
    }
    Executor[] ret = new Executor[list.size()];
    return (Executor[]) list.toArray(ret);
  }

  /**
   * Testing scenario:
   * 1 registry, n client, 1 servers
   *
   * Tests performed:
   *   standard rmi invocation, 1 server - n client
   */
  static void test0(Executor[] executors, String[] names, DistributedSystem ds)
    throws Exception
  {
    Object[] args;
    Object[] ids = new Object[executors.length];

    args = new Object[0];
    executors[0].start(RMIServer.class, args);

    // Performance test
    args = new Object[] { names[0] };
    for (int i=1; i < executors.length-1; i++) {

        // Log information    
        System.out.println("Test started: N. Server: 1 - N. Clients: " + i);

        // Start clients
        for (int j=0; j < i; j++) 
            ids[j-1] = executors[executors.length-1].start(RMIClient.class, args);

        // Stop clients
        for (int j=0; j < i; j++) 
            executors[j].halt(ids[j]);
            
        // Log information    
        System.out.println("Test ended: N. Server: 1 - N. Clients: " + i);
    }
  }


  /**
   * Testing scenario:
   * 1 registry, 1 client, n servers
   *
   * Tests performed:
   * - anycast void invocations
   *   - single-threaded client
   * - anycast byte[] invocations
   *   - single-threaded client
   * - multicast void invocations
   *   - single-threaded client
   * - multicast byte[] invocations
   *   - multi-threaded client
   */
  static void test1(Executor[] executors, String[] names, DistributedSystem ds)
    throws Exception
  {
    Object[] args;

    // Starts a dependable registry in the first machine
    args = new Object[] { ds };
    executors[0].start(RegistryStarter.class, args);

    // Performance test
    for (int i=1; i < executors.length; i++) {

        // Log information
        System.out.println("Test performed: N. Server: " + i + " N. Clients: 1");

        // Starts a new speed server
        args = new Object[] { ds };
        executors[i].start(SpeedTestServer.class, args);

        // Starts a new speed client
        args = new Object[] { ds, "N. Server: " + i + " N. Clients: 1" };
        Object id = executors[0].start(SpeedTestClient.class, args);

        // Stops the speed client; before being stopped, the speed client waits
        // to complete its task.
        executors[0].halt(id);

        // Log information
        System.out.println("Test completed: N. Server: " + i + " N. Clients: 1");
        
    }

  }

} // END Console


















