/*
 * Copyright (c) 1998-2000 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package test.performance;

import java.io.*;
import java.net.*;
import java.util.*;
import java.rmi.registry.*;


/**
 *  Test for measuring the performance of standard RMI invocations.
 *  Can be executed remotely through the <t>Executor</t> class,
 *  since it implements the <t>Server</t> interface.
 *
 *  @author       Alberto Montresor
 *  @since        Jgroup 1.1
 */
public class RMIClient 
extends Thread
implements Server
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Main
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** 
   * Start client from command line
   */
  public static void main(String[] argv)
    throws Exception
  {
    String hostname = null;
		int port = Registry.REGISTRY_PORT;

    for (int i = 0 ; i < argv.length ; i++) {
      if  (argv[i].equals("-rport")) {
        if ((i + 1) == argv.length) {
          System.err.println("Options -rport must be followed by a port number");
          usage();
          return;
        }
        try {
          port = Integer.parseInt(argv[++i]);
        } catch (Exception e) {
          System.err.println("Value " + argv[i] + " is not a number");
          usage();
          return;
        }
      } else if (argv[i].startsWith("-")) {
        System.err.println("Unknown option " + argv[i]);
        usage();
      } else {
        if (hostname != null) {
          System.err.println("Hostfile already specified");
          usage();
          return;
        }
        hostname = argv[i];
      }
    }

    RMIClient client = new RMIClient();
    Object[] objs = new Object[] { hostname, new Integer(port) };
    client.start(objs);
  }
  
  private static void usage()
  {
    System.out.println("Usage: rmispeedclient <hostname> [-rport <port number>]");
  }
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Remote reference for the RMI server */
  SpeedTest test;

  /** True iff the test is completed */
  boolean completed = false;

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Default constructor, used by the executor service
   * to create a service.
   */
  public RMIClient()
  {
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Method
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Start the service associated with this server object.
   */
  public void start(Object[] args)
    throws Exception
  {
    // Check arguments
    if (args == null || args.length != 2)
      throw new IllegalArgumentException("Arguments required: String");
    String hostname = (String) args[0];
		int port = ((Integer) args[1]).intValue();
    
    // Locate server
    Registry reg = LocateRegistry.getRegistry(hostname);
    test = (SpeedTest) reg.lookup("Jgroup/SpeedTest");
    System.out.println("Obtained stub: " + test);

    this.start();
  }
  
  /*
   * Run method of the thread
   */
  public synchronized void run() 
  {
    try {
      // Open log file
      PrintStream out = null;
      try {
        String name = InetAddress.getLocalHost().toString();
        int t1 = name.indexOf('.');
        int t2 = name.indexOf('/');
        t1 = (t1 < t2 ? t1 : t2);
        GregorianCalendar gc = new GregorianCalendar();
        name = "client." + name.substring(0, t1) + gc.get(GregorianCalendar.YEAR) +
          gc.get(GregorianCalendar.MONTH) + gc.get(GregorianCalendar.DAY_OF_MONTH) +
          "-" + gc.get(GregorianCalendar.HOUR_OF_DAY) + gc.get(GregorianCalendar.MINUTE)+".log";
        out = new PrintStream(new FileOutputStream(name));
      } catch (Exception e) {
        e.printStackTrace();
      }

      // Perform test
      ClientTest.test1(test, out);
      ClientTest.test2(test, out);
  
      // Close log file
      out.close();
      completed = true;
      notify();
    } catch (Exception e) {
      e.printStackTrace();
    }
  }

  /**
   * Stop the service associated with this server object.
   */
  public synchronized void halt()
  {
    while (!completed)
      try { wait(); } catch (Exception e) { }
  }

} // END RMICLient

