/*
 * Copyright (c) 1998-2000 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package test.performance;

import jgroup.*;
import jgroup.registry.*;

import java.io.*;
import java.util.*;
import java.net.*;
import java.rmi.registry.*;

import test.performance.*;

/**
 *  Test for measuring the performance of GMI invocations.
 *  Can be executed remotely through the <t>Executor</t> class,
 *  since it implements the <t>Server</t> interface.
 *
 *  @author       Alberto Montresor
 *  @since        Jgroup 1.1
 */
public class SpeedTestClient
implements Server
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Main
  ////////////////////////////////////////////////////////////////////////////////////////////

  public static void main(String[] argv)
    throws Exception
  {
    String hostfile = null;
    int rport = 0;
    String log = "";

    for (int i = 0 ; i < argv.length ; i++) {
      if  (argv[i].equals("-log")) {
        if ((i + 1) == argv.length) {
          System.err.println("Options -log must be followed by a log file name");
          usage();
          return;
        }
        log = argv[++i];
      } else if (argv[i].equals("-rport")) {
        if ((i + 1) == argv.length) {
          System.err.println("Options -rport must be followed by a port number");
          usage();
          return;
        }
        try {
          rport = Integer.parseInt(argv[++i]);
        } catch (Exception e) {
          System.err.println("Value " + argv[i] + " is not a number");
          usage();
          return;
        }
      } else if (argv[i].startsWith("-")) {
        System.err.println("Unknown option " + argv[i]);
        usage();
      } else {
        if (hostfile != null) {
          System.err.println("Hostfile already specified");
          usage();
          return;
        }
        hostfile = argv[i];
      }
    }
    if (hostfile == null) {
      System.err.println("You must specify an hostfile");
      usage();
      return;
    }

    SpeedTestClient client = new SpeedTestClient();
    DistributedSystem ds = new DistributedSystem(hostfile);
    Object[] objs = new Object[] { ds, log, new Integer(rport) };
    client.start(objs);
  }

  private static void usage()
  {
    System.out.println("Usage: speedclient <hostfile>  [-comment <log comment>] [-rport <port number>]");
  }

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** True when the action requested is completed */
  private boolean completed = false;

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  public SpeedTestClient()
  {
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  public synchronized void start(Object args[])
    throws Exception
  {
    // Check arguments
    if (args == null || args.length != 3)
      throw new IllegalArgumentException("Arguments required: DistributedSystem, String");
    DistributedSystem ds = (DistributedSystem) args[0];
    String comment = (String) args[1];
    int rport = ((Integer) args[2]).intValue();
    

    // Locate server
    DependableRegistry reg = RegistryFactory.getRegistry(ds, rport);
    SpeedTest test = (SpeedTest) reg.lookup("Jgroup/SpeedTest");

    // Open log file
    PrintStream out = null;
    try {
      String name = InetAddress.getLocalHost().toString();
      int t1 = name.indexOf('.');
      int t2 = name.indexOf('/');
      t1 = (t1 < t2 ? t1 : t2);
      GregorianCalendar gc = new GregorianCalendar();
      StringBuffer buffer = new StringBuffer();
      buffer.append("log-");
      buffer.append(name.substring(0, t1));
      buffer.append("-");
      buffer.append(Integer.toString(gc.get(GregorianCalendar.YEAR)+100).substring(1));
      buffer.append(Integer.toString(gc.get(GregorianCalendar.MONTH)+100).substring(1));
      buffer.append(Integer.toString(gc.get(GregorianCalendar.DAY_OF_MONTH)+100).substring(1));
      buffer.append("-");
      buffer.append(Integer.toString(gc.get(GregorianCalendar.HOUR)+100).substring(1));
      buffer.append(Integer.toString(gc.get(GregorianCalendar.MINUTE)+100).substring(1));
      buffer.append(".txt");
      
      out = new PrintStream(new FileOutputStream(buffer.toString()));
    } catch (Exception e) {
      e.printStackTrace();
    }

    // Perform test
    out.println(comment);
    ClientTest.test1(test, out);
    ClientTest.test2(test, out);
    ClientTest.mtest1(test, out);
    ClientTest.mtest2(test, out);

    completed = true;
    notify();
  }

  public synchronized void halt()
    throws Exception
  {
    while (!completed)
      try { wait(); } catch (Exception e) { }

  }


} // END SpeedTestClient







