/*
 * Copyright (c) 1998-2000 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package test.performance;

import java.rmi.*;
import java.io.*;
import jgroup.*;
import jgroup.registry.*;

import debug.Debug;

/**
 *  Jgroup server for measuring the performance of GMI invocations.
 *  Can be executed remotely through the <t>Executor</t> class,
 *  since it implements the <t>Server</t> interface.
 *
 *  @author       Alberto Montresor
 *  @since        Jgroup 1.1
 */
public class SpeedTestServer
implements SpeedTest, MembershipListener, Server
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Main
  ////////////////////////////////////////////////////////////////////////////////////////////

  public static void main(String[] argv)
    throws Exception
  {
    String hostfile = null;
    int jport = 0;
    int rport = 0;

    for (int i = 0 ; i < argv.length ; i++) {
      if  (argv[i].equals("-jport")) {
        if ((i + 1) == argv.length) {
          System.err.println("Options -jport must be followed by a port number");
          usage();
          return;
        }
        try {
          jport = Integer.parseInt(argv[++i]);
        } catch (Exception e) {
          System.err.println("Value " + argv[i] + " is not a number");
          usage();
          return;
        }
      } else if (argv[i].equals("-rport")) {
        if ((i + 1) == argv.length) {
          System.err.println("Options -rport must be followed by a port number");
          usage();
          return;
        }
        try {
          rport = Integer.parseInt(argv[++i]);
        } catch (Exception e) {
          System.err.println("Value " + argv[i] + " is not a number");
          usage();
          return;
        }
      } else if (argv[i].startsWith("-")) {
        System.err.println("Unknown option " + argv[i]);
        usage();
      } else {
        if (hostfile != null) {
          System.err.println("Hostfile already specified");
          usage();
          return;
        }
        hostfile = argv[i];
      }
    }
    if (hostfile == null) {
      System.err.println("You must specify an hostfile");
      usage();
      return;
    }

    SpeedTestServer server = new SpeedTestServer();
    Object[] args = new Object[] { new DistributedSystem(hostfile),
      new TransportConfig(jport), new Integer(rport) };
    server.start(args);
  }

  private static void usage()
  {
    System.out.println("Usage: speedserver <hostfile>  [-jport <port number>] [-rport <port number>]");
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  private MembershipService membershipService;
  private ExternalGMIService externalGMIService;
  private IID registration;

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  public SpeedTestServer()
  {
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Server
  ////////////////////////////////////////////////////////////////////////////////////////////


  public void start(Object[] args)
    throws Exception
  {
    try {
      if (args == null || args.length != 3)
        throw new IllegalArgumentException("Arguments required: distributed system, transport config, port number");
      DistributedSystem ds = (DistributedSystem) args[0];
      TransportConfig tc = (TransportConfig) args[1];
      int rport = ((Integer) args[2]).intValue();
  
      Object[] configuration = new Object[] { ds, tc };
      Object[] listeners = new Object[] { this };
      GroupManager gm = GroupManager.getGroupManager(configuration, listeners);
      membershipService = (MembershipService) gm.getService(MembershipService.class);
      externalGMIService = (ExternalGMIService) gm.getService(ExternalGMIService.class);
      membershipService.join(12);
      DependableRegistry dr = RegistryFactory.getRegistry(ds, rport);
      registration = externalGMIService.bind("Jgroup/SpeedTest", dr);
      System.out.println("SpeedTestServer registered");
    } catch (Exception e) { 
      e.printStackTrace(); 
    }
  }


  public void halt()
    throws Exception
  {
    // ExternalGMIService.unbind("Jgroup/SpeedTest");
    membershipService.leave();
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from SpeedTest
  ////////////////////////////////////////////////////////////////////////////////////////////

  public void mtest()
    throws RemoteException, McastRemoteException
  {
  }

  public byte[] mtest(byte[] x)
    throws RemoteException, McastRemoteException
  {
    return x;
  }

  public String mtest(String x)
    throws RemoteException, McastRemoteException
  {
    return x;
  }

  public void test()
    throws RemoteException
  {
  }

  public byte[] test(byte[] x)
    throws RemoteException
  {
    return x;
  }

  public String test(String x)
    throws RemoteException
  {
    return x;
  }

  /**
   *  Shut down the server
   */
  public void shutdown()
    throws RemoteException
  {
    new Thread() {
      public void run() {
        try {
          sleep(3000);
        } catch (InterruptedException e) {
        }
        System.exit(0);
      }
    }.start();
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from MembershipListener
  ////////////////////////////////////////////////////////////////////////////////////////////

  public void viewChange(View v)
  {
    System.out.println("- New View ------------------------");
    MemberId[] adds = v.getMembers();
    for (int i=0; i < adds.length ; i++)
      System.out.println(adds[i]);
    System.out.println("-----------------------------------");
  }

  public void leaved()
  {
  }

  public void prepareChange()
  {
  }

} // END SpeedTestServer

